#!/usr/bin/env python

import os
import re

# Need to import these to register all settings;
# correct order is crucial
from vixn import main
from vixn import vixbuf, vixwin, vixview 
from vixn import make, throwback

from vixn import bindings, dirs, excmd, settings, vicflags, vicmd, rc


def _(s): return s


DOC_DATA_DIR = os.path.join(dirs.vixn_dir, "docdata")
DOC_DIR = os.path.join(dirs.vixn_dir, "docs")
HTML_DIR = os.path.join(DOC_DIR, "html")

if not os.path.isdir(HTML_DIR):
    os.makedirs(HTML_DIR)

########################################

def sort_settings(stngs):
    stngs.sort(None, lambda s: s.desc.key)

class Setting(object):
    def __init__(self, desc):
        self.desc = desc
        self.synonyms = []
    
    def add_synonym(self, setting):
        for s in [self] + self.synonyms:
            if not s in setting.synonyms:
                setting.synonyms.append(s)
        sort_settings(setting.synonyms)
        for s in self.synonyms:
            if not setting in s.synonyms:
                s.synonyms.append(s)
                sort_settings(s.synonyms)
        self.synonyms.append(setting)
        sort_settings(self.synonyms)
        

scoped_settings = [[], [], [], []]
all_settings = []

def get_sorted_settings():
    global all_settings, scoped_settings
    for d in settings.registry.settings.values():
        s = Setting(d)
        all_settings.append(s)
        scoped_settings[d.get_scope()].append(s)
    # Have to make second pass in case we encounter a synonym before the desc
    # it's a synonym for
    for s in all_settings:
        if hasattr(s.desc, "synonym_for"):
            d = s.desc.synonym_for
            for s0 in all_settings:
                if s0.desc.key == d.key:
                    s0.add_synonym(s)
                    break
    sort_settings(all_settings)
    for s in scoped_settings:
        sort_settings(s)


###################################

_header = None

def write_header(fp, title):
    global _header
    if not _header:
        hfp = open(os.path.join(DOC_DATA_DIR, "header.html"), 'r')
        _header = hfp.readlines()
        hfp.close()
    for l in _header:
        if "<title>" in l:
            fp.write("<title>" + _("Vixn manual - ") + title + "</title>\n")
        else:
            fp.write(l)
    fp.write("\n<h1>" + title + "</h1>\n\n")


def open_html(leafname, title, divid = None):
    if not leafname.endswith(".html"):
        leafname += ".html"
    fp = open(os.path.join(HTML_DIR, leafname), 'w')
    write_header(fp, title)
    if divid:
        fp.write('<div id="%s">\n' % divid)
    return fp


def close_html(fp, divid = None):
    if divid:
        fp.write('</div> <!-- id="%s" -->\n' % divid)
    fp.write('</body>\n</html>\n')
    fp.close()


def format_help(s_in):
    s_out = ""
    n = 0
    l = len(s_in)
    while n < l:
        c = s_in[n]
        if c == '[':
            link = ""
            n += 1
            while 1:
                c = s_in[n]
                if c == ']':
                    break
                link += c
                n += 1
            link = link.split(':')
            s_out += '<a href="%s">%s</a>' % \
                    (link[0] + "settings.html#" + link[1], link[1])
        else:
            s_out += c
        n += 1
    return s_out
                

def tabulate_items(fp, items, printer, columns = None):
    COLUMNS = 4
    
    if not columns:
        columns = COLUMNS
    
    fp.write('<table class="index">\n')
    col = 0
    for i in items:
        if col == 0:
            fp.write('<tr>\n')
        fp.write('<td>')
        printer(fp, i)
        fp.write('</td>\n')
        col += 1
        if col == columns:
            fp.write('</tr>\n')
            col = 0
    if col < columns:
        while col < columns:
            fp.write('<td> </td>\n')
            col += 1
        fp.write('</tr>\n')
    fp.write('</table>\n')
    

#####################################

SCOPE_NAMES_FOR_URLS = [ "glob", "buf", "view", "win" ] 

LONG_SCOPE_NAMES = [ \
        _("Global scope"),
        _("Buffer scope"),
        _("View scope"),
        _("Window scope") \
]

def url_for_setting(desc):
    return SCOPE_NAMES_FOR_URLS[desc.get_scope()] + \
            "_settings.html#" + desc.get_key()


def tabulate_settings(fp, stngs):
    
    def print_setting(fp, s):
        fp.write('<a href="' + url_for_setting(s.desc) + '">' + \
                s.desc.key + '</a>')
        if hasattr(s.desc, "synonym_for"):
            fp.write(' <a href="#synonym" class="footnote">*</a>')
    
    tabulate_items(fp, stngs, print_setting)
    

def scoped_settings_index():
    fp = open_html("scoped_settings",
            _("Index of settings grouped by scope"),
            "scoped_settings")
    global scoped_settings
    for i in range(settings.NUM_SCOPES):
        fp.write('<h2>' + LONG_SCOPE_NAMES[i] + '</h2>\n')
        tabulate_settings(fp, scoped_settings[i])
    
    fp.write('</div> <!-- id="scoped_settings" -->\n')
    
    fp.write('<p id="synonym_footnote">')
    fp.write('<span id="synonym">*</span> ')
    fp.write('Settings marked thus are synonyms</p>\n')
    
    close_html(fp)


def detailed_settings(scope):
    global scoped_settings
    sn = SCOPE_NAMES_FOR_URLS[scope]
    fp = open_html("%s_settings" % sn,
            _("Settings: %s") % LONG_SCOPE_NAMES[scope],
            "%s_settings" % sn)
    for s in scoped_settings[scope]:
        if hasattr(s.desc, "synonym_for"):
            continue
        fp.write('<div id="%s" class="defn">\n' % s.desc.key)
        fp.write('<h2>' + s.desc.key)
        if s.desc.auto_save:
            fp.write('<span class="autosave">' + _(' (auto-saved)') + '</span>')
        fp.write('</h2>\n')
        fp.write('<p class="type">')
        fp.write('<span class="defn_label">' + _("Python type: ") + '</span>')
        t = str(s.desc.type)
        t = t[t.index("'") + 1 : -2]
        fp.write(t + '</p>\n')
        if s.desc.description:
            fp.write('<p>' + format_help(s.desc.description) + '</p>\n')
        if s.synonyms:
            fp.write('<p class="synonyms">')
            fp.write('<span class="defn_label">' + _("Synonyms: ") + '</span>')
            syns = ""
            for sy in s.synonyms:
                syns += sy.desc.key + ', '
            fp.write(syns[:-2])
            fp.write('</p>\n')
        fp.write('</div>\n')
    close_html(fp, '%s_settings' % sn)
                

###########################################

ex_commands = excmd.commands.items()
ex_commands.sort(None, lambda x: x[0])
    
def print_ex_commands():
    global ex_commands
    fp = open_html("ex_commands", _("Ex commands"), "excommands")
    for ex in ex_commands:
        fp.write('<div id="%s" class="defn">\n' % ex[0])
        fp.write('<h2>:' + ex[0] + '</h2>\n')
        if ex[1].help:
            fp.write('<p>' + ex[1].help + '</p>\n')
        fp.write('</div>\n')
    close_html(fp, "excommands")

###########################################

# User's PoV of vi command flags is different from system's
COMMAND = 0
CHANGE = 1
INSERT = 2
MOVEMENT = 3
SELECT = 4
COMPOUND = 5
NUM_VI_TYPES = 6

FLAG_NAMES_FOR_URLS = [ "std", "change", "ins", "mv", "sel", "compound" ] 
LONG_FLAG_NAMES = [ \
        _("Ordinary commands"),
        _("Commands that change text"),
        _("Commands that start an interactive edit (insertion)"),
        _("Commands that move the cursor"),
        _("Commands that start a selection"),
        _("Commands that change the text then start an insertion") \
]

vi_commands_by_flags = [ [], [], [], [], [], [] ]
all_vi_commands = []

def get_all_vi_commands():
    global all_vi_commands, vi_commands_by_flags
    vicmds = vicmd.commands.values()
    vicmds.sort(None, lambda x: x.method.func_name)
    for vc in vicmds:
        if vc.flags & vicflags.COMPOUND:
            f = COMPOUND
        elif vc.flags & vicflags.SELECT:
            f = SELECT
        elif vc.flags & vicflags.INSERT:
            f = INSERT
        elif vc.flags & vicflags.CHANGE:
            f = CHANGE
        elif vc.flags & vicflags.MOVEMENT:
            f = MOVEMENT
        else:
            f = COMMAND
        all_vi_commands.append([vc, f])
        vi_commands_by_flags[f].append(vc)


def vi_commands_indexes():
    global vi_commands_by_flags
    fp = open_html("sorted_vi_commands",
            _("Index of vi commands grouped by type"),
            "sorted_vi_commands")
    global scoped_settings
    for i in range(NUM_VI_TYPES):
        def print_vi(fp, vc):
            name = vc.method.func_name
            fp.write('<a href="vi_%s_commands.html#%s">%s</a>' % \
                    (FLAG_NAMES_FOR_URLS[i], name, name))
        fp.write('<h2>' + LONG_FLAG_NAMES[i] + '</h2>\n')
        tabulate_items(fp, vi_commands_by_flags[i], print_vi)
    
    close_html(fp, "sorted_vi_commands")


def detailed_vi_commands(flag, keystrokes):
    global vi_commands_by_flags
    fn = FLAG_NAMES_FOR_URLS[flag]
    fp = open_html("vi_%s_commands.html" % fn,
            LONG_FLAG_NAMES[flag],
            "vi_%s_flags" % fn)
    for vc in vi_commands_by_flags[flag]:
        name = vc.method.func_name
        fp.write('<div id="%s" class="defn">\n' % name)
        fp.write('<h2>' + name + '</h2>\n')
        if vc.help:
            fp.write('<p>' + format_help(vc.help) + '</p>\n')
        if vc.dflt_key:
            fp.write('<p class="dflt_key">')
            fp.write('<span class="defn_label">' + \
                    _("Default keystroke(s): ") + '</span>')
            keys = vc.dflt_key
            if hasattr(keys, 'isspace'):
                keys = [keys]
            ks = ""
            ors = '<span class="ks_or"> or </span>'
            for k in keys:
                k = re.sub(r'\\\\', r'\\', k)
                k = re.sub(r'\&', '&amp;', k)
                k = re.sub('<', '&lt;', k)
                k = re.sub('>', '&gt;', k)
                k = re.sub(' ', '&lt;Space&gt;', k)
                ks += ('<span class="keystroke">%s</span>' % k) + ors
                keystrokes.append([k, vc.method.func_name, flag])
            ks = ks[:-len(ors)]
            fp.write(ks + '</p>\n')
        fp.write('</div>\n')
    
    close_html(fp, "vi_%s_commands.html" % fn)
                
###########################################

def print_functions():
    global ex_commands
    fp = open_html("functions", _("Public python functions"), "functions")
    for f in rc.public_functions:
        fp.write('<div id="%s" class="defn">\n' % f[0])
        fp.write('<h2>' + f[0] + '(')
        comma = False
        for a in f[1]:
            if comma:
                fp.write(', ')
            comma = True
            fp.write(a)
        fp.write(')</h2>\n')
        fp.write('<p>' + f[2] + '</p>\n')
        fp.write('</div>\n')
    close_html(fp, "functions")

###########################################

def master_index(keystrokes):
    fp = open_html("index2", _("Main index"))
    global all_settings
    fp.write('<div id="settings_index">\n')
    fp.write('<h2>' + _("All settings") + '</h2>\n')
    tabulate_settings(fp, all_settings)
    fp.write('</div> <!-- id="settings_index" -->\n')
    fp.write('<p id="synonym_footnote">')
    fp.write('<span id="asterisk">*</span> ')
    fp.write('Settings marked thus are synonyms</p>\n')
    
    global all_vi_commands
    def print_vi(fp, vc):
        name = vc[0].method.func_name
        fp.write('<a href="vi_%s_commands.html#%s">%s</a>' % \
                (FLAG_NAMES_FOR_URLS[vc[1]], name, name))
    fp.write('<div id="vi_index">\n')
    fp.write('<h2>' + _("Vi commands") + '</h2>\n')
    tabulate_items(fp, all_vi_commands, print_vi)
    fp.write('</div> <!-- id="vi_index" -->\n')
    
    global ex_commands
    def print_ex(fp, e):
        fp.write(':<a href="ex_commands.html#%s">%s</a>' % (e[0], e[0]))
    fp.write('<div id="ex_index">\n')
    fp.write('<h2>' + _("Ex commands") + '</h2>\n')
    tabulate_items(fp, ex_commands, print_ex)
    fp.write('</div> <!-- id="ex_index" -->\n')
    
    fp.write('<div id="keybindings">\n')
    fp.write('<h2>' + _("Default vi key bindings") + '</h2>\n')
    keystrokes.sort(None, lambda x: x[0])
    def print_key_link(fp, key):
        fp.write('<a href="vi_%s_commands.html#%s">%s</a>' % \
                (FLAG_NAMES_FOR_URLS[key[2]], key[1], key[0]))
    tabulate_items(fp, keystrokes, print_key_link)
    fp.write('</div> <!-- id="keybindings" -->\n')
    
    fp.write('<div id="functions">\n')
    fp.write('<h2>' + _("Public python functions") + '</h2>\n')
    def print_function(fp, f):
        fp.write('<a href="functions.html#%s">%s</a>' % (f[0], f[0]))
    tabulate_items(fp, rc.public_functions, print_function)
    fp.write('</div> <!-- id="functions" -->\n')
    
    close_html(fp)


###########################################

get_sorted_settings()
get_all_vi_commands()
scoped_settings_index()
for scope in range(settings.NUM_SCOPES):
    detailed_settings(scope)
vi_commands_indexes()
keystrokes = []
for f in range(NUM_VI_TYPES):
    detailed_vi_commands(f, keystrokes)
print_ex_commands()
print_functions()
rc.public_functions.sort()
master_index(keystrokes)

