from gi.repository import GLib, GObject
from gi.repository import Gdk, Gtk

from . import history
from . import mode

PADDING = 4

""" Each bottombar widget must have a special hide() method which takes
a restore_focus argument so that focusable widgets know whether to restore
focus to window's main (view) widget when being hidden. """

class SingleStatusbar(Gtk.Statusbar):
    """ Provides a Statusbar with a set_text() method. """
    def __init__(self):
        Gtk.Statusbar.__init__(self)
        def find_label(w):
            if isinstance(w, Gtk.Label):
                return w
            elif hasattr(w, "get_children"):
                for child in w.get_children():
                    label = find_label(child)
                    if label:
                        return label
            return None
        self.label = find_label(self)

    def set_text(self, text, bold = False):
        if bold:
            self.label.set_markup("<b>%s</b>" % GLib.markup_escape_text(text))
        else:
            self.label.set_text(text)

    def hide(self, arg = None):
        Gtk.Statusbar.hide(self)


class StatusBar(Gtk.HBox):
    """ An HBox containing 4 status bar fields: Mode, Filetype,
    Keys pressed, Position """
    def __init__(self):
        GObject.GObject.__init__(self)
        self.set_homogeneous(True)
        self.mode = SingleStatusbar()
        self.filetype = SingleStatusbar()
        self.key_string = SingleStatusbar()
        self.position = SingleStatusbar()
        self.pack_start(self.mode, True, True, PADDING)
        self.pack_start(self.filetype, True, True, PADDING)
        self.pack_start(self.key_string, True, True, PADDING)
        self.pack_start(self.position, True, True, PADDING)
        self.mode.show()
        self.filetype.show()
        self.key_string.show()
        self.position.show()
        self.filetype_s = ""
        self.dos = False
        self.important = False

    def __replace_text(self, which, text, bold = False):
        widget = getattr(self, which)
        widget.set_text(text, bold)

    def set_mode(self, new_mode = None):
        """ See vixview for valid modes. """
        if new_mode == None:
            self.mode.set_text("")
            return
        bold = True
        ins = new_mode & mode.INSERT_MASK
        sel = new_mode & mode.SELECT_MASK
        ro = new_mode & mode.RO_MASK
        self.important = False
        if ins == mode.INSERT:
            s = _("INSERT ")
            self.important = True
        elif ins == mode.OVERWRITE:
            s = _("OVRWRITE ")
            self.important = True
        elif not sel:
            s = _("vi ")
            bold = False
        else:
            s = ""
        if sel == mode.SELECT:
            s += _('SELECT ')
        elif sel == mode.SELBYLINE:
            s += _('SELECT LINES ')
        elif sel == mode.SELRECT:
            s += _('SELECT RECT ')
        if new_mode & mode.RECORDING:
            s += _("*REC* ")
            bold = True
            self.important = True
        if ro:
            s += _('RO')
        self.__replace_text("mode", s, bold)

    def __show_filetype_and_dos(self):
        t = self.filetype_s
        if self.dos:
            t += " (DOS)"
        self.__replace_text("filetype", t)

    def set_filetype(self, filetype):
        self.filetype_s = filetype
        self.__show_filetype_and_dos()

    def set_dos(self, dos):
        self.dos = dos
        self.__show_filetype_and_dos()

    def set_key_string(self, keys = None):
        if not keys:
            keys = ""
        self.__replace_text("key_string", keys)

    def set_position(self, column, line):
        """ Input values start from 0, but are displayed starting from 1
        ie they have +1 added. """
        self.__replace_text("position",
                _("Ln %d, Col %d") % (line + 1, column + 1))

    def set_position_from_iter(self, iter):
        self.set_position(iter.get_visible_line_offset(), iter.get_line())

    def hide(self, arg = None):
        Gtk.HBox.hide(self)


class Focusable(object):
    """ Base class for bottombar widgets which handle focus. """
    def __init__(self):
        self.focus_handler_id = None

    def take_focus(self, view):
        """ If view is given focus, the bottombar operation is
        cancelled. """
        if self.focus_handler_id != None:
            self.view.disconnect(self.focus_handler_id)
        self.view = view
        self.focus_handler_id = view.connect("focus-in-event",
                self.focus_lost_handler)
        # Grabbing focus causes selection to be cleared
        buf = view.get_buffer()
        selection = buf.get_selection_bounds()
        self.grab_focus()
        if selection:
            buf.select_range(*selection)

    def focus_lost_handler(self, widget, event):
        self.cancel()

    def disconnect_focus(self):
        if self.focus_handler_id != None:
            self.view.disconnect(self.focus_handler_id)
            self.focus_handler_id = None

    def cancel(self):
        self.disconnect_focus()

    def hide(self, restore_focus = True):
        """ You should override this in derived classes which also derive from
        widgets to make sure Gtk.Widget.hide doesn't override this. """
        self.disconnect_focus()
        Gtk.Widget.hide(self)
        if restore_focus:
            self.view.grab_focus()


class InputBar(history.TabEntry, Focusable):
    """ A text entry field temporarily displayed at the bottom of the window
    in place of the status bar. """
    def __init__(self, activate_callback = None, cancelled_callback = None):
        """ When activated (Enter pressed) activate_callback is called with
        contents of widget as arg.
        cancelled_callback is a function to be called when
        the entry is cancelled by everything in it being deleted or Esc
        pressed. It is responsibility of either callback to hide widget. """
        history.TabEntry.__init__(self, "")
        Focusable.__init__(self)
        self.set_has_frame(False)
        self.activate_callback = activate_callback
        self.cancelled_callback = cancelled_callback
        self.connect("activate", self.__activate)
        self.connect("changed", self.__changed)
        self.connect("key-press-event", self.__key_press)
        self.connect("parent-set", self.__parent_set)

    def start(self, view, cwd, initial, hist):
        """ Shows the widget with initial char. If view gets focus while
        InputBar is showing, InputBar is cancelled. """
        self.attach_history(hist)
        self.set_prefix(initial[0])
        self.reset_matches(cwd)
        self.set_text_unselected(initial)
        self.take_focus(view)

    def cancel(self):
        Focusable.cancel(self)
        if self.cancelled_callback:
            self.cancelled_callback()

    def __activate(self, widget):
        self.save_history()
        if self.activate_callback:
            self.activate_callback(self.get_text())
        else:
            self.cancel()

    def __changed(self, widget):
        if not self.ignore_changes and not self.get_text():
            self.cancel()

    def __key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.cancel()
            return True
        return False

    def __parent_set(self, widget, old_parent):
        if not old_parent:
            settings = self.get_settings()
            settings.set_string_property("gtk-key-theme-name", "Emacs", "vixn")

    def hide(self, restore_focus = True):
        Focusable.hide(self, restore_focus)


class QueryBar(Gtk.HBox, Focusable):
    """ Shows a prompt and gets answers one key at a time. Automatically gets
    cancelled if a key is pressed. """
    def __init__(self, key_callback = None, cancelled_callback = None):
        """ When activated (Enter pressed) activate_callback is called with
        contents of widget as arg.
        cancelled_callback is a function to be called when
        the entry is cancelled by everything in it being deleted or Esc
        pressed. It is responsibility of either callback to hide widget. """
        GObject.GObject.__init__(self)
        self.set_homogeneous(False)
        self.set_spacing(PADDING)
        Focusable.__init__(self)
        self.prompt = Gtk.Label()
        self.prompt.set_use_markup(True)
        self.entry = Gtk.Entry()
        self.entry.set_has_frame(False)
        self.key_callback = key_callback
        self.cancelled_callback = cancelled_callback
        self.entry.connect("key-press-event", self.__key_press)
        self.pack_start(self.prompt, False, True, 0)
        self.pack_start(self.entry, True, True, 0)
        self.prompt.show()
        self.entry.show()

    def start(self, view, prompt):
        self.prompt.set_markup(prompt)
        self.take_focus(view)

    def cancel(self):
        Focusable.cancel(self)
        if self.cancelled_callback:
            self.cancelled_callback()

    def __key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.cancel()
        elif self.key_callback:
            self.key_callback(event.get_state(), event.keyval, event.string)
        return True

    def hide(self, restore_focus = True):
        Focusable.hide(self, restore_focus)

    def grab_focus(self):
        self.entry.grab_focus()


class MessageBar(SingleStatusbar):
    """ For briefly showing a short message. """
    pass


class BottomBar(Gtk.HBox):
    """ Displays either the status bar (default), input bar or output bar,
    automatically switching between them as necessary. Also acts as proxy
    for their function calls. """
    def __init__(self):
        GObject.GObject.__init__(self)
        self.set_homogeneous(False)
        self.set_spacing(8)
        self.status_bar = StatusBar()
        self.input_bar = InputBar(self.__input_activate, self.__cancel_input)
        self.input_callback = None
        self.message_bar = MessageBar()
        self.query_bar = QueryBar(self.__query_key_press, self.__query_cancel)
        self.query_key_cb = None
        self.query_cancel_cb = None
        self.query_args = []
        self.showing = None
        self.pack_start(self.status_bar, True, True, 0)
        self.pack_start(self.input_bar, True, True, 0)
        self.pack_start(self.message_bar, True, True, 0)
        self.pack_start(self.query_bar, True, True, 0)
        # Need to wait for a couple of idles before allowing status to replace
        # message, because status gets restored on mark-set and we can't
        # help that, so kludge around it
        self.sentinel_tag = None
        self.message_sentinel = 0
        self.__show_only()
        self.connect("destroy", self.__cancel_sentinel)

    def __wname(self, w):
        if w == self.status_bar:
            return "status"
        elif w == self.message_bar:
            return "message"
        elif not w:
            return "None"
        else:
            return "other"

    def __dec_sentinel(self, data):
        self.message_sentinel -= 1
        if self.message_sentinel == 0:
            self.sentinel_tag = None
        return self.message_sentinel != 0

    def __show_only(self, widget = None):
        if not widget:
            widget = self.status_bar
        if widget != self.showing:
            for w in self.status_bar, self.input_bar, \
                    self.message_bar, self.query_bar:
                if w == widget:
                    w.show()
                elif w == self.showing:
                    # Don't restore focus to view if we're replacing one
                    # Focusable with another
                    w.hide(not isinstance(widget, Focusable))
            self.showing = widget
        if widget == self.message_bar:
            if self.sentinel_tag == None:
                self.sentinel_tag = GLib.idle_add(self.__dec_sentinel, None)
            self.message_sentinel = 2
        else:
            self.__cancel_sentinel()

    def __cancel_sentinel(self, gobj = None):
        if self.sentinel_tag != None:
            GObject.source_remove(self.sentinel_tag)
            self.sentinel_tag = None
            self.message_sentinel = 0

    def __input_activate(self, value):
        self.__show_only(self.status_bar)
        if self.input_callback:
            if self.input_callback_args:
                self.input_callback(value, *self.input_callback_args)
            else:
                self.input_callback(value)

    def __cancel_input(self):
        self.__show_only(self.status_bar)

    def __query_cancel(self):
        self.__show_only(self.status_bar)
        if self.query_cancel_cb:
            if self.query_args:
                self.query_cancel_cb(*self.query_args)
            else:
                self.query_cancel_cb()

    def __query_key_press(self, state, keyval, keystring):
        if self.query_key_cb:
            if self.query_args:
                self.query_key_cb(state, keyval, keystring, *self.query_args)
            else:
                self.query_key_cb(state, keyval, keystring)

    def restore_status(self, force = False):
        """ Replaces message_bar with status_bar but leaves others in place
        unless force is True. """
        if force or (self.showing == self.message_bar and \
                (self.status_bar.important or not self.message_sentinel)):
            self.__show_only(self.status_bar)

    def set_mode(self, new_mode = None):
        self.status_bar.set_mode(new_mode)
        self.restore_status()

    def set_filetype(self, filetype = None):
        self.status_bar.set_filetype(filetype)
        self.restore_status()

    def set_dos(self, dos = True):
        self.status_bar.set_dos(dos)
        self.restore_status()

    def set_key_string(self, keys = None):
        self.status_bar.set_key_string(keys)
        self.restore_status()

    def set_position(self, column, line):
        self.status_bar.set_position(column, line)
        self.restore_status()

    def set_position_from_iter(self, iter):
        self.restore_status()
        self.status_bar.set_position_from_iter(iter)

    def open_input(self, view, cwd, initial, callback, *args):
        self.input_callback = callback
        self.input_callback_args = args
        self.__show_only(self.input_bar)
        ic = initial[0]
        if ic == '/' or ic == '\\' or ic == '?':
            hist = history.searches
        else:
            hist = history.commands
        self.input_bar.start(view, cwd, initial, hist)

    def show_message(self, text, immediate = False):
        if self.showing == self.status_bar and self.status_bar.important:
            return
        self.message_bar.set_text(text)
        self.__show_only(self.message_bar)

    def open_query(self, view, *args):
        """ First 3 *args should be prompt, key_press_cb, cancel_cb. """
        if args:
            prompt = args[0]
        if len(args) > 1:
            self.query_key_cb = args[1]
        else:
            self.query_key_cb = None
        if len(args) > 2:
            self.query_cancel_cb = args[2]
        else:
            self.query_cancel_cb = None
        if len(args) > 3:
            self.query_args = args[3:]
        else:
            self.query_args = args
        self.__show_only(self.query_bar)
        self.query_bar.start(view, prompt)

