""" Keeps track of the most recent change for the repeat ('.') command. """

last = None

class Change(object):
    def __init__(self, buf, keychain = None, seltype = 0, selsize = 0):
        """ keychain of None is basic insert - in case i has been rebound """
        self.buffer = buf
        self.keychain = keychain
        self.seltype = seltype
        self.selsize = selsize
        self.deleted_before = 0
        self.deleted_after = 0
        self.inserted = None
        self.start_mark = None
        self.end_mark = None
        self.delete_cb_tag = None
        self.mark_cb_tag = None
        self.changed_cb_tag = None
        self.no_change = keychain == None
    
    def prepare_for_insert(self):
        """ Note we don't call begin_user_action() because the insert may be
        part of a compound operation, so leave responsibility to vicmd.
        Nor does finalise call end_user_action(). """
        buf = self.buffer
        iter = buf.get_iter_at_insert()
        self.start_mark = buf.create_mark(None, iter, True)
        self.end_mark = buf.create_mark(None, iter, False)
        self.delete_cb_tag = self.buffer.connect("delete-range",
                self.delete_range)
        self.mark_cb_tag = self.buffer.connect("mark-set", self.mark_set)
        if self.no_change:
            self.changed_cb_tag = self.buffer.connect("changed", self.changed)
    
    def finalise(self):
        buf = self.buffer
        if not buf:
            return
        buf.change = None
        if self.start_mark and self.end_mark:
            self.inserted = buf.get_text(buf.get_iter_at_mark(self.start_mark),
                    buf.get_iter_at_mark(self.end_mark))
        if self.start_mark:
            buf.delete_mark(self.start_mark)
            self.start_mark = None
        if self.end_mark:
            buf.delete_mark(self.end_mark)
            self.end_mark = None
        if self.delete_cb_tag != None:
            buf.disconnect(self.delete_cb_tag)
            self.delete_cb_tag = None
        if self.mark_cb_tag != None:
            buf.disconnect(self.mark_cb_tag)
            self.mark_cb_tag = None
        if self.changed_cb_tag != None:
            buf.disconnect(self.changed_cb_tag)
            self.changed_cb_tag = None
        if not self.no_change:
            global last
            last = self
    
    def delete_range(self, buf, start, end):
        if not end.compare(buf.get_iter_at_mark(self.start_mark)):
            self.deleted_before += len(buf.get_text(start, end))
        elif not start.compare(buf.get_iter_at_mark(self.end_mark)):
            self.deleted_after += len(buf.get_text(start, end))

    def mark_set(self, buf, iter, mark):
        if buf.ignore_mark_set:
            return
        # If cursor moves to separate part of buf without coming out of
        # insert mode, reset change to treat this as a separate insert
        # operation
        if mark.get_name() == 'insert':
            if iter.compare(buf.get_iter_at_mark(self.start_mark)) < 0 or \
                    iter.compare(buf.get_iter_at_mark(self.end_mark)) > 0:
                buf.end_user_action()
                self.finalise()
                buf.change = Change(buf)
                buf.begin_user_action()
                buf.change.prepare_for_insert()
    
    def changed(self, buf):
        self.no_change = False
        buf.disconnect(self.changed_cb_tag)
        self.changed_cb_tag = None
    
    def repeat_deletions(self):
        """ Deletes deleted_before and deleted_after chars from either side of
        current insert point. """
        buf = self.buffer
        ins = buf.get_iter_at_insert()
        start = None
        end = None
        if self.deleted_before:
            start = ins.copy()
            start.backward_cursor_positions(self.deleted_before)
        if self.deleted_after:
            end = ins.copy()
            end.forward_cursor_positions(self.deleted_after)
        if start or end:
            if not start:
                start = ins
            elif not end:
                end = ins
            buf.delete(start, end)
