""" All sorts of dialog boxes. """

import locale
import os
import sys
import traceback

from gi.repository import GdkPixbuf, GObject, Gtk
from gi.repository import Gtk

from . import buffers
from . import dirs
from . import msgs
from . import misc
from . import version

_EXC_DIAG_WIDTH = 480
_TRACEBACK_HEIGHT = 240

_F = Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT

""" DCS (Discard/Cancel/Save) Response constants. """
DCS_CANCEL = 1
DCS_SAVE = 2
DCS_DISCARD = 3

def run_dcs_dialog(dialog, save_function, discard_function):
    """ Runs a DCS dialog for all buffers and takes appropriate action on
    response. """
    response = dialog.run()
    if response == DCS_SAVE:
        save_function()
    if response == DCS_DISCARD or response == DCS_SAVE:
        discard_function()
    return response


def add_no_repeat(box, label, handler, use_underline = True,
        expand = True, fill = True, padding = 0):
    """ Creates a checkbutton with given label and connects signal handler. Adds
    checkbutton to box container. Returns checkbutton widget. """
    button = Gtk.CheckButton(label, use_underline)
    button.connect("toggled", handler)
    button.show()
    box.pack_start(button, expand, fill, padding)
    return button


class ADialog(Gtk.Dialog):
    """ Mimic GtkDialog's constructor which we can't use in derived
    python classes. """
    def __init__(self, title = None, parent = None, flags = 0, buttons = None):
        Gtk.Dialog.__init__(self)
        if title:
            self.set_title(title)
        if parent:
            self.set_transient_for(parent)
        self.set_modal(flags & Gtk.DialogFlags.MODAL)
        self.set_destroy_with_parent( \
                flags & Gtk.DialogFlags.DESTROY_WITH_PARENT)
        if buttons:
            self.add_buttons(buttons)



class Exception(ADialog):
    """ Base class for reporting exception. Creates dialog
    with DCS buttons. """
    def __init__(self, exc_info, fatal = False, parent_win = None,
            message = None, no_repeat = None):
        """ no_repeat is tuple (label, handler, use_underline) for
        add_no_repeat().  use_underline is optional, default is True. """
        if msgs.recursive_exception:
            msgs.fatal_exception(message)
        self.message = message
        try:
            self.exc_info = exc_info
            ltxt = str(self.exc_info[1])

            ADialog.__init__(self,  _("Vixn Exception"), parent_win, _F)

            button = misc.custom_stock_button(_("_Quit without saving"),
                    Gtk.STOCK_QUIT)
            self.add_action_widget(button, DCS_DISCARD)

            button = misc.custom_stock_button(_("_Save All + Quit"),
                    Gtk.STOCK_SAVE)
            self.add_action_widget(button, DCS_SAVE)
            if not buffers.any_modified():
                button.set_sensitive(False)

            button = misc.custom_stock_button(_("_Continue"), Gtk.STOCK_OK)
            self.add_action_widget(button, DCS_CANCEL)
            if fatal:
                button.set_sensitive(False)

            if fatal:
                self.set_default_response(Gtk.ResponseType.OK)
            else:
                self.set_default_response(Gtk.ResponseType.CLOSE)

            hbox = Gtk.HBox()
            img = Gtk.Image()
            if fatal:
                stock = Gtk.STOCK_DIALOG_ERROR
            else:
                stock = Gtk.STOCK_DIALOG_WARNING
            img.set_from_stock(stock, Gtk.IconSize.DIALOG)
            hbox.pack_start(img, False, False, 10)

            vbox = Gtk.VBox()
            self.sub_vbox = vbox

            label = Gtk.Label(label=_('<b>Vixn exception</b>'))
            label.set_use_markup(True)
            label.set_alignment(0, 0.5)
            vbox.pack_start(label, True, True, 8)

            if not ltxt and not message:
                message = _("An exception has occurred")
            if message:
                label = Gtk.Label(label=message)
                label.set_alignment(0, 0)
                label.set_line_wrap(True)
                label.set_size_request(_EXC_DIAG_WIDTH, -1)
                vbox.pack_start(label, True, True, 8)

            if ltxt:
                label = Gtk.Label(label=ltxt)
                label.set_alignment(0, 0)
                label.set_line_wrap(True)
                label.set_size_request(_EXC_DIAG_WIDTH, -1)
                vbox.pack_start(label, True, True, 8)

            if no_repeat:
                if len(no_repeat) == 2:
                    no_repeat.append(True)
                add_no_repeat(vbox, no_repeat[0], no_repeat[1], no_repeat[2],
                        False, False, 8)

            button = Gtk.Button(_("_Details >>"))
            alignment = Gtk.Alignment.new(1, 0.5, 0, 1)
            alignment.add(button)
            vbox.pack_start(alignment, False, False, 4)
            button.connect("clicked", self.__details_clicked)

            self.win_size = None
            self.traceback = None

            hbox.pack_start(vbox, True, True, 10)
            self.get_content_area().pack_start(hbox, True, True, 0)
        except:
            msgs.fatal_exception(message)

    def run(self):
        if msgs.recursive_exception:
            msgs.fatal_exception(self.message)
        try:
            response = Gtk.Dialog.run(self)
            self.traceback = None
        except:
            msgs.fatal_exception(message)
        return response

    def __details_clicked(self, button):
        if not self.traceback:
            etxt = traceback.format_exception(self.exc_info[0],
                            self.exc_info[1], self.exc_info[2]) + \
                    ["***********\n"] + traceback.format_stack()
            buf = Gtk.TextBuffer()
            buf.set_text("\n".join(etxt))
            view = Gtk.TextView.new_with_buffer(buf)
            view.set_editable(False)
            view.set_cursor_visible(False)
            view.set_left_margin(4)
            view.set_right_margin(4)
            scrport = Gtk.ScrolledWindow()
            scrport.set_policy(Gtk.PolicyType.AUTOMATIC,
                    Gtk.PolicyType.AUTOMATIC)
            scrport.set_shadow_type(Gtk.ShadowType.IN)
            scrport.add(view)
            sph = _TRACEBACK_HEIGHT
            vh = view.get_preferred_height()[1] + 4
            if sph > vh:
                sph = vh
            scrport.set_size_request(_EXC_DIAG_WIDTH, sph)
            self.sub_vbox.pack_start(scrport, True, True, 4)
            self.traceback = scrport
        if self.traceback.get_property("visible"):
            self.traceback.hide()
            button.set_label(_("_Details >>"))
            if self.win_size:
                self.resize(*self.win_size)
        else:
            self.win_size = self.get_size()
            self.traceback.show_all()
            button.set_label(_("_Details <<"))

def _exception(exc_info = None, fatal = False, parent_win = None,
        message = None, no_repeat = None):
    if not exc_info:
        exc_info = sys.exc_info()
    for l in traceback.format_exception(exc_info[0], exc_info[1], exc_info[2]):
        sys.stderr.write(l + " ")
    exc = Exception(exc_info, fatal, parent_win, message, no_repeat)
    exc.show_all()
    run_dcs_dialog(exc,
            lambda: buffers.save_all(True),
            lambda: sys.exit(0))
    exc.destroy()

def exception(parent_win = None, message = None, no_repeat = None):
    _exception(None, False, parent_win, message, no_repeat)

def fatal_exception(parent_win = None, message = None, no_repeat = None):
    _exception(None, True, parent_win, message, no_repeat)

def excepthook(type, value, tb):
    _exception([type, value, tb])


def message_dialog(parent_win = None, type = Gtk.MessageType.INFO,
        buttons = None, message = None):
    if not buttons:
        buttons = Gtk.ButtonsType.CLOSE
    dialog = Gtk.MessageDialog(parent_win, _F, type, buttons, message)
    t = None
    if type == Gtk.MessageType.INFO:
        t = _("Message from Vixn")
    elif type == Gtk.MessageType.WARNING:
        t = _("Warning from Vixn")
    elif type == Gtk.MessageType.ERROR:
        t = _("Error from Vixn")
    if t:
        dialog.set_title(t)
    response = dialog.run()
    dialog.destroy()
    return response

def error(parent_win = None, message = None):
    return message_dialog(parent_win, Gtk.MessageType.ERROR,
            None, message)

def warning(parent_win = None, message = None):
    return message_dialog(parent_win, Gtk.MessageType.WARNING,
            None, message)

def _create_file_chooser(title = None, parent_win = None, action = None,
        buttons = None, default_file = None):
    """ If default_file ends with / it's used for set_current_folder(),
    otherwise set_current_name(). """
    if action == None:
        action = Gtk.FileChooserAction.OPEN
    dialog = Gtk.FileChooserDialog(title, parent_win, action, buttons)
    if default_file:
        if default_file.endswith(os.path.sep):
            dialog.set_current_folder(default_file)
        else:
            dialog.set_current_name(str(default_file))
    return dialog

def choose_file(title = None, parent_win = None,
        action = None, button_label = None, default_file = None):
    """ If default_file ends with / it's used for set_current_folder(),
    otherwise set_current_name().
    button_label should be Gtk.STOCK_OPEN (default) or ..._SAVE;
    it will be assigned Gtk.ResponseType.ACCEPT.
    Any response other than Cancel is presumed to mean use this file.
    Returns chosen file or None. """
    if not button_label:
        button_label = Gtk.STOCK_OPEN
    dialog = _create_file_chooser(title, parent_win, action,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            button_label, Gtk.ResponseType.ACCEPT), default_file)
    if default_file:
        if default_file.endswith(os.path.sep):
            dialog.set_current_folder(default_file)
        else:
            dialog.set_current_name(str(default_file))
    if dialog.run() != Gtk.ResponseType.CANCEL:
        filename = dialog.get_filename()
    else:
        filename = None
    dialog.destroy()
    return filename

def get_filename_open(title = None, parent_win = None,
        default_file = None):
    if not title:
        title = _("Edit file")
    return choose_file(title, parent_win,
            Gtk.FileChooserAction.OPEN, Gtk.STOCK_OPEN, default_file)

def get_filename_save(title = None, parent_win = None,
        default_file = None):
    if not title:
        title = _("Save As")
    return choose_file(title, parent_win,
            Gtk.FileChooserAction.SAVE, Gtk.STOCK_SAVE_AS, default_file)


def _dcs_dialog(parent_win = None, buttons = None, title = None,
        message = None):
    """ Buttons is an array of button labels in order Continue, Discard, Save.
    Save is optional.
    Runs the dialog and returns response. """
    if not buttons:
        buttons = (None, None, None)
    if not title:
        title = _("Unsaved text")
    if not message:
        message = _("Your edits will be lost if you do not save " \
                "before continuing.")
    dialog = Gtk.MessageDialog(parent_win, _F,
                    Gtk.MessageType.WARNING, Gtk.ButtonsType.NONE,
                    title)
    dialog.format_secondary_text(message)
    if buttons[0]:
        b = misc.custom_stock_button(buttons[0], Gtk.STOCK_CANCEL)
    else:
        b = Gtk.Button(None, Gtk.STOCK_CANCEL)
    dialog.add_action_widget(b, DCS_CANCEL)
    if buttons[1]:
        b = misc.custom_stock_button(buttons[1], Gtk.STOCK_DELETE)
    else:
        b = Gtk.Button(None, Gtk.STOCK_DELETE)
    dialog.add_action_widget(b, DCS_DISCARD)
    if len(buttons) > 2:
        if buttons[2]:
            b = misc.custom_stock_button(buttons[2], Gtk.STOCK_SAVE)
        else:
            b = Gtk.Button(None, Gtk.STOCK_SAVE)
        dialog.add_action_widget(b, DCS_SAVE)
        # dialog.set_default_response(DCS_SAVE)
    dialog.show_all()
    response = dialog.run()
    dialog.destroy()
    return response


def dcs_one_file(buf, parent_win = None,
        title = None, message = None):
    """ Returns True if current file should be kept. """
    if not title:
        title = _("Unsaved file")
    if not message:
        if buf.filename:
            message = _("File '%s' has been altered.") % buf.filename
        else:
            message = _("This file has been altered.")
        message +=  _(" If you continue " \
                "without saving it your edits will be lost.")
    response = _dcs_dialog(parent_win,
            (None, _("_Discard"), _("_Save first")),
            title, message)
    if response == DCS_SAVE:
        if buf.filename:
            buf.save()
        else:
            filename = get_filename_save(None, parent_win, None)
            if filename:
                buf.save_as(filename)
            else:
                response = DCS_CANCEL
    return response == DCS_CANCEL

def dcs_win_multi_files(parent_win = None, title = None, message = None):
    """ Returns True if files should be kept. """
    if not title:
        title = _("Unsaved files")
        message =  _("There are unsaved files in this window. If you " \
                "continue without saving them your edits will be lost.")
    response = _dcs_dialog(parent_win, (None, _("_Discard")), title, message)
    return response == DCS_CANCEL

def dcs_multi_files(parent_win = None, title = None, message = None):
    """ Returns True if files should be kept. """
    if not title:
        title = _("Unsaved files")
        message =  _("There are unsaved files in vixn. If you continue " \
                "without saving them your edits will be lost.")
    response = _dcs_dialog(parent_win, (None, _("_Discard")), title, message)
    return response == DCS_CANCEL


def check_overwrite(parent_win = None, filename = None):
    if filename:
        message = _("File '%s' already exists" % filename)
    else:
        message = _("This file already exists")
    message += _(" and appears not to belong to this buf.")
    message += _(" Are you sure you want to overwrite it?")
    dialog = Gtk.MessageDialog(parent_win, _F,
            Gtk.MessageType.QUESTION, Gtk.ButtonsType.YES_NO, _("File exists"))
    dialog.format_secondary_text(message)
    dialog.set_default_response(Gtk.ResponseType.YES)
    response = dialog.run()
    dialog.destroy()
    return response == Gtk.ResponseType.YES

class EncodingDialog(ADialog):
    def __init__(self, title = None, parent_win = None, filename = None,
            encoding = None, message = None):
        if not title:
            title = _("Text conversion failed")
        ADialog.__init__(self, title, parent_win, _F,
                (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY))
        self.set_default_response(Gtk.ResponseType.APPLY)
        hbox = Gtk.HBox()
        img = Gtk.Image()
        img.set_from_stock(Gtk.STOCK_DIALOG_WARNING, Gtk.IconSize.DIALOG)
        hbox.pack_start(img, False, False, 10)

        vbox = Gtk.VBox()

        if filename:
            filename = "'" + filename + "'"
        else:
            filename = _("this file")
        if not message:
            if not encoding:
                encoding = locale.getdefaultlocale()[1]
            if encoding.lower() == 'utf8' or encoding.lower() == 'utf-8':
                hint = \
                    _("(the most common encoding in the West is ISO-8859-1) ")
            else:
                hint = ""
            message = _("Unable to convert %s to/from UTF-8" \
                    " using the encoding %s. Try again with a different" \
                    " encoding %sand/or replace invalid characters.") % \
                    (filename, encoding, hint)

        label = Gtk.Label(label=message)
        label.set_alignment(0, 0.5)
        label.set_line_wrap(True)
        label.set_size_request(_EXC_DIAG_WIDTH, -1)
        vbox.pack_start(label, True, True, 8)

        sub_hbox = Gtk.HBox(spacing = 8)
        label = Gtk.Label(label=_("Encoding"))
        sub_hbox.pack_start(label, False, False, 0)
        self.encoding = Gtk.Entry()
        self.encoding.set_text(encoding)
        self.encoding.set_activates_default(True)
        sub_hbox.pack_start(self.encoding, True, True, 0)
        vbox.pack_start(sub_hbox, False, False, 8)

        self.replace = Gtk.CheckButton(_("_Replace invalid characters"))
        vbox.pack_start(self.replace, False, False, 8)

        hbox.pack_start(vbox, True, True, 10)
        self.vbox.pack_start(hbox, True, True, 0)

def encoding(parent_win = None, filename = None, encoding = None):
    """ Returns None for Cancel or tuple: (encoding, replace (bool)) """
    dialog = EncodingDialog(None, parent_win, filename, encoding)
    dialog.show_all()
    response = dialog.run()
    if response != Gtk.ResponseType.CANCEL:
        result = (dialog.encoding.get_text(), dialog.replace.get_active())
    else:
        result = None
    dialog.destroy()
    return result

def close_request(parent_win = None, message = None):
    """ Returns True if window should be kept open """
    if not message:
        message = _("Are you sure you want to close that window?")
    dialog = Gtk.MessageDialog(parent_win, _F, Gtk.MessageType.QUESTION,
            Gtk.ButtonsType.NONE, message)
    dialog.set_title(_("Question from Vixn"))
    dialog.add_buttons(_("Keep Open"), Gtk.ResponseType.CANCEL,
            Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
    dialog.set_default_response(Gtk.ResponseType.CLOSE)
    response = dialog.run()
    dialog.destroy()
    return response == Gtk.ResponseType.CANCEL

def show_about(parent = None):
    dialog = Gtk.AboutDialog()
    if parent:
        dialog.set_transient_for(parent)
        dialog.set_destroy_with_parent(True)
    dialog.set_program_name("Vixn")
    dialog.set_version("%s (%s)" % (version.version_string, version.date))
    dialog.set_copyright("\u00a9 2007-2013 Tony Houghton")
    dialog.set_comments(_("A vi-like programmer's editor"))
    dialog.set_license_type(Gtk.License.GPL_3_0)
    dialog.set_website("http://www.vixn.org")
    dialog.set_website_label("www.vixn.org")
    dialog.set_logo(GdkPixbuf.Pixbuf.new_from_file_at_size(
            os.path.join(dirs.vixn_dir, "vixn_logo.svg"), 128, 128))
    dialog.run()
    dialog.destroy()

