""" Directories used by vixn, and some file-related utilities. """

import locale
import os
import sys

from .vixdomain import *

_home = os.environ.get('HOME', '/')

xdg_data_home = os.environ.get('XDG_DATA_HOME',
            os.path.join(_home, '.local', 'share'))

xdg_data_dirs = \
    os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')

xdg_config_home = os.environ.get('XDG_CONFIG_HOME',
            os.path.join(_home, '.config'))

xdg_config_dirs = \
    os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg').split(':')

xdg_cache_home = os.environ.get('XDG_CACHE_HOME',
            os.path.join(_home, '.cache'))

def _filter_existing_dirs(dir):
    return dir and os.path.isdir(dir)

def _bind_paths(base_paths, leafname):
    bound = []
    for d in base_paths:
        dir = os.path.join(d, leafname)
        if os.path.isdir(dir):
            bound.append(dir)
    return bound

vixn_data_dirs = _bind_paths([xdg_data_home] + xdg_data_dirs, VIXN_RESOURCE)
vixn_config_dirs = _bind_paths([xdg_config_home] + xdg_config_dirs,
        VIXN_RESOURCE)

def _init_vixn_dir():
    def _parent(d):
        return os.path.split(d)[0]
    global vixn_dir
    arg0path = os.path.realpath(sys.argv[0])
    if os.path.basename(arg0path) == 'main.py':
        arg0path = os.path.dirname(arg0path)
    vixn_dir = os.path.dirname(arg0path)
    # If installed we want to use $prefix/share instead of $prefix/lib
    d = _parent(vixn_dir)
    if os.path.basename(d) == 'lib':
        d = os.path.join(_parent(d), 'share', 'vixn')
        if os.path.isdir(d):
            vixn_dir = d

_init_vixn_dir()


def save_path(basedir, *resource):
    """Ensures parent directory of <basedir>/VIXN_RESOURCE/<resource>/ exists,
    and return its path. resource may be a single string, sequence of strings,
    or a tuple/list of strings, or None."""
    while resource and not hasattr(resource[0], 'isspace'):
        resource = resource[0]
    path = os.path.join(basedir, VIXN_RESOURCE, *resource)
    dir = os.path.dirname(path)
    if not os.path.isdir(dir):
        os.makedirs(dir, 0o755)
    return path

def save_config_path(*resource):
    """Ensures dirname($XDG_CONFIG_HOME/VIXN_RESOURCE/<resource>/) exists,
    and returns its path."""
    return save_path(xdg_config_home, resource)

def save_data_path(*resource):
    """Ensures dirname($XDG_DATA_HOME/VIXN_RESOURCE/<resource>/) exists,
    and returns its path."""
    return save_path(xdg_data_home, resource)

def save_cache_path(*resource):
    """Ensures dirname($XDG_CACHE_HOME/VIXN_RESOURCE/<resource>/) exists,
    and returns its path."""
    return save_path(xdg_cache_home, resource)

def get_all_existing(dirs, install_leaf, *resource):
    """Gets all existing instances of resource in <dirs>
    or in <vixn_dir>/<install_leaf>."""
    while not hasattr(resource[0], 'isspace'):
        resource = resource[0]
    resource = os.path.join(*resource)
    dirs = [os.path.join(d, resource) for d in dirs]
    dirs.append(os.path.join(vixn_dir, install_leaf, resource))
    return [x for x in dirs if os.path.exists(x)]

def get_first_existing(dirs, install_leaf, *resource):
    """Gets first existing instance of resource in dirs
    or in <vixn_dir>/<install_leaf>."""
    while not hasattr(resource[0], 'isspace'):
        resource = resource[0]
    resource = os.path.join(*resource)
    for d in dirs:
        res = os.path.join(d, resource)
        if os.path.exists(res):
            return res
    res = os.path.join(vixn_dir, install_leaf, resource)
    if os.path.exists(res):
        return res
    return None

def get_all_config(*resource):
    """Gets all instances of resource in vixn_config_dirs
    or in <vixn_dir>/config"""
    return get_all_existing(vixn_config_dirs, 'config', resource)

def get_first_config(*resource):
    """Gets first instance of resource in vixn_config_dirs
    or in <vixn_dir>/config"""
    return get_first_existing(vixn_config_dirs, 'config', resource)

def get_all_data(*resource):
    """Gets all instances of resource in vixn_data_dirs
    or in <vixn_dir>/data"""
    return get_all_existing(vixn_data_dirs, 'data', resource)

def get_first_data(*resource):
    """Gets first instance of resource in vixn_data_dirs
    or in <vixn_dir>/data"""
    return get_first_existing(vixn_data_dirs, 'data', resource)

def get_file_for_lang(prefix, lang, suffix = None):
    """ Searches for the most appropriate file for the given
    language. Note suffix must include separator. For example:
    get_file_for_lang('/some/path/index', 'en_GB', '.html')
    would look for /some/path/index.en_GB.html, then /some/path/index.en.html,
    then /some/path/index.html. """
    if not suffix:
        suffix = ''
    path = prefix + '.' + lang + suffix
    if os.path.exists(path):
        return path
    if '_' in lang:
        lang = lang.split('_', 1)[0]
        path = prefix + '.' + lang + suffix
        if os.path.exists(path):
            return path
    path = prefix + suffix
    if os.path.exists(path):
        return path
    return None

def get_file_for_current_lang(prefix, suffix = None):
    """ Calls get_file_for_lang() with the current language as returned by
    locale.getdefaultlocale(). """
    return get_file_for_lang(prefix, locale.getdefaultlocale()[0], suffix)

def get_vixn_share_dir(*resource):
    """ Looks for a data file in vixn's system share dir, either where it's
    run from, or /usr/share for packages. """
    if resource:
        return os.path.join(vixn_dir, *resource)
    else:
        return vixn_dir

def get_vixn_logo_icon():
    return get_vixn_share_dir("vixn_logo.svg")

