import os

from gi.repository import Gdk
from gi.overrides import keysyms as ks

from . import buffers
from . import dialog
from . import dirs
from . import hooks
from . import rc
from . import search
from . import settings
from . import vixiter
from . import vixview
from . import vixwin

def trailing_backslash(s):
    n = 0
    i = len(s) - 1
    while i >= 0 and s[i] == '\\':
        n += 1
        i -= 1
    return (n % 2) == 1

def rejoin_escaped_delimiters(args, sep):
    n = 0
    while n < len(args) - 1:
        while trailing_backslash(args[n]):
            if len(args) > n + 1:
                args[n] = args[n][:-1] + sep + args[n + 1]
                del args[n + 1]
        n += 1

def get_flag(flags, flag):
    result = flag in flags
    if result:
        n = flags.index(flag)
        flags = flags[:n] + flags[n + 1:]
    return [result, flags]

#################################################

class ExCommand(object):
    def __init__(self, method, help = None):
        self.method = method
        self.help = help



commands = {}


def command(help = None):
    """ Use as a decorator to register a function as an ex command. Each
    function is called with an ExInterpreter as its first argument. """
    def decorator(method):
        commands[method.__name__] = ExCommand(method, help)
        return method
    return decorator


""" Built-in ex commands. """

def _check_filename_for_loading(interp, filename):
    if filename == '*':
        default_dir = os.path.dirname(interp.buffer.get_default_filename())
        if not default_dir.endswith(os.path.sep):
            default_dir += os.path.sep
        filename = dialog.get_filename_open(None, interp.window, default_dir)
    return filename

@command(_("Load named file in a new tab. If no filename argument is " \
        "given, an empty buf is created instead. If the argument is '*' " \
        "a file dialog is used."))
def e(interp, filename = None):
    if filename:
        filename = _check_filename_for_loading(interp, filename)
        if not filename:
            return
    if filename and interp.buffer and interp.buffer.pristine:
        interp.buffer.load(filename, True)
    else:
        s = interp.window
        if not s:
            s = interp.get_settings()
        interp.open_file(s, None, filename)

@command(_("Insert named file (if the filename '*' is given, a file dialog " \
        "is used) into current buf. If no filename is given, a new window " \
        "and view is opened for the current buf."))
def r(interp, filename = None):
    if filename:
        filename = _check_filename_for_loading(interp, filename)
        if filename:
            interp.buffer.insert_file(filename)
    else:
        win = vixwin.VixnWindow(interp.buffer, settings.PRIORITY_DEFAULT)
        win.present()
        win.view.update_title()

@command(_("Save buf contents to a file"))
def w(interp, filename = None):
    if filename == '*' or (not filename and \
            (not interp.buffer.filename or interp.bounds)):
        interp.save_as()
    elif filename:
        if interp.force:
            force = 0
        else:
            force = 2
        if interp.bounds:
            interp.buffer.save_range(filename,
                    interp.bounds[0], interp.bounds[1], force)
        else:
            interp.buffer.save_as(filename)
    else:
        # First condition prevents this branch being reached if there's
        # a bounds
        interp.buffer.save(False)

@command(_("Close current tab, discarding buf if it has no other views"))
def q(interp, arg = None):
    if interp.force or not interp.window.close_request():
        interp.window.destroy_view(interp.view)

@command(_("Save current buffer, then close it"))
def wq(interp, arg = None):
    w(interp, arg)
    q(interp)

@command(_("Save current buffer, then close it"))
def x(interp, arg = None):
    w(interp, arg)
    q(interp)

@command(_("Close all windows, discarding buffers"))
def qa(interp, arg = None):
    if not interp.force and buffers.any_modified():
        if dialog.dcs_multi_files(interp.window):
            return
    vixwin.quit_all()

@command(_("Edit user rc file (in  a new window)"))
def econf(interp, arg = None):
    f = dirs.save_config_path('rc')
    interp.open_file(interp.get_settings(), os.path.dirname(f), f)

@command(_("Run a python statement in current window's context"))
def py(interp, arg = None):
    if arg:
        rc.context.settings = interp.get_settings()
        rc.context.eval(arg)

@command(_("Apply setting to curernt buf, affecting all its views"))
def set(interp, arg = None):
    interp.set(arg, False)

@command(_("Apply setting to curernt buf, affecting only current view"))
def setlocal(interp, arg = None):
    interp.set(arg, True)

@command(_("Show value of a setting"))
def show(interp, arg = None):
    if not arg:
        return
    desc = settings.registry.lookup(arg)
    if desc:
        interp.show_message('%s=%s' % \
                (desc.get_key(),
                    str(interp.get_settings().get_setting(arg))))
    else:
        interp.show_message(_("No such setting '%s'") % arg)

@command(_("Search and replace"))
def s(interp, arg = None):
    if not arg:
        return
    sep = arg[0]
    if sep == '\\':
        interp.show_message(_("\\ is not a valid substitution delimiter"))
        return
    args = arg[1:].split(sep)
    rejoin_escaped_delimiters(args, sep)
    if len(args) < 2:
        interp.show_message( \
                _("Missing replacement string for substitution"))
        return
    if len(args) > 3:
        interp.show_message( \
                _("Invalid trailing characters in substitution"))
        return
    pattern = args[0]
    replace = args[1]
    if len(args) > 2:
        flags = args[2]
    else:
        flags = ""
    orig_flags = flags
    confirm, flags = get_flag(flags, 'c')
    glob, flags = get_flag(flags, 'g')
    ic, flags = get_flag(flags, 'i')
    noic, flags = get_flag(flags, 'I')
    if flags:
        interp.show_message( \
                _("Invalid flags or trailing characters in substitution"))
        return
    if ic and noic:
        interp.show_message( \
                _("I and i substitution flags given simultaneously"))
        return
    if interp.bounds:
        start = interp.bounds[0]
        end = interp.bounds[1]
    else:
        start = interp.buffer.get_iter_at_insert()
        end = start.copy()
        start.set_line_offset(0)
        vixiter.iter_to_eol(end)
    if not ic and not noic:
        ic = search.get_ignore_case( \
                interp.get_settings(), pattern)
    sub = search.Substitute(pattern, replace, interp.buffer, start, end,
            ic, glob)
    if confirm:
        bounds = sub.get_next_match()
        if bounds:
            interp.highlight_match(bounds)
            interp.window.status_bar.open_query(interp.view,
                    _("<b>Replace (y/n/q/a/l/^E/^Y)?</b>"),
                    interp.search_query_cb, None, sub)
    else:
        interp.subst_all(sub)

@command(_("Run make in current buf's working directory"))
def make(interp, arg = None):
    """ A new VixnBuffer may have to be made, so get our buf to do it to
    avoid circular dependencies. """
    interp.buffer.launch_make(arg)

@command(_("Run scons in current buf's working directory"))
def scons(interp, arg = None):
    """ A new VixnBuffer may have to be made, so get our buf to do it to
    avoid circular dependencies. """
    interp.buffer.launch_scons(arg)

@command(_("Trim trailing spaces from lines"))
def trim(interp, arg = None):
    interp.buffer.trim()

@command(_("Show information about Vixn"))
def about(interp, arg = None):
    dialog.show_about(interp.window)

""" End built in ex commands. """


class ExInterpreter(object):
    """ Interprets ex-style commands and runs appropriate method on
    buf, view or window. """

    def __init__(self, window, priority = settings.PRIORITY_USER):
        """ window may be a VixnBuffer instead of a window so that settings
        can be loaded from files and hooks run on buffers before they're
        attached to windows. """
        self.buffer = None
        self.view = None
        self.window = None
        if hasattr(window, "is_buffer"):
            self.buffer = window
        elif hasattr(window, "is_window"):
            self.window = window
            self.view = window.view
            self.buffer = self.view.get_buffer()
        self.priority = priority
        # Workaround for circular dependency
        from .openfile import open_file
        self.open_file = open_file

    def set_window(self, window):
        self.window = window
        self.set_view(window.view)

    def set_view(self, view):
        self.view = view
        self.buffer = view.get_buffer()

    def get_settings(self):
        return self.buffer

    def show_status(self, message, immediate = False):
        if self.window:
            self.window.status_bar.show_message(message, immediate)
            return True
        else:
            return False

    def show_message(self, message):
        if not self.show_status(message):
            dialog.message_dialog(message = message)

    def __sep_line_num(self, command):
        """ Returns [iter for line number, rest of command] """
        # This method is very inefficient performance-wise, but nice and simple,
        # and performance doesn't really matter here
        num = ""
        while command[0].isdigit():
            num += command[0]
            command = command[1:]
        try:
            iter = self.buffer.get_iter_at_line(int(num))
        except:
            self.show_message(_("Invalid line number %s") % num)
            iter = self.buffer.get_iter_at_insert()
        return [iter, command]

    def __sep_place(self, command, end):
        """ Returns [iter, rest of command]. If command doesn't begin with a
        'place' iter is None and command is unchanged. end determines whether
        whole-line places should be at start or end of line. """
        buf = self.buffer
        iter = None
        if command[0].isdigit():
            iter, command = self.__sep_line_num(command)
        elif command[0] == '.':
            # Special case; don't strip if looking for start,
            # because it applies to end too
            if end:
                command = command[1:]
            iter = buf.get_iter_at_insert()
        if iter:
            if end:
                vixiter.iter_to_eol(iter)
            else:
                iter.set_line_offset(0)
            return [iter, command]
        if command[0] == '%':
            # Special case; don't strip if looking for start,
            # because it applies to end too
            if end:
                return [buf.get_end_iter(), command[1:]]
            else:
                return [buf.get_start_iter(), command]
        mark = None
        if command[0] == "<":
            # Allow < or > on own as shortcuts for '< and '>
            mark = buf.get_vi_mark('<')
        elif command[0] == ">":
            mark = buf.get_vi_mark('>')
        elif command[0] == "$":
            iter = buf.get_start_iter()
        if mark:
            iter = buf.get_iter_at_mark(mark)
        if iter:
            return [iter, command[1:]]
        if command[0] == "'":
            if len(command) == 1:
                return [None, ""]
            else:
                mark = buf.get_vi_mark(command[1])
                if mark:
                    iter = buf.get_iter_at_mark(mark)
            command = command[2:]
        return [iter, command]

    def __get_bounds(self, command):
        """ Sets bounds (pair of iters) if appropriate and returns modified
        version of command with bounds stripped. """
        start, command = self.__sep_place(command, False)
        if start and (command[0] == ',' or command[0] == '%'):
            if command[0] == ',':
                command = command[1:]
            end, command = self.__sep_place(command, True)
        else:
            end = None
        if start and not end:
            end = start.copy()
            vixiter.iter_to_eol(end)
        if start:
            bounds = [start, end]
        else:
            bounds = None
        return [bounds, command]

    def execute(self, command, num = 1):
        """ Executes the complete command on the appropriate target. num is
        number arg from vicmd, currently ignored. """
        if command[0] == ':':
            command = command[1:]
        elif command[0] == '/' or command[0] == '?' or command[0] == '\\':
            self.search(command, num)
            return
        command = command.lstrip()
        bounds, command = self.__get_bounds(command)
        command = command.lstrip()
        if not command:
            if bounds:
                self.view.move_cursor_to_iter(bounds[0], True)
            return
        for n in range(len(command)):
            if not command[n].isalpha():
                method = command[:n]
                if command[n] == '!':
                    force = True
                    args = command[n + 1:]
                else:
                    force = False
                    args = command[n:]
                args = args.lstrip()
                break
        else:
            method = command
            args = None
            force = False
        if method:
            self.execute_method(method, bounds, force, args)

    def execute_method(self, method_name, bounds, force = False, args = None):
        if self.view:
            self.view.active_for_buffer(True)
        cmd = commands.get(method_name)
        if cmd:
            self.bounds = bounds
            self.force = force
            cmd.method(self, args)
        else:
            self.show_message(_("Unrecognised ex command ':%s'") % method_name)

    def search(self, pattern, num = 1):
        if self.view:
            self.view.active_for_buffer(True)
        if pattern[0] == '?':
            backward = True
        else:
            backward = False
        if pattern[0] == '\\':
            tback = True
        else:
            tback = False
        pattern = pattern[1:]
        if not pattern:
            return
        if self.window:
            self.show_status(_("Searching..."), True)
        result = self.buffer.start_search(pattern, backward, tback, num)
        if self.window:
            self.window.status_bar.restore_status(True)
        if not tback:
            self.view.show_search_result(result)

    def run_hooks(self, hook, match_with):
        hks = hook.get_matching_hooks(match_with)
        if not hks:
            return
        rc.context.settings = self.get_settings()
        rc.context.push_priority(hook.get_settings_priority())
        for h in hks:
            if hasattr(h, 'isspace'):
                self.execute(h)
            else:
                rc.context.run_function(h)
        rc.context.pop_priority()

    def save_as(self, args = None):
        filename = dialog.get_filename_save(None, self.window,
                self.buffer.filename)
        if filename:
            filename = self.get_settings().expandpath(filename)
            b = buffers.find_filename(filename)
            if b and b != self.buffer:
                dialog.error(self.window,
                _("'%s' is already being edited in another buf"),
                filename)
                return
            if self.force:
                force = 0
            else:
                force = 2
            if self.bounds:
                self.buffer.save_range(filename,
                        self.bounds[0], self.bounds[1], force)
            else:
                self.buffer.save_as(filename)

    def set(self, arg, local):
        if not arg:
            return
        if '=' in arg:
            k, v = arg.split('=', 1)
            k = k.strip()
            v = v.strip()
            self.get_settings().set_setting(k, v, self.priority, local)
        elif arg.endswith('?'):
            show(self, arg[:-1])
        else:
            if arg.startswith('no'):
                v = False
                k = arg[2:]
            else:
                v = True
                k = arg
            self.get_settings().set_setting(k, v, self.priority, local)

    def end_search(self):
        self.window.status_bar.restore_status(True)
        self.buffer.clear_selection()

    def highlight_match(self, bounds):
        self.buffer.select_range(*bounds)
        self.view.scroll_cursor_on_screen()

    def subst_all(self, sub):
        self.show_status(_("Substituting..."), True)
        n = 0
        iua = self.buffer.ignore_user_action
        self.buffer.ignore_user_action = True
        self.buffer.begin_user_action()
        while True:
            if not sub.get_next_match():
                break
            sub.replace()
            n += 1
        self.buffer.clear_selection()
        self.buffer.end_user_action()
        self.buffer.ignore_user_action = iua
        self.show_message(_("%d substitution(s) made") % n)

    def search_query_cb(self, state, keysym, keystr, sub):
        replace = False
        get_next = False
        cancel = False
        if (keysym == ks.Y or keysym == ks.y) and \
                (state & Gdk.ModifierType.CONTROL_MASK):
            self.view.scroll_by_line(-1)
        elif (keysym == ks.E or keysym == ks.e) and \
                (state & Gdk.ModifierType.CONTROL_MASK):
            self.view.scroll_by_line(1)
        elif keystr == 'y' or keystr == 'Y':
            replace = True
            get_next = True
        elif keystr == 'n' or keystr == 'N':
            get_next = True
        elif keystr == 'q' or keystr == 'Q':
            cancel = True
        elif keystr == 'l' or keystr == 'L':
            replace = True
            cancel = True
        elif keystr == 'a' or keystr == 'A':
            iua = self.buffer.ignore_user_action
            self.buffer.ignore_user_action = True
            self.buffer.begin_user_action()
            sub.replace()
            self.subst_all(sub)
            self.end_search()
            self.buffer.end_user_action()
            self.buffer.ignore_user_action = iua
            return
        if replace:
            sub.replace()
        if get_next:
            bounds = sub.get_next_match()
            if bounds:
                self.highlight_match(bounds)
            else:
                cancel = True
        if cancel:
            self.end_search()


