""" Formats (reflows) text, using special rules according to language (in
gtksourceview sense) where appropriate, eg taking quoting into account for
email. """

import re


def _split(text):
    """ This returns a list of pairs: [text, separator space sequence].
    Any separator space sequence containing a \n is replaced by a single
    space, unless it's at the end, when it's replaced by a single \n. """
    pairs = []
    n = 0
    l = len(text)
    nl = 0
    while n < l:
        t = ""
        s = ""
        nl = 0
        while n < l and not text[n].isspace():
            t += text[n]
            n += 1
        while n < l and text[n].isspace():
            c = text[n]
            if c == '\n':
                nl += 1
            s += c
            n += 1
        if nl == 1:
            s = ' '
        elif nl > 1:
            s = '\n' * nl
        pairs.append([t, s])
    # If there were trailing \n's, replace with a single \n for each one
    # instead of a space for all of them
    if nl:
        pairs[-1][1] = '\n' * nl
    return pairs


def len_with_tabs(text, tabstop = 8, start_col = 0):
    l = 0
    for c in text:
        if c == '\t':
            l += tabstop - (start_col + l) % tabstop
        else:
            l += 1
    return l


def basic_format(text, width, tabstop = 8, prefix = None, start_col = 0):
    pairs = _split(text)
    line_len = start_col
    n_pairs = len(pairs)
    text = ""
    line1 = True
    if prefix:
        prefix_len = len_with_tabs(prefix, tabstop, start_col)
    for n in range(n_pairs):
        if not line1 and prefix and line_len == start_col:
            text += prefix
            line_len = prefix_len + start_col
        line1 = False
        text += pairs[n][0]
        line_len += len_with_tabs(pairs[n][0], tabstop, line_len)
        s = pairs[n][1]
        space_len = len_with_tabs(s, tabstop, line_len)
        extra_len = space_len
        if n < n_pairs - 1:
            extra_len += len_with_tabs(pairs[n + 1][0],
                    tabstop, line_len + extra_len)
        if line_len + extra_len > width:
            text += '\n'
            line_len = start_col
        else:
            text += s
            if '\n' in s:
                line_len = start_col
                line1 = True
            else:
                line_len += space_len
    return text


_space_re = None

def _get_space_re():
    global _space_re
    if not _space_re:
        _space_re = re.compile(r"(\s+)")
    return _space_re


def advanced_format(text, width, rex, tabstop, start_col = 0):
    """ Splits text into lines and then paragraphs so it can detect paragraphs
    starting with bullet points. """
    lines = text.split('\n')
    n_lines = len(lines)
    paras = []
    # para = [prefix for all except first line, lines]
    para = [None, []]
    old_prefix = ""
    for n in range(n_lines):
        l = lines[n]
        # If not on last line, l2 is next line (l2 = None if next line is blank)
        l2 = None
        if n < n_lines - 1:
            l2 = lines[n + 1]
            if l2.isspace():
                l2 = None
        prev_para = None
        prefix = ""
        for c in l:
            if c.isspace():
                prefix += c
            else:
                break
        if not l or l.isspace():
            # Start new para if current line is blank
            if para[1]:
                prev_para = para
            para = [None, []]
            l = '\n'
        elif rex.match(l) and (l2 or n == 0):
            # Line begins with a bullet
            if l2:
                # Another line follows, bullet only counts if the next line
                # begins with space; use the space at the start of that next
                # line as prefix
                m =  _get_space_re().match(l2)
                if m:
                    prefix = m.group()
            else:
                # Line is on its own; prefix is start of this line with a space
                # in place of the bullet
                prefix = rex.sub(r"\1 \3", l)
        if prefix and prefix != old_prefix:
            # Also start a new para if we've found a prefix for a bullet
            if para[1]:
                prev_para = para
            para = [prefix, [l]]
        elif l:
            # Otherwise add current line to current para
            para[1].append(l)
        if prev_para:
            # Save last para
            paras.append(prev_para)
        old_prefix = prefix
    if para:
        # If we exit loop with a current para, save it
        paras.append(para)
    text_out = ""
    for p in paras:
        text = '\n'.join(p[1])
        text_out += basic_format(text, width, tabstop, p[0], start_col)
    return text_out


_quote_prefix_re = None

def get_quote_prefix_re():
    global _quote_prefix_re
    if not _quote_prefix_re:
        _quote_prefix_re = re.compile("[>: |]+")
    return _quote_prefix_re 


def _split_by_quote_prefix(text):
    """ Returns a list of [prefix, block] pairs where block is a block of text
    grouped by lines having identical mail/news quote style prefixes. The
    prefixes are stripped from these lines. """
    qpre = get_quote_prefix_re()
    lines = text.split('\n')
    paras = []
    para = ""
    prefix = None
    for l in lines:
        m = qpre.match(l)
        if m:
            m = m.group()
            l = l[len(m):]
        else:
            m = ""
        if m == prefix:
            para += '\n' + l
        else:
            if para:
                paras.append([m, para])
            para = ""
        prefix = m
    if prefix == None:
        prefix = ""
    if para:
        paras.append([prefix, para])
    return paras


def format_quoted(text, width, rex, tabstop):
    para_pairs = _split_by_quote_prefix(text)
    text_out = ""
    for prefix, text in para_pairs:
        lines = advanced_format(text, width - len(prefix), rex, tabstop)
        lines = lines.split('\n')
        for n in range(len(lines)):
            lines[n] = prefix + lines[n]
        text_out += '\n'.join(lines)
    return text_out


def format(text, width, buf, tabstop):
    """ Main format function """
    ft = buf.get_filetype()
    if ft == 'email':
        formatter = format_quoted
    else:
        formatter = advanced_format
    return formatter(text, width, buf.get_fmt_bullet_re(), tabstop)

