import atexit
import os

from gi.repository import GLib, GObject
from gi.repository import Gdk, Gtk

from . import dirs
from .state import state

MAX_LINES = 32

def string_is_ex_prefix_char(s):
    return len(s) == 1 and s in r"\/:?"

class History(object):
    def __init__(self, lines):
        self.lines = lines

    def append(self, line):
        if not line or string_is_ex_prefix_char(line):
            return
        if self.lines and self.lines[-1] == line:
            return
        try:
            i = self.lines.index(line)
            del self.lines[i]
        except:
            pass
        if len(self.lines) >= MAX_LINES:
            self.lines.pop(0)
        self.lines.append(line)

    def dec_iter(self, index):
        index -= 1
        if index < -1:
            index = len(self.lines) - 1
        return index

    def inc_iter(self, index):
        index += 1
        if index == len(self.lines):
            index = -1
        return index

    def search_backward(self, index, prefix):
        """ If none found, index returned is -1 """
        i = index
        while True:
            i = self.dec_iter(i)
            if i == -1 and index != -1:
                i = self.dec_iter(i)
            if i == index:
                return -1
            if self.lines[i].startswith(prefix):
                return i

    def search_forward(self, index, prefix):
        """ If none found, index returned is -1 """
        i = index
        while True:
            i = self.inc_iter(i)
            if i == -1 and i != index:
                i = self.inc_iter(i)
            if i == index:
                return -1
            if self.lines[i].startswith(prefix):
                return i


commands = History(state.command_history)
searches = History(state.search_history)


class Entry(Gtk.Entry):
    """ A Gtk.Entry with a history buffer. Not GTK's history. This one works
    like bash history with PageUp and PageDown as search history. Any derived
    classes of this should check the ignore_changes semaphore when responding
    to the "changed" signal. prefix is whichever char was used to open the
    widget (: or /). """
    def __init__(self, prefix = "", history = None):
        GObject.GObject.__init__(self)
        self.prefix = prefix
        self.ignore_changes = 0
        self.attach_history(history)
        self.search_prefix = None
        self.connect("key-press-event", self.key_press_cb)
        self.connect("changed", self.changed_cb)
        self.changed_since_last_save = False

    def get_text(self):
        return str(Gtk.Entry.get_text(self))

    def set_prefix(self, prefix):
        self.prefix = prefix

    def attach_history(self, history):
        self.history = history
        self.history_index = -1
        # Using -1 rather than 0 makes sure going forwards through history
        # starts at first item

    def clear_selection(self, pos = -1):
        self.select_region(0, 0)
        self.set_position(pos)

    def __post_set_text_unselected(self, args):
        pos, clipboard, clip_text = args
        self.set_position(pos)
        self.ignore_changes -= 1
        if clip_text:
            clipboard.set_text(clip_text, -1)

    def __pre_set_text_unselected(self, clipboard, clip_text, data):
        self.set_text(data[0])
        GLib.idle_add(self.__post_set_text_unselected,
                (data[1], clipboard, clip_text))

    def set_text_unselected(self, text, pos = -1):
        # Need a bit of trickery to preserve PRIMARY clipboard
        self.ignore_changes += 1
        Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY).request_text( \
                self.__pre_set_text_unselected, (text, pos))

    def save_history(self):
        if not self.changed_since_last_save:
            return
        text = self.get_text()
        if string_is_ex_prefix_char(text[0]):
            text = text[1:]
        if not len(text):
            return
        self.history.append(text)
        self.changed_since_last_save = False
        self.history_index = -1

    def changed_cb(self, widget):
        self.changed_since_last_save = True
        if not self.ignore_changes:
            self.history_index = -1

    def post_history_move(self, pos):
        if self.history_index == -1:
            text = ""
        else:
            text = self.history.lines[self.history_index]
        char1 = self.get_text()
        if char1:
            if text:
                text = char1[0] + text
            else:
                text = char1[0]
        self.set_text_unselected(text, pos)

    def history_backward(self, pos = -1):
        if not self.history.lines:
            return
        self.history_index = self.history.dec_iter(self.history_index)
        self.post_history_move(pos)

    def history_forward(self, pos = -1):
        if not self.history.lines:
            return
        self.history_index = self.history.inc_iter(self.history_index)
        self.post_history_move(pos)

    def __get_history_search_prefix(self):
        text = self.get_text()
        pos = self.get_position()
        if pos > len(text) or pos < 0:
            pos = len(text)
        if string_is_ex_prefix_char(text[0]):
            return text[1:pos]
        else:
            return text[:pos]

    def __search_history(self, search_method, empty_method):
        prefix = self.__get_history_search_prefix()
        l = len(prefix)
        if not l or string_is_ex_prefix_char(prefix):
            empty_method(l)
            return
        i = search_method(self.history_index, prefix)
        if i != -1:
            self.history_index = i
            self.post_history_move(l + 1)

    def search_history_backward(self):
        self.__search_history(self.history.search_backward,
                self.history_backward)

    def search_history_forward(self):
        self.__search_history(self.history.search_forward,
                self.history_forward)

    def key_press_cb(self, widget, event):
        k = event.keyval
        if k == Gdk.KEY_Up or k == Gdk.KEY_KP_Up:
            self.history_backward()
            return True
        elif k == Gdk.KEY_Down or k == Gdk.KEY_KP_Down:
            self.history_forward()
            return True
        elif k == Gdk.KEY_Page_Up or k == Gdk.KEY_KP_Page_Up:
            self.search_history_backward()
            return True
        elif k == Gdk.KEY_Page_Down or k == Gdk.KEY_KP_Page_Down:
            self.search_history_forward()
            return True
        return False



class TabEntry(Entry):
    """ As Entry but with tab filename completion too. """

    class Matcher(object):
        def __init__(self, cwd, matchstr):
            self.cwd = cwd
            self.matchstr = matchstr
            self.matches = None
            self.index = -1

        def is_valid_for(self, matchstr):
            return self.matchstr == matchstr

        def get_matches(self):
            if self.matches != None:
                return
            self.index = -1
            matchstr = self.matchstr
            sep = matchstr.rfind(os.path.sep)
            if sep == 0:
                if os.path.isabs(matchstr):
                    # Assume root is os.path.sep on its own
                    subdir = matchstr[0]
                    dir = matchstr[0]
                else:
                    # Assume relative to cwd
                    subdir = matchstr[0]
                    dir = self.cwd
            elif sep != -1:
                subdir = matchstr[:sep]
                if os.path.isabs(subdir):
                    dir = subdir
                else:
                    dir = os.path.join(self.cwd, subdir)
            else:
                subdir = None
                dir = self.cwd
            matches = os.listdir(dir)
            self.matches = []
            for m in matches:
                if os.path.isdir(os.path.join(dir, m)) and not m.endswith('/'):
                    m += '/'
                if subdir:
                    m = os.path.join(subdir, m)
                if m.startswith(matchstr):
                    self.matches.append(m)

        def get_next_match(self):
            self.get_matches()
            if not self.matches:
                return None
            self.index += 1
            if self.index >= len(self.matches):
                self.index = 0
            return self.matches[self.index]


    def __init__(self, cwd, prefix = "", history = None):
        self.cwd = cwd
        self.matcher = None
        Entry.__init__(self, prefix, history)

    def key_press_cb(self, widget, event):
        if event.keyval == Gdk.KEY_Tab:
            self.tab_completion()
            return True
        else:
            return Entry.key_press_cb(self, widget, event)

    def tab_completion(self):
        contents = self.get_text()
        prefix = contents[0]
        contents = contents[1:]
        l = len(contents)
        pos = self.get_position() - 1
        if pos < 0:
            pos = l
        start = contents.find(' ') + 1
        if start > pos:
            end = start
            start = 0
        else:
            end = contents.find(' ', pos)
            if end == -1:
                end = l
        matchstr = contents[start:pos]
        if matchstr and matchstr[0] == '~':
            l0 = len(matchstr)
            matchstr = os.path.expanduser(matchstr)
            dl = len(matchstr) - l0
            pos += dl
            l += dl
        if self.matcher and not self.matcher.is_valid_for(matchstr):
            self.matcher = None
        if not self.matcher:
            self.matcher = self.Matcher(self.cwd, matchstr)
        matchstr = self.matcher.get_next_match()
        if not matchstr:
            return
        if start:
            matchstr = prefix + contents[:start] + matchstr
        else:
            matchstr = prefix + matchstr
        if end < l:
            matchstr += contents[end:]
        self.set_text_unselected(matchstr, pos + 1)

    def reset_matches(self, cwd):
        self.cwd = cwd
        self.matcher = None

