import fnmatch
import os

from . import dialog
from . import settings

_exact_filetype_hooks = {}
_exact_filename_hooks = {}
_exact_scheme_hooks = {}
_glob_filename_hooks = {}
_glob_filetype_hooks = {}
_glob_scheme_hooks = {}


class Hooks(object):
    def __init__(self, priority, uses_leaf = False):
        self.exact_hooks = {}
        self.glob_hooks = {}
        self.uses_leaf = uses_leaf
        self.priority = priority
    
    def choose_hooks(self, match_str):
        """ Returns exact_hooks or glob_hooks depending on whether match_str
        contains glob characters. """
        if '*' in match_str or '?' in match_str or '[' in match_str:
            return self.glob_hooks
        else:
            return self.exact_hooks

    def add_hook(self, match_with, action = None):
        """ If action is None, this acts as a decorator. Otherwise action may
        be a string (treated as an ex command) or a python function. """
        hooks = self.choose_hooks(match_with)

        def add(a):
            matches = hooks.get(match_with)
            if matches:
                matches.append(a)
            else:
                hooks[match_with] = [a]

        def decorator(fn):
            add(fn)
            return fn

        if not action:  # Being used as a decorator
            return decorator
        else:
            add(action)
            return None

    def get_matching_hooks(self, match_with):
        """ Hooks are matched in priority order; only hooks in the first
        matching priority are returned. Priorities are:
          1. Exact (with path)
          2. Leafname
          3. Glob wildcards (with path)
          4. Glob wildcards (leafname)
        """
        matches = self.exact_hooks.get(match_with)
        if matches:
            return matches
        leaf = self.uses_leaf
        if leaf:
            leaf = os.path.basename(match_with)
            if leaf != match_with:
                matches = self.exact_hooks.get(leaf)
            else:
                leaf = False
        if matches:
            return matches
        matches = []
        for k, v in list(self.glob_hooks.items()):
            if fnmatch.fnmatch(match_with, k):
                matches += v
        if not matches and leaf:
            for k, v in list(self.glob_hooks.items()):
                if leaf and fnmatch.fnmatch(leaf, k):
                    matches += v
        return matches

    def clear_hooks(self, match_with, action = None):
        hooks = self.choose_hooks(match_with)
        matches = hooks.get(match_with)
        if not matches:
            dialog.warning(None,
                    _("No hooks matching '%s' to clear") % match_with)
            return
        if action:
            try:
                i = matches.index(action)
            except:
                dialog.warning(None,
                    _("No hooks matching '%s' with action '%s' to clear") % \
                            (match_with, str(action)))
            else:
                del matches[i]
                if not matches:
                    del hooks[match_with]
        else:
            del hooks[match_with]
    
    def get_settings_priority(self):
        return self.priority


filename = Hooks(settings.PRIORITY_FILENAME_HOOK, True)
filetype = Hooks(settings.PRIORITY_FILETYPE_HOOK, False)
scheme = Hooks(settings.PRIORITY_SCHEME_HOOK, False)

