""" Maps GDK key symbols to vim key-notation """

from gi.repository import Gdk
from gi.overrides.keysyms import *

""" All the strings in the values (RHS column) of this dict form the encoded
names of special keys when surrounded with <>. Ctrl and Alt are represented by
C- and/or A- (or M- for Meta) respectively after the < followed by one of the
following strings or a plain key, followed by >. Eg <C-B>, <A-C-Del>. """
gkeys = {
    BackSpace:  'BS',
    Tab:        'Tab',
    Linefeed:   'NL',
    Return:     'Return',
    Escape:     'Esc',
    Delete:     'Del',
    Up:         'Up',
    Down:       'Down',
    Left:       'Left',
    Right:      'Right',
    F1:         'F1',
    F2:         'F2',
    F3:         'F3',
    F4:         'F4',
    F5:         'F5',
    F6:         'F6',
    F7:         'F7',
    F8:         'F8',
    F9:         'F9',
    F10:        'F10',
    F11:        'F11',
    F12:        'F12',
    Insert:     'Insert',
    Home:       'Home',
    End:        'End',
    Page_Up:    'PageUp',
    Page_Down:  'PageDown',
    KP_Insert:  'kInsert',
    KP_Delete:  'kDel',
    KP_Home:    'kHome',
    KP_End:     'kEnd',
    KP_Page_Up: 'kPageUp',
    KP_Page_Down:   'kPageDown',
    KP_0:       'k0',
    KP_1:       'k1',
    KP_2:       'k2',
    KP_3:       'k3',
    KP_4:       'k4',
    KP_5:       'k5',
    KP_6:       'k6',
    KP_7:       'k7',
    KP_8:       'k8',
    KP_9:       'k9',
    KP_Up:      'kUp',
    KP_Down:    'kDown',
    KP_Left:    'kLeft',
    KP_Right:   'kRight',
    KP_Decimal: 'kPoint',
    KP_Enter:   'kEnter',
    KP_Add:     'kPlus',
    KP_Subtract:'kMinus',
    KP_Divide:  'kDivide',
    KP_Multiply:'kMultiply'
}

# Control keys
for n in range(27):
    gkeys[n] = 'C-' + chr(n + 64)


def lookup(gkey):
    global gkeys
    vkey = gkeys.get(gkey)
    if vkey:
        vkey = str('<' + vkey + '>')
    return vkey


def encode_event_key(keyval, ctrl, alt):
    """ Encode a keystroke from event data. """
    k = lookup(keyval)
    if k == None:
        name = Gdk.keyval_name(keyval)
        if "Shift" in name or "Control" in name or \
                "Lock" in name or "Alt" in name or \
                "Meta" in name or "Super" in name or \
                "Hyper" in name:
            return None
        k = chr(keyval)
        if ctrl or alt:
            # Convert Ctrl or Alt to upper-case
            if k.islower():
                k = k.upper()
        if ctrl and alt:
            k = '<A-C-' + k + '>'
        elif ctrl:
            k = '<C-' + k + '>'
        elif alt:
            k = '<A-' + k + '>'
        elif k == '\\' or k == '<':
            k = '\\' + k
    else:
        # Currently Ctrl is only modifier with special mappings
        # in keysyms so don't need to check whether A- is already
        # in k
        if ctrl and alt:
            if k.startswith('<C-'):
                k = '<A-' + k[1:]
            else:
                k = '<A-C-' + k[1:]
        elif ctrl and not k.startswith('<C-'):
            k = '<C-' + k[1:]
        elif alt:
            k = '<A-' + k[1:]
    return k



def includes_ctrl_alt(k):
    """ Whether encoded string k (without <>) has leading C- and/or M-/A-.
    Returns array pair: [Ctrl, Alt] """
    enc_ctrl = False
    enc_alt = False
    l = len(k)
    if l > 3:
        if k[1] == '-':
            if k[0] == 'C':
                enc_ctrl = True
                if l > 4 and (k[2] == 'M' or k[2] == 'A') and k[3] == '-':
                    enc_alt = True
            elif k[0] == 'A' or k[0] == 'M':
                enc_alt = True
                if l > 4 and k[2] == 'C' and k[3] == '-':
                    enc_ctrl = True
    return [enc_ctrl, enc_alt]

def split_ctrl_alt(k):
    """ Combination of includes_ctrl_alt() and strip_ctrl_alt. Returns [Ctrl,
    Alt, stripped k] """
    c, a = includes_ctrl_alt(k)
    if c and a:
        k = k[4:]
    elif c or a:
        k = k[2:]
    return [c, a, k]

def strip_ctrl_alt(k):
    """ Strips any leading C- and/or M-/A- from encoded string.
    Input k and return value don't include <>. """
    return split_ctrl_alt(k)[2]

def split_command(command):
    # Command is a string of encoded keystrokes which may include sequences etc,
    # eg <C-A> or <Esc>
    seq = []
    while len(command) > 0:
        if command[0] == '\\':
            if len(command) >= 2:
                seq.append(command[1])
                command = command[2:]
            else:
                seq.append('\\')
                command = ""
        elif command[0] == '<':
            k = ""
            l = ""
            c, a = includes_ctrl_alt(command[1:])
            if c:
                k += 'C-'
            if a:
                k += 'A-'
            if c and a:
                n = 5
            elif c or a:
                n = 3
            else:
                n = 1
            if (c or a) and command[n] == '>':
                n += 1
                k += '>'
            while len(command) > n and command[n] != '>':
                k += command[n]
                l += command[n]
                n += 1
            brktd = False
            if n < len(command):
                if ((c or a) and len(l) == 1) or l in list(gkeys.values()):
                    brktd = True
                    seq.append('<' + k + '>')
                    command = command[n + 1:]
            if not brktd:
                seq.append('<')
                command = command[1:]
        else:
            seq.append(command[0])
            command = command[1:]
    return seq

def serialise(keys):
    """ keys is a list of keystrokes as returned by split_command. This
    function turns it back into a single string with consistent use of
    backslashes. """
    result = ""
    if not keys:
        return result
    for k in keys:
        if k == '\\' or k == '<':
            result += '\\' + k
        else:
            result += k
    return result

