import os
import re

from . import dialog
from . import settings
from . import tbstb
from . import throwback
from . import vixexcept
from . import vixi18n


class MakeError(vixexcept.UserError):
    pass


class BuildToBuffer(tbstb.TermToBuffer):
    def __init__(self, buf, dir, cmd_opt = 'makeprg', default_cmd = 'make',
            cmd_args = None, stop_callback = None):
        cmd = buf.lookup_setting(cmd_opt, default_cmd)
        if cmd_args:
            cmd += ' ' + cmd_args
        tbstb.TermToBuffer.__init__(self, buf, stop_callback, cmd, True,
                dir, True)

    def add_stdout_line(self, line):
        self.tb_extn.add_line(line, self.tb_extn.check_stdout_message)

    def add_stderr_line(self, line):
        self.tb_extn.add_line(line, self.tb_extn.check_c_message)


class BufExtn(throwback.BufExtn):
    def __init__(self, buf):
        throwback.BufExtn.__init__(self, buf, 1)
        self.makedir = buf.get_cwd()
        self.ed_re = None
        self.c_re = None
        self.inc_re = None
        self.j_re = None

    def parse_line(self, line, *args, **kwargs):
        return args[0](line)

    def process_match(self, m, i_f, i_l, i_c, i_t, i_m):
        """ i_* variables represent match indexes for:
        f: filename
        l: line number
        c: column
        t: type
        m: message
        """
        emsg = m.group(i_m)
        f = m.group(i_f)
        line = int(m.group(i_l))
        line_start_offset = m.start(i_l)
        line_end_offset = m.end(i_l)
        if i_c > 0:
            col = m.group(i_c)
        else:
            col = 0
        if col and col.isdigit():
            col = int(col)
            col_start_offset = m.start(i_c)
            col_end_offset = m.end(i_c)
        else:
            col = 0
            col_start_offset = None
            col_end_offset = None
        if not os.path.isabs(f):
            f = os.path.join(self.makedir, f)
        if line < 1:
            line = 1
        if col < 1:
            col = 1
        if i_t <= 0 or m.group(i_t) == "note:":
            cat = "info"
        elif m.group(i_t) == "warning:":
            cat = "warning"
        else:
            cat = "error"
        return [f, line, col, line_start_offset, line_end_offset,
                col_start_offset, col_end_offset, cat]

    def check_c_message(self, s):
        if not self.c_re:
            self.c_re = re.compile( \
                    r"(?:\s*\[[a-z]*\]\s*)?" \
                    r"(.*?):" \
                    r"([0-9]+):" \
                    r"(?:([0-9]+):)? " \
                    r"(fatal error:|error:|warning:|note:|" \
                        r"undefined reference to) " \
                    r"(.*)",
                    re.I)
        m = self.c_re.match(s)
        if m:
            return self.process_match(m, 1, 2, 3, 4, 5)
        else:
            if not self.inc_re:
                self.inc_re = re.compile( \
                        r"(?:(?:In file included |\s*)from\s)?" \
                        r"(.*?):" \
                        r"([0-9]+)[:,]" \
                        r"(?:([0-9]+)[:,])? ?" \
                        r"(.*)?",
                        re.I)
            m = self.inc_re.match(s)
            if m:
                return self.process_match(m, 1, 2, 3, -1, 4)
            else:
                return None

    def check_j_message(self, s):
        if not self.j_re:
            self.j_re = re.compile( \
                    r"^\s*\[javac\] " \
                    r"(/[^:]+):" \
                    r"([0-9]+):\s*" \
                    r"(error:|warning:)\s*" \
                    r"(.*)",
                    re.I)
        m = self.j_re.match(s)
        if m:
            return self.process_match(m, 1, 2, -1, 3, 4)
        else:
            return None

    def check_enter_leave_dir(self, s):
        return None


    def check_stdout_message(self, s):
        result = self.check_enter_leave_dir(s)
        if not result:
            result = self.check_j_message(s)
        if not result:
            result = self.check_c_message(s)
        return result


class MakeBufExtn(BufExtn):
    def check_enter_leave_dir(self, s):
        if not self.ed_re:
            self.ed_re = re.compile( \
                    r"make(\[[0-9]+\])?: Entering directory .(.*)")
            # NB matched filename will include trailing quote
        m = self.ed_re.match(s)
        if m:
            dir = m.group(2)
            # Remember that trailing quote
            self.makedir = dir[:-1]
        return None


class SConsBufExtn(BufExtn):
    pass


class Build(object):
    def __init__(self, buf, buf_extn_class,
            cmd_opt, default_cmd,
            leafname_opt, default_leafname, path_opt,
            stop_callback = None, window = None, arg = None):
        """ buf is a new buf for holding the output of the command;
        window should be the window the command was run in. If self.dir is
        None after construction, no Makefile/SConstruct etc was found. """
        self.stop_callback = stop_callback
        self.buffer = buf
        self.args = arg
        self.dir = None
        self.buf_extn_class = buf_extn_class
        self.cmd_opt = cmd_opt
        self.default_cmd = default_cmd
        cwd = buf.get_cwd()
        leafname = buf.lookup_setting(leafname_opt, default_leafname).lower()
        for p in buf.lookup_setting(path_opt,
                ".:src:build:../build").split(':'):
            dir = os.path.join(cwd, p)
            if os.path.isdir(dir):
                files = os.listdir(dir)
                for f in files:
                    if f.lower() == leafname:
                        self.dir = dir
                        return
        #dialog.warning(window, _("No Makefile found"))
        self.dir = cwd

    def run(self):
        self.buf_extn_class(self.buffer)
        throwback.add_view_extn_to_buf(self.buffer)
        self.to_buf = BuildToBuffer(self.buffer, self.dir,
                self.cmd_opt, self.default_cmd,
                self.args, self.stop_callback)


class Make(Build):
    def __init__(self, buf, stop_callback = None, window = None, arg = None):
        Build.__init__(self, buf, MakeBufExtn,
                'makeprg', 'make',
                'makefile', 'Makefile', 'makepath',
                stop_callback, window, arg)


class SCons(Build):
    def __init__(self, buf, stop_callback = None, window = None, arg = None):
        Build.__init__(self, buf, SConsBufExtn,
                'sconsprg', 'scons',
                'sconstruct', 'SConstruct', 'sconspath',
                stop_callback, window, arg)



settings.register(('makefilesearchpath', 'makepath', 'mkfspth'), str,
        None, None, _("Colon-separated relative (to buf's cwd) paths " \
                "to search for a Makefile"), settings.SCOPE_BUFFER)

settings.register(('makeprg', 'mp', 'makeprog', 'mkprg', 'mkprog'), str,
        None, None, _("Program run by :make command"), settings.SCOPE_BUFFER)

settings.register(('makefile', 'mkfile'), str,
        None, None, _("Leafname of Makefiles (case-insensitive)"),
        settings.SCOPE_BUFFER)

settings.register(('makeshowextra', 'mkshowx'), int,
        None, None,
        _("Ignored legacy option from old throwback system."),
        settings.SCOPE_BUFFER)

settings.register(('sconsfilesearchpath', 'sconspath', 'scnspth'), str,
        None, None, _("Colon-separated relative (to buf's cwd) paths " \
                "to search for a SConstruct file"), settings.SCOPE_BUFFER)

settings.register(('sconsprg', 'sconsprog'), str,
        None, None, _("Program run by :scons command"), settings.SCOPE_BUFFER)

settings.register(('sconsfile', 'sconstruct'), str,
        None, None, _("Leafname of SConstruct files (case-insensitive)"),
        settings.SCOPE_BUFFER)
