from gi.repository import Gtk

from . import shortcuts

menus = {}
items = {}

def menu_path(menu):
    """ Find the full path of a menu (may be passed in as id or
    Menu object (see below)), returned as array of ids eg
    ["/", "File"]. Returns None if unable to construct path. """
    global items, menus
    if not isinstance(menu, Menu):
        menu = menus[menu]
    if menu.id == '/':
        return ["/"]
    elif menu.parent_item:
        item = items[menu.parent_item]
        if item.parent_menu:
            ancestors = menu_path(item.parent_menu)
            if ancestors:
                return ancestors + [menu.parent_item, menu.id]
    return None

def item_path(item):
    """ Find the full path of a menu item (may be passed in as id or
    MenuItem object (see below)), returned as array of ids eg
    ["/", "File", "Save"]. Returns None if unable to construct path. """
    global items
    if not isinstance(item, MenuItem):
        item = items[item]
    if item.parent_menu:
        path = menu_path(item.parent_menu)
        if path:
            return path + [item.id]
    return None


class Menu(object):
    """ A menu's description rather than the actual GTK object. """
    def __init__(self, id = "/", *child_items):
        """ Default id represents top-level: created as menu bar rather than
        menu. Otherwise id must be globally unique. each member of child_items
        is either a MenuItem id or an array of args to create a MenuItem. """
        self.id = id
        global menus, items
        menus[id] = self
        self.parent_item = None
        self.items = []
        if child_items:
            for i in child_items:
                if hasattr(i, 'isspace'):
                    item_id = i
                    item = items[i]
                else:
                    item = MenuItem(*i)
                    item_id = item.id
                self.items.append(item_id)
                item.parent_menu = self.id

    def create(self, target, menus_dict = None, items_dict = None,
            accel_group = None):
        """ See MenuItem for a description of target. This creates all items
        and submenus recursively, so externally it should only be called on
        top-level. dict args, if given, are dictionaries where each created
        widget is stored keyed by its id. """
        was_connected = shortcuts.disconnect_handler()
        if self.id == '/':
            gmenu = Gtk.MenuBar()
        else:
            gmenu = Gtk.Menu()
            if accel_group:
                gmenu.set_accel_group(accel_group)
        global items
        for i in self.items:
            gitem = items[i].create(target, menus_dict, items_dict, accel_group)
            gmenu.append(gitem)
        if menus_dict != None:
            menus_dict[self.id] = gmenu
        if was_connected:
            shortcuts.connect_handler()
        return gmenu


class MenuItem(object):
    """ A menu item's description rather than the actual GTK object. """
    def __init__(self, id, label = None, method_name = None):
        """ If label is None, the item is a separator. id must be globally
        unique. If method_name begins with ':' it's an ex command, otherwise
        it's a vi command. """
        self.id = id
        self.label = label
        if method_name and method_name[0] == ':':
            self.ex = True
            self.method_name = method_name[1:]
        else:
            self.ex = False
            self.method_name = method_name
        global items
        items[id] = self
        self.parent_menu = None
        self.submenu = None
        shortcuts.register_item(id, method_name)

    def attach_submenu(self, submenu):
        """ submenu may either be Menu or id """
        global menus
        if isinstance(submenu, Menu):
            id = submenu.id
        else:
            id = submenu
            submenu = menus[id]
        self.submenu = id
        submenu.parent_item = self.id

    def create(self, target, menus_dict = None, items_dict = None,
            accel_group = None):
        """ target is the object that will receive the command when the menu
        item is activated. It must have these methods:
        activate_vi_method(gobject, method_name).
        activate_ex_method(gobject, method_name).
        See Menu.create() for description of dict args. """
        if self.label:
            gitem = Gtk.MenuItem.new_with_mnemonic(self.label)
            gitem.set_accel_path(shortcuts.MAIN_PATH + self.id)
        else:
            gitem = Gtk.SeparatorMenuItem()
        if items_dict != None:
            items_dict[self.id] = gitem
        if (self.method_name):
            if self.ex:
                gitem.connect("activate", target.activate_ex_method,
                        self.method_name)
            else:
                gitem.connect("activate", target.activate_vi_method,
                        self.method_name)
        if self.submenu:
            global menus
            gmenu = menus[self.submenu].create(target, menus_dict, items_dict,
                    accel_group)
            gitem.set_submenu(gmenu)
        gitem.show_all()
        return gitem

def create_menu_tree(target, accel_group):
    """ Creates all the menus recursively from the top-level and returns
    a tuple: (menu bar widget, menus_dict, items_dict). See Menu.create()
    for description of dicts. """
    menus_dict = {}
    items_dict = {}
    return (menus['/'].create(target, menus_dict, items_dict, accel_group),
            menus_dict, items_dict)

