""" Executes rc files, parses commands etc, and provides contexts for these
operations. """

import os.path
import sys

from gi.repository import Gtk
from gi.repository import GtkSource

from . import bindings
from . import dialog
from . import dirs
from . import hooks
from . import langs
from . import menus
from . import msgs
from . import settings
from . import shortcuts
from . import vicmd
from . import vixexcept

public_functions = [ \
    ('_', ("s", ), _("Mark string s as translatable")),
    ('set', ("s", "v"),
        _("Set setting s to value v, applying to all views of current buffer")),
    ('setlocal', ("s", "v"),
        _("Set setting s to value v, applying only to current view")),
    ('get_setting', ("s", ), _("Retrieve the value of setting s")),
    ('bind', ("ks", "cmd"),
        _("Bind vi keystroke or sequence ks to command cmd")),
    ('def_menu', ('id = "/"', "*items"),
        _("Define menu with given id (default is top-level) " \
            "and a list of items (objects returned by def_menu_item)")),
    ('def_menu_item', ("id", "label = None", "method_name = None"),
        _("Define a menu item with given id. If label is None the item is " \
            "a separator. method_name is the name of a vi command or " \
            "an ex command if it begins with :")),
    ('attach_submenu', ("item", "submenu"),
        _("Attach a submenu to a menu item")),
    ('filetype_hook', ("filetype", "action"),
        _("Declare code to be run when loading a file of the given type. " 
            "action may be a python function or a string beginning with " \
            ": for an ex command")),
    ('filename_hook', ("pattern", "action"),
        _("Declare code to be run when loading a file whose name matches " \
            "pattern. action may be a python function or a string beginning " \
            "with : for an ex command")),
    ('scheme_hook', ("scheme", "action"),
        _("Declare code to be run when a certain colour scheme is applied. " \
            "action may be a python function or a string beginning " \
            "with : for an ex command")),
    ('clear_filetype_hook', ("filetype", "action = None"),
        _("Clear filetype hook for filetype and action. If " \
            "action isn't given, clears all hooks matching the filetype")),
    ('clear_filename_hook', ("pattern", "action = None"),
        _("Clear filename hook for matching pattern and action. If " \
            "action isn't given, clears all hooks matching pattern")),
    ('clear_scheme_hook', ("scheme", "action = None"),
        _("Clear scheme hook for scheme and action. If " \
            "action isn't given, clears all hooks matching the scheme"))
]
    

class RCContext(object):
    """ This should be used as a singleton, because any functions etc defined
    by the user will always refer to the same RCContext. """

    locals = {}

    def __init__(self, settings_client = settings.defaults,
            priority = settings.PRIORITY_RC_SYSTEM, binds = bindings.defaults):
        self.settings = settings_client
        self.priority = priority
        self.bindings = binds
        self.priority_stack = []
        self.globals = { 'vixn_context': self }
        for f, n1, n2 in public_functions:
            self.globals[f] = getattr(self, f)
    
    def push_priority(self, priority):
        self.priority_stack.append(self.priority)
        self.priority = priority
    
    def pop_priority(self):
        self.priority = self.priority_stack.pop()
    
    def execfile(self, filename, win = None):
        """ Executes the given file. """
        try:
            exec(compile(open(filename).read(), filename, 'exec'), self.globals, self.locals)
        except:
            dialog.exception(win, _("When parsing '%s': ") % filename)
    
    def eval(self, command):
        """ Executes the given python command. """
        eval(command, self.globals, self.locals)
    
    def run_function(self, func):
        "Runs given function in this context. Bit of a kludge, but it works."
        self.globals['vixn_user_func__'] = func
        eval('vixn_user_func__()', self.globals, self.locals)
    
    def __load_config_file(self, leafname, win = None):
        settings.auto_save.initial_load = True
        try:
            vixrc = dirs.get_first_config(leafname)
            if vixrc:
                self.execfile(vixrc, win)
        except:
            dialog.exception(win, _("Error loading config file '%s'") % vixrc)
        settings.auto_save.initial_load = False
    
    def load_config(self, win = None):
        self.__load_config_file("rc", win)
    
    def load_auto_save(self, win = None):
        self.__load_config_file("autosave", win)
    
    def load_menus(self):
        self.create_lang_menu()
        self.create_scheme_menu()
        menusrc = None
        default_menusrc = dirs.get_file_for_current_lang(
                os.path.join(dirs.vixn_dir, "config", "menus"))
        if default_menusrc:
            self.execfile(default_menusrc)
        for d in dirs.vixn_config_dirs:
            menusrc = dirs.get_file_for_current_lang(os.path.join(d, "menus"))
            if menusrc:
                self.execfile(menusrc)
                break
        else:
            if not default_menusrc:
                raise vixexcept.FileError(
                        _("No menus definition file found"))
        # shortcuts need to be reloaded each time menus are loaded
        shortcuts.load_defaults()
    
    def create_lang_menu(self):
        def section_sort_key(s):
            k = s[0].lower()
            if k == 'others' : k = 'zzzz'
            return k

        def lang_sort_key(l):
            k = l[1].lower()
            if k[0] == '.' : k = k[1:]
            return k

        glm = langs.get_manager()
        sections = {}
        for i in glm.get_language_ids():
            l = glm.get_language(i)
            sec_name = l.get_section()
            s = sections.get(sec_name)
            if not s:
                s = []
                sections[sec_name] = s
            s.append(['%s/%s/%s' % ('Filetype', sec_name, i),
                    l.get_name(),
                    ':set filetype=' + i])
        sections = list(sections.items())
        sections.sort(key = section_sort_key)
        sec_menu_items = [('Settings/Filetype/' + s[0], s[0]) for s in sections]
        self.def_menu('Filetype', *sec_menu_items)
        for s in sections:
            s[1].sort(key = lang_sort_key)
            self.def_menu('Filetype_' + s[0], *s[1])
            self.attach_submenu('Settings/Filetype/' + s[0], 'Filetype_' + s[0])
        
    def create_scheme_menu(self):
        gssm = GtkSource.StyleSchemeManager.get_default()
        schemes = gssm.get_scheme_ids()
        schemes = [('Settings/Scheme/' + i,
                gssm.get_scheme(i).get_name(),
                ':set scheme=' + i) \
                for i in schemes]
        schemes.sort(key = lambda x: x[1].lower())
        self.def_menu('Scheme', *schemes)

    
    #
    # Functions which can be called from rc files
    #
    def set(self, key, value):
        self.settings.set_setting(key, value, self.priority)
    
    def setlocal(self, key, value):
        self.settings.set_setting(key, value, self.priority, True)
    
    def get_setting(self, key):
        return self.settings.get_setting(key, value)
    
    def bind(self, keystroke, command):
        try:
            vicmd.bind_command(self.bindings, keystroke, command)
        except:
            dialog.exception(_("Unable to bind"))
    
    def def_menu(self, *args):
        return menus.Menu(*args)
    
    def def_menu_item(self, *args):
        return menus.MenuItem(*args)

    def attach_submenu(self, item, submenu):
        item = menus.items[item]
        submenu = menus.menus[submenu]
        item.attach_submenu(submenu)
    
    def _(self, s):
        return _(s)
    
    def filetype_hook(self, *args):
        return hooks.filetype.add_hook(*args)

    def filename_hook(self, *args):
        return hooks.filename.add_hook(*args)

    def scheme_hook(self, *args):
        return hooks.scheme.add_hook(*args)

    def clear_filetype_hook(self, *args):
        hooks.filetype.clear_hooks(*args)

    def clear_filename_hook(self, *args):
        hooks.filename.clear_hooks(*args)

    def clear_scheme_hook(self, *args):
        hooks.scheme.clear_hooks(*args)
    
    #
    
context = RCContext()
