from gi.repository import Gtk

from . import vixiter

""" Constants representing which corner cursor is at """
C_LEFT = 0
C_TOP = 1
C_BOTTOM = 2
C_RIGHT = 3

""" Which edge to extend/shrink the selection; 0 = no change """
EXTEND_RIGHT = 1
SHRINK_RIGHT = 2
EXTEND_LEFT = 3
SHRINK_LEFT = 4
EXTEND_TOP = 5
SHRINK_TOP = 6
EXTEND_BOTTOM = 7
SHRINK_BOTTOM = 8

class RectSel(object):
    def __init__(self, buf, coords = None):
        self.buffer = buf
        style = buf.get_view().get_style_context()
        bg = style.get_background_color(Gtk.StateFlags.SELECTED)
        fg = style.get_color(Gtk.StateFlags.SELECTED)
        self.tag = buf.create_tag(background_full_height = True,
                background_full_height_set = True,
                background_rgba = bg,
                background_set = True,
                foreground_rgba = fg,
                foreground_set = True)
        if coords:
            self.left = coords[0]
            self.top = coords[1]
            self.right = coords[2]
            self.bottom = coords[3]
            for l in range(self.top, self.bottom + 1):
                self.__apply_at_row(self.left, self.right + 1, l)
            iter = buf.get_checked_iter(self.left, self.top)
            buf.place_cursor(iter)
        else:
            start = buf.get_iter_at_insert()
            end = start.copy()
            end.forward_cursor_position()
            buf.apply_tag(self.tag, start, end)
            self.left, self.top = vixiter.get_iter_pos(start)
            # Bottom and right are inclusive
            self.right = self.left
            self.bottom = self.top
            if self.right < self.left:
                left = self.right
                self.right = self.left
                self.left = left
                self.corner_x = C_LEFT
            else:
                self.corner_x = C_RIGHT
            self.corner_y = C_BOTTOM
        self.__set_marks()

    def cancel(self):
        buf = self.buffer
        start = buf.get_checked_iter(self.left, self.top)
        end = buf.get_checked_iter(self.right + 1, self.bottom)
        buf.remove_tag(self.tag, start, end)
        buf.rectsel = None

    def move_to_iter(self, iter):
        pos = vixiter.get_iter_pos(iter)
        self.move_to_pos(pos[0], pos[1], iter.ends_line())

    def move_to_pos(self, x, y, eol = False):
        if x < self.left:
            if eol:
                h = 0
            else:
                h = EXTEND_LEFT
                self.corner_x = C_LEFT
        elif x > self.right:
            h = EXTEND_RIGHT
            self.corner_x = C_RIGHT
        elif x == self.left:
            if self.corner_x == C_RIGHT:
                h = SHRINK_RIGHT
            else:
                h = 0
            self.corner_x = C_LEFT
        elif x == self.right:
            if self.corner_x == C_LEFT:
                h = SHRINK_LEFT
            else:
                h = 0
            self.corner_x = C_RIGHT
        elif self.corner_x == C_RIGHT:
            h = SHRINK_RIGHT
        elif self.corner_x == C_LEFT:
            h = SHRINK_LEFT
        if y < self.top:
            v = EXTEND_TOP
            self.corner_y = C_TOP
        elif y > self.bottom:
            v = EXTEND_BOTTOM
            self.corner_y = C_BOTTOM
        elif y == self.top:
            if self.corner_y == C_BOTTOM:
                v = SHRINK_BOTTOM
            else:
                v = 0
            self.corner_y = C_TOP
        elif y == self.bottom:
            if self.corner_y == C_TOP:
                v = SHRINK_TOP
            else:
                v = 0
            self.corner_y = C_BOTTOM
        elif self.corner_y == C_BOTTOM:
            v = SHRINK_BOTTOM
        elif self.corner_y == C_TOP:
            v = SHRINK_TOP
        self.__change_application(x, y, h, v)

    def __change_application(self, x, y, h, v):
        if h == EXTEND_LEFT or h == SHRINK_LEFT:
            new_left = x
            new_right = self.right
        elif h == EXTEND_RIGHT or h == SHRINK_RIGHT:
            new_right = x
            new_left = self.left
        else:
            new_left = self.left
            new_right = self.right
        if y < self.top:
            for y0 in range(y, self.top):
                self.__apply_at_row(new_left, new_right + 1, y0)
            self.top = y
        elif v == SHRINK_TOP:
            for y0 in range(self.top, y):
                self.__remove_from_row(self.left, self.right + 1, y0)
            self.top = y
        if y > self.bottom:
            for y0 in range(self.bottom + 1, y + 1):
                self.__apply_at_row(new_left, new_right + 1, y0)
            self.bottom = y
        elif v == SHRINK_BOTTOM:
            for y0 in range(y + 1 , self.bottom + 1):
                self.__remove_from_row(self.left, self.right + 1, y0)
            self.bottom = y
        if h:
            for y0 in range(self.top, self.bottom + 1):
                if h == EXTEND_LEFT:
                    self.__apply_at_row(x, self.left + 1, y0)
                elif h == EXTEND_RIGHT:
                    self.__apply_at_row(self.right + 1, x + 1, y0)
                elif h == SHRINK_LEFT:
                    self.__remove_from_row(self.left, x, y0)
                elif h == SHRINK_RIGHT:
                    self.__remove_from_row(x + 1, self.right + 1, y0)
        self.left = new_left
        self.right = new_right
        self.__set_marks()

    def __set_marks(self):
        buf = self.buffer
        buf.set_vi_mark('<', buf.get_checked_iter(self.left, self.top))
        buf.set_vi_mark('>', buf.get_checked_iter(self.right + 1, self.bottom))

    def __change_row(self, left, right, row, func):
        buf = self.buffer
        i1 = buf.get_checked_iter(left, row)
        i2 = buf.get_checked_iter(right, row)
        if i1 and i2 and not (i1.ends_line() and func == buf.apply_tag):
            func(self.tag, i1, i2)

    def __apply_at_row(self, left, right, row):
        self.__change_row(left, right, row, self.buffer.apply_tag)

    def __remove_from_row(self, left, right, row):
        self.__change_row(left, right, row, self.buffer.remove_tag)

    def get_bounds(self):
        return [self.buffer.get_checked_iter(self.left, self.top),
                self.buffer.get_checked_iter(self.right + 1, self.bottom)]

    def get_iters_by_line(self):
        """ Returns pairs of iters at the start and end of each line in the
        selection block. """
        pairs = []
        for l in range(self.top, self.bottom + 1):
            pairs.append([self.buffer.get_checked_iter(self.left, l),
                    self.buffer.get_checked_iter(self.right + 1, l)])
        return pairs

    def get_marks_by_line(self):
        """ As get_iters_by_line() but creates marks. """
        marks = []
        for start, end in self.get_iters_by_line():
            marks.append([self.buffer.create_mark(None, start, True),
                    self.buffer.create_mark(None, end, False)])
        return marks

    def get_text(self):
        """ Returns a block of text with a newline at the end of each line
        from the selection block (no NL on last line). """
        lines = []
        for start, end in self.get_iters_by_line():
            lines.append(self.buffer.get_text(start, end))
        return '\n'.join(lines)

    def extend_to_eol(self):
        """ Extends right edge to end of longest line in selection. """
        for l in range(self.top, self.bottom + 1):
            i = self.buffer.get_iter_at_line(l)
            if not i.ends_line():
                i.forward_to_line_end()
                i = i.get_line_offset() + 1
                if i > self.right:
                    self.right = i


