""" Context for searching and results. """

import re

from . import dialog

" Status values "
INVALID = 0
NO_MATCH = 1
MATCHED = 2
WRAPPED_FORWARD = 3
WRAPPED_BACKWARD = 4


class Search(object):
    """ results are stored as a list of [start, end] iters. """
    def __init__(self, buf):
        self.buffer = buf
        self.invalidate()

    def invalidate(self):
        self.text = None
        self.results = None
        self.pattern = None
        self.start = None
        self.end = None
        self.flags = 0
        self.status = INVALID

    def is_valid(self):
        return self.status != INVALID

    def get_text(self, start, end):
        """ Returns False if text was already valid for this search """
        if self.text and \
                self.start.equal(start) and self.end.equal(end):
            return False
        else:
            self.text = self.buffer.get_text(start, end)
            self.start = start
            self.end = end
            self.results = None
            self.pattern = None
            return True

    def get_matches(self, pattern, backward = False,
            selection = False, flags = 0):
        global last_search
        flags |= re.M
        last_search = [pattern, flags, backward]
        if selection:
            bounds = self.buffer.get_selection_bounds()
            if bounds:
                start = bounds[0]
                end = bounds[1]
                self.get_text(start, end)
                if self.text == pattern:
                    selection = False
            else:
                selection = False
        if not selection:
            start = self.buffer.get_start_iter()
            end = self.buffer.get_end_iter()
        self.get_text(start, end)
        if self.pattern != pattern or self.flags != flags:
            self.results = []
            self.pattern = pattern
            self.flags = flags
            try:
                miter = re.finditer(pattern, self.text, flags)
            except re.error as e:
                dialog.error(self.buffer.get_vixwin(),
                        _("Invalid expression: %s") % str(e))
            else:
                o = start.get_offset()
                for m in miter:
                    self.results.append( \
                            [self.buffer.get_iter_at_offset(m.start() + o),
                            self.buffer.get_iter_at_offset(m.end() + o)])
        if self.results:
            self.status = MATCHED
        else:
            self.status = NO_MATCH

    def search_forward(self, iter, num = 1):
        if not self.results:
            self.status = NO_MATCH
            return None
        self.status = MATCHED
        l = len(self.results)
        for i in range(l):
            if iter.compare(self.results[i][0]) <= 0:
                break
        else:
            self.status = WRAPPED_FORWARD
            i = 0
        i += num - 1
        if i >= l:
            self.status = WRAPPED_FORWARD
        i %= l
        return self.results[i]

    def search_backward(self, iter, num = 1):
        if not self.results:
            self.status = NO_MATCH
            return None
        self.status = MATCHED
        rev = self.results[:]
        rev.reverse()
        l = len(rev)
        for i in range(l):
            if iter.compare(rev[i][0]) >= 0:
                break
        else:
            self.status = WRAPPED_BACKWARD
            i = 0
        i += num - 1
        if i >= l:
            self.status = WRAPPED_BACKWARD
        i %= l
        return rev[i]

last_search = None
" [pattern, flags, backward] "

last_lnsearch = None
" [type (method name as str), char to search for] "


class Substitute(object):
    def __init__(self, pattern, replace, buf, start, end,
            ignore_case, all_on_a_line):
        flags = re.M
        if ignore_case:
            flags |= re.I
        #if "\\n" in pattern:
        #   self.nl = True
        #else:
        #   self.nl = False
        self.re = re.compile(pattern, flags)
        self.replace_pat = replace
        self.buffer = buf
        self.iter = start
        self.end = end
        self.all = all_on_a_line
        self.match = None

    def get_next_match(self):
        """ Returns [start, end]:- iters indicating where match was found.
        Or None. """
        if self.iter.compare(self.end) >= 0:
            return None
        text = self.buffer.get_text(self.iter, self.end)
        self.match = self.re.search(text)
        if self.match:
            self.match_start_iter = self.iter.copy()
            self.match_start_iter.forward_chars(self.match.start())
            self.match_end_iter = self.iter.copy()
            self.match_end_iter.forward_chars(self.match.end())
            self.iter = self.match_start_iter.copy()
            if self.all:
                # By default we want to start searching from one character
                # after current match, but replace will move on to end of
                # replaced string to avoid possibility of infinite loops
                # (eg s/./00/).
                self.iter.forward_cursor_position()
            else:
                self.iter.forward_line()
                self.iter.set_line_offset(0)
            return [self.match_start_iter, self.match_end_iter]
        else:
            return None

    def replace(self):
        # Set marks at places we need to preserve across replace
        match_end_mark = self.buffer.create_mark(None, self.match_end_iter)
        end_mark = self.buffer.create_mark(None, self.end)
        text = self.buffer.get_text(self.match_start_iter, self.match_end_iter)
        text = self.match.expand(self.replace_pat)
        self.buffer.begin_user_action()
        self.buffer.delete(self.match_start_iter, self.match_end_iter)
        self.buffer.insert(self.buffer.get_iter_at_mark(match_end_mark), text)
        self.iter = self.buffer.get_iter_at_mark(match_end_mark)
        if not self.all:
            self.iter.forward_line()
            self.iter.set_line_offset(0)
        self.end = self.buffer.get_iter_at_mark(end_mark)
        self.buffer.delete_mark(match_end_mark)
        self.buffer.delete_mark(end_mark)
        self.buffer.end_user_action()


def get_ignore_case(stngs, pattern):
    """ Returns whether to ignore case based on smartcase and/or ignorecase in
    stngs (SettingsClient) and pattern (in case of smartcase). """
    if stngs.lookup_setting("smartcase", False):
        for c in pattern:
            if c.isupper():
                return False
        else:
            return True
    else:
        return stngs.lookup_setting("ignorecase", False)
