import atexit
import os.path

from . import dialog
from . import dirs
from . import vixexcept

class SettingsError(vixexcept.CodeError):
    pass

""" When a setting is applied it applies to one of these. """
SCOPE_NONE = 0
SCOPE_BUFFER = 1
SCOPE_VIEW = 2
SCOPE_WINDOW = 3
NUM_SCOPES = 4


""" Priorities. Setting an option is ignored if it's already been set at a
higher priority (ordinally higher). This nicely solves two problems: user hooks
can override, eg filetype defaults, and settings explicitly set by user can
override rc files. """
PRIORITY_CLEAR = 0          # Overrides all other values but can then be
                            # overridden by others
PRIORITY_DEFAULT = 10       # Set from a default when looking up
PRIORITY_AUTO_FILENAME = 20 # For setting filetype by filename automatically
PRIORITY_RC_SYSTEM = 30     # rc file automatically loaded
PRIORITY_EMBEDDED = 40      # From commands embedded in a loaded file
PRIORITY_SCHEME_HOOK = 50   # Set from a scheme hook
PRIORITY_FILETYPE_HOOK = 60 # Set from a filetype hook
PRIORITY_FILENAME_HOOK = 70 # Set from a filename hook
PRIORITY_USER = 80          # Explicitly set by user
PRIORITY_RC_USER = 100      # Additional value to add when an rc file is
                            # explicitly loaded by user; only used to test
                            # whether to apply setting; this value is
                            # subtracted again before the priority is stored
                            # with the setting. """


class AutoSave(object):
    """ Handles saving of settings with auto_save flag. Actual save is deferred
    until exit. """
    def __init__(self):
        self.settings = {}
        self.dirty = False
        self.initial_load = False

    def set(self, key, value):
        old = self.settings.get(key)
        if old == value:
            return
        self.settings[key] = value
        if not self.dirty and not self.initial_load:
            atexit.register(self.save)
            self.dirty = True

    def save(self):
        try:
            filename = dirs.save_config_path("autosave")
            fp = open(filename, "w")
            for k, v in list(self.settings.items()):
                fp.write("set('%s', %s)\n" % (k, repr(v)))
            fp.close()
        except:
            dialog.exception(None, _("Unable to save %s") % filename)

auto_save = AutoSave()


class Description(object):
    """ Base class for the description of a setting: how to apply the value to
    and get a value from the target an implementation of the setting is bound
    to; the value is not stored in objects of this class or its derivations. """

    def __init__(self, key, type = str, get_method = None, set_method = None,
            description = None, scope = SCOPE_NONE, methods_need_key = False,
            auto_save = False):
        """ Typically to set/get a value a method is called on the target
        a setting is bound to. Supply those methods here
        if you intend to use the standard set() and get().
        If skip is True the setting will be skipped when running apply_all().
        If methods_need_key is True get_method is called as get_method(key)
        otherwise just get_method(). Similarly set_method(k,v)/set_method(v).
        """
        self.key = key
        self.type = type
        self.get_method = get_method
        self.set_method = set_method
        self.description = description
        self.scope = scope
        self.methods_need_key = methods_need_key
        self.auto_save = auto_save

    def get_from(self, target):
        """ target is the object a setting is bound to. """
        if self.get_method:
            if self.methods_need_key:
                return self.get_method(target, self.key)
            else:
                return self.get_method(target)
        else:
            return None

    def apply_to(self, target, value):
        """ target is the object a setting is bound to. Returns False if
        there's no way to apply setting. """
        if self.set_method and target:
            if self.methods_need_key:
                self.set_method(target, self.key, value)
            else:
                self.set_method(target, value)
            return True
        else:
            return False

    def get_key(self):
        " Always use this method to get the key name in case it's a synonym. "
        return self.key

    def get_scope(self):
        " Always use this method to get the scope in case it's a synonym. "
        return self.scope

    def get_type(self):
        " Always use this method to get the type in case it's a synonym. "
        return self.type

    def get_auto_save(self):
        " Always use this method to get the type in case it's a synonym. "
        return self.auto_save


class Synonym(object):
    """ A setting's synonym. """

    def __init__(self, key, synonym_for):
        """ synonym_for is a Description object, not a key name. """
        self.key = key
        self.synonym_for = synonym_for

    def get_from(self, target):
        return self.synonym_for.get_from(target)

    def apply_to(self, target, value):
        self.synonym_for.apply_to(target, value)

    def get_scope(self):
        return self.synonym_for.get_scope()

    def get_key(self):
        return self.synonym_for.get_key()

    def get_type(self):
        return self.synonym_for.get_type()

    def get_auto_save(self):
        return self.synonym_for.get_auto_save()


class Registry(object):
    """ A singleton where all settings descriptions are registered. """

    def __init__(self):
        self.settings = {}

    def register(self, setting):
        """ Registers a Description object. If one with the same name already
        exists an exception is raised. """
        if setting.key in self.settings:
            raise SettingsError(_("Setting '%s' already registered") %
                    setting.key)
        self.settings[setting.key] = setting

    def lookup(self, key):
        """ Looks up a Description in the registry. """
        setting = self.settings.get(key)
        if not setting:
            raise SettingsError(_("No such setting '%s'") % key)
        return setting

registry = Registry()
" Singleton. Register all settings descriptions here before creating settings. "

def register(keys, type = str,
        get_method = None, set_method = None,
        description = None, scope = SCOPE_NONE, methods_need_key = False,
        auto_save = False):
    """ Convenience function to create and register a standard Description.
    keys can be a string (one key), or list of keys; extra members are
    synonyms. """
    if hasattr(keys, 'isspace'):
        main_key = keys
        synonyms = False
    else:
        main_key = keys[0]
        synonyms = True
    setting = Description(main_key, type, get_method, set_method,
            description, scope, methods_need_key, auto_save)
    global registry
    registry.register(setting)
    if synonyms:
        for k in keys[1:]:
            registry.register(Synonym(k, setting))


class Setting(object):
    def __init__(self, value = None, priority = PRIORITY_USER,
            applied = False):
        self.value = value
        self.priority = priority
        self.applied = False

    def copy(self, priority = None):
        if priority == None:
            priority = self.priority
        return Setting(self.value, priority)


class SettingsDict(object):
    """ Like a dict but with intelligent copy. """
    def __init__(self, copy_from = None, priority = None):
        self.dict = {}
        if copy_from:
            for k, v in list(copy_from.dict.items()):
                self.dict[k] = v.copy(priority)

    def copy(self, priority = None):
        return SettingsDict(self, priority)

    def __setitem__(self, key, value):
        self.dict[key] = value

    def __getitem__(self, key):
        return self.dict[key]

    def get(self, key, default = None):
        return self.dict.get(key, default)

    def items(self):
        return list(self.dict.items())

    def keys(self):
        return list(self.dict.keys())

    def values(self):
        return list(self.dict.values())

    def merge(self, other):
        " Settings from other override those in self irrespective of priority "
        if other is self:
            return
        for k, v in list(other.items()):
            self.dict[k] = v.copy()

_global_sd = SettingsDict()


""" client below is a buffer, view or setting. Each has methods get_buffer(),
get_view() and get_vixwin() and get_vixbuf_settings(), get_vixview_settings(),
get_vixwin_settings() ('vix' part to avoid clashing with GTK methods eg
get_settings(), get_window()) and copy_settings() (returning a valid client) so
that client need not be same as setting's scope.  Buffer also needs methods
get_views() and get_windows(). sd stands for SettingsDict. """


def _get_target_and_sd(client, scope):
    global _global_sd
    if client:
        if scope == SCOPE_BUFFER:
            target = client.get_buffer()
            sd = client.get_vixbuf_settings()
        elif scope == SCOPE_VIEW:
            target = client.get_view()
            sd = client.get_vixview_settings()
        elif scope == SCOPE_WINDOW:
            target = client.get_vixwin()
            sd = client.get_vixwin_settings()
        else:
            target = None
            sd = _global_sd
    return [target, sd]


def __lookup_or_get(client, key, use_target, default = None, applying = False):
    global registry
    desc = registry.lookup(key)
    scope = desc.get_scope()
    key = desc.get_key()
    target, sd = _get_target_and_sd(client, scope)
    if use_target:
        value = desc.get_from(target)
    else:
        value = None
    if value == None:
        # If can't get value from target, look up in dict or use default
        s = sd.get(key)
        if s:
            value = s.value
        else:
            value = default
    if value != None and applying:
        s = sd.get(key)
        if s:
            s.applied = True
        else:
            sd[key] = Setting(value, PRIORITY_DEFAULT, True)
    if value == None:
        return None
    else:
        return desc.get_type()(value)


def lookup(client, key, default = None, applying = False):
    """ Gets a setting's value from the dict. default is used if there's no
    value in dict. """
    return __lookup_or_get(client, key, False, default, applying)


def get(client, key):
    """ Gets a setting's value. If it has a get_method that's used in
    preference to value stored in dict. """
    return __lookup_or_get(client, key, True)


def set(client, key, value, priority = PRIORITY_USER, local = False):
    """ Sets a setting's value and applies it to target if possible. If local
    is False, view and window settings are applied to all buffer's views. """
    global registry
    desc = registry.lookup(key)
    scope = desc.get_scope()
    key = desc.get_key()
    value = desc.get_type()(value)
    ts = [_get_target_and_sd(client, scope)]
    if not local and client:
        client = client.get_buffer()
        if client:
            targets = None
            if scope == SCOPE_VIEW:
                targets = client.get_views()
            elif scope == SCOPE_WINDOW:
                targets = client.get_windows()
            if targets:
                ts = []
                for t in targets:
                    ts.append(_get_target_and_sd(t, scope))
    applied = False
    for target, sd in ts:
        s = sd.get(key)
        if s:
            if s.priority > priority and priority != PRIORITY_CLEAR:
                continue
            else:
                s.value = value
                s.priority = priority
        else:
            s = Setting(value, priority, False)
            sd[key] = s
        s.applied = desc.apply_to(target, value)
        applied = True
    if applied and desc.get_auto_save():
        global auto_save
        auto_save.set(key, value)



def apply_all(target, sd):
    global registry
    for key, s in list(sd.items()):
        if not s.applied:
            s.applied = registry.lookup(key).apply_to(target, s.value)


def mark_all_unapplied(sd):
    for s in list(sd.values()):
        s.applied = False


class SettingsClient(object):
    def lookup_setting(self, key, default = None, applying = False):
        return lookup(self, key, default, applying)

    def get_setting(self, key):
        return get(self, key)

    def set_setting(self, key, value, priority = PRIORITY_USER, local = False):
        set(self, key, value, priority, local)

    def copy_settings(self, priority = None):
        return SettingsDefaults(self, priority)

    def expandpath(self, filename):
        filename = os.path.expanduser(os.path.expandvars(filename))
        if not os.path.isabs(filename):
            filename = os.path.join(self.get_setting('cwd'), filename)
        return os.path.realpath(os.path.normpath(filename))

    def merge(self, other):
        " Settings from other override those in self irrespective of priority "
        if other is self:
            return
        self.get_vixbuf_settings().merge(other.get_vixbuf_settings())
        self.get_vixview_settings().merge(other.get_vixview_settings())
        self.get_vixwin_settings().merge(other.get_vixwin_settings())



class SettingsDefaults(SettingsClient):
    """ A pseudo client allowing defaults to be loaded from rc files. """
    def __init__(self, copy_from = None, priority = None):
        if copy_from:
            self.buffer_sd = copy_from.get_vixbuf_settings().copy(priority)
            self.view_sd = copy_from.get_vixview_settings().copy(priority)
            self.window_sd = copy_from.get_vixwin_settings().copy(priority)
        else:
            self.buffer_sd = SettingsDict()
            self.view_sd = SettingsDict()
            self.window_sd = SettingsDict()

    def get_vixbuf_settings(self):
        return self.buffer_sd

    def get_vixview_settings(self):
        return self.view_sd

    def get_vixwin_settings(self):
        return self.window_sd

    def get_buffer(self):
        return None

    def get_view(self):
        return None

    def get_vixwin(self):
        return None

    def copy_settings(self, priority = None):
        return SettingsDefaults(self, priority)


defaults = SettingsDefaults()

