import atexit
import os.path

#import xml.dom.ext
from xml.dom import minidom

from gi.repository import Gdk

from . import change
from . import clipboard
from . import dialog
from . import dirs
from . import keysyms
from . import misc
from . import mode
from . import msgs
from . import qrec
from . import search

STATE_FILE = dirs.save_config_path("state.xml")

MAX_FILES = 20


def get_xml_child_text(node):
    text = ""
    for t in node.childNodes:
        if t.nodeType == minidom.Node.TEXT_NODE:
            text += t.data
    return text
    

class File(object):
    def __init__(self, filename):
        self.filename = filename
        self.pos = None
        self.marks = {}


class State(object):
    def __init__(self, filename):
        self.filename = filename
        self.command_history = []
        self.search_history = []
        self.files = []
        if os.path.exists(STATE_FILE):
            try:
                dom = minidom.parse(filename)
                self.load_histories(dom)
                self.load_files(dom)
                self.load_last_search(dom)
                self.load_last_lnsearch(dom)
                self.load_last_change(dom)
                self.load_recordings(dom)
                self.load_clipboards(dom)
            except:
                dialog.exception(None,
                    _("State file '%s' appears to be invalid") % STATE_FILE)
        
    def load_histories(self, dom):
        histories = dom.getElementsByTagName("history")
        if not histories:
            return
        for hist in histories:
            if hist.hasChildNodes():
                for node in hist.childNodes:
                    if node.nodeType == minidom.Node.ELEMENT_NODE:
                        if node.nodeName == "command":
                            self.command_history = self.load_history(node)
                        elif node.nodeName == "search":
                            self.search_history = self.load_history(node)
    
    @staticmethod
    def load_history(hist_node):
        entries = []
        if not hist_node.hasChildNodes():
            return []
        for node in hist_node.childNodes:
            if node.nodeType == minidom.Node.ELEMENT_NODE and \
                    node.nodeName == "entry" and node.hasChildNodes:
                text = get_xml_child_text(node)
                if text:
                    entries.append(text)
        return entries

    @staticmethod
    def parse_pos(node):
        line = node.getAttribute("line")
        if line:
            line = int(line)
        else:
            line = 0
        column = node.getAttribute("column")
        if column:
            column = int(column)
        else:
            column = 0
        return [column, line]

    def load_files(self, dom):
        files = dom.getElementsByTagName("file")
        if not files:
            return
        for f in files:
            name = f.getAttribute("name")
            file = File(name)
            pos = f.getElementsByTagName("pos")
            if pos:
                file.pos = self.parse_pos(pos[0])
            marks = f.getElementsByTagName("mark")
            if marks:
                for m in marks:
                    file.marks[m.getAttribute("name")] = self.parse_pos(m)
            self.files.append(file)
    
    @staticmethod
    def load_last_search(dom):
        snode = dom.getElementsByTagName("last_search")
        if not snode:
            return
        snode = snode[0]
        search.last_search = [get_xml_child_text(snode),
                int(snode.getAttribute("flags")),
                snode.getAttribute("backward") != "0"]
                
    @staticmethod
    def load_last_lnsearch(dom):
        snode = dom.getElementsByTagName("last_lnsearch")
        if not snode:
            return
        snode = snode[0]
        search.last_lnsearch = [snode.getAttribute("method"),
                snode.getAttribute("char")]
                
    @staticmethod
    def load_last_change(dom):
        cnode = dom.getElementsByTagName("change")
        if not cnode:
            return
        cnode = cnode[0]
        seltype = int(cnode.getAttribute("seltype"))
        selsize = cnode.getAttribute("selsize")
        if seltype == mode.SELRECT:
            x, y = selsize.split(',')
            selsize = [int(x[1:]), int(y[:-1])]
        else:
            selsize = int(selsize)
        change.last = change.Change(None,
                keysyms.split_command(cnode.getAttribute("keychain")),
                seltype, selsize)
        change.last.deleted_before = int(cnode.getAttribute("deleted_before"))
        change.last.deleted_after = int(cnode.getAttribute("deleted_after"))
        ins = cnode.getElementsByTagName("insert")
        if ins:
            change.last.inserted = get_xml_child_text(ins[0])
    
    @staticmethod
    def load_recordings(dom):
        nodes = dom.getElementsByTagName("qrec")
        if not nodes:
            return
        for qnode in nodes:
            qr = qrec.Recording(qnode.getAttribute("reg"))
            knodes = qnode.getElementsByTagName("ks")
            for knode in knodes:
                e = Gdk.EventKey()
                e.type = Gdk.EventType.KEY_PRESS
                e.state = Gdk.ModifierType(int(knode.getAttribute("state")))
                e.keyval = int(knode.getAttribute("keyval"))
                hardware_keycode = knode.getAttribute("hardware_keycode")
                if hardware_keycode:
                    e.hardware_keycode = int(hardware_keycode)
                group = knode.getAttribute("group")
                if group:
                    e.group = int(group)
                qr.add_keystroke(e)
    
    @staticmethod
    def load_clipboards(dom):
        nodes = dom.getElementsByTagName("clipboard")
        if not nodes:
            return
        for node in nodes:
            name = node.getAttribute("name")
            if not name:
                msgs.warning( \
                        _("Saved state file contains clipboard with no name"))
                continue
            mo = node.getAttribute("mode")
            if mo:
                mo = int(mo)
            else:
                mo = mode.SELECT
            text = get_xml_child_text(node)
            if text:
                clipboard.clipboards[name] = [text, mo]
        
    """ By 'save' we mean convert to XML strings; return "" rather than None
        for empty sections """
    
    @staticmethod
    def save_history(tag_name, history):
        if not history:
            return ""
        h = ""
        for entry in history:
            h += "<entry>" + misc.esc_xml_text(entry) + "</entry>\n"
        return "<" + tag_name + ">\n" + h + "</" + tag_name + ">\n"
    
    def save_histories(self):
        c = self.save_history("command", self.command_history)
        s = self.save_history("search", self.search_history)
        if c or s:
            return "<history>\n" + c + s + "</history>\n"
        else:
            return ""
    
    @staticmethod
    def get_pos_attr(pos):
        return misc.xml_attr_to_str("column", str(pos[0])) + \
            misc.xml_attr_to_str("line", str(pos[1]))
    
    def save_files(self):
        if not self.files:
            return ""
        result = ""
        for f in self.files:
            result += "<file" + misc.xml_attr_to_str("name", f.filename) + ">\n"
            if f.pos:
                result += "<pos" + self.get_pos_attr(f.pos) + "/>\n"
            for n, p in list(f.marks.items()):
                result += "<mark" + \
                        misc.xml_attr_to_str("name", n) + \
                        self.get_pos_attr(p) + "/>\n"
            result += "</file>\n"
        return result
    
    @staticmethod
    def save_last_search():
        if not search.last_search:
            return ""
        return "<last_search" + \
                misc.xml_attr_to_str("flags", search.last_search[1]) + \
                misc.xml_attr_to_str("backward", int(search.last_search[2])) + \
                ">" + misc.esc_xml_text(search.last_search[0]) + \
                "</last_search>\n"
    
    @staticmethod
    def save_last_lnsearch():
        if not search.last_lnsearch:
            return ""
        return "<last_lnsearch" + \
                misc.xml_attr_to_str("method", search.last_lnsearch[0]) + \
                misc.xml_attr_to_str("char", search.last_lnsearch[1]) + "/>\n"
    
    @staticmethod
    def save_last_change():
        if not change.last:
            return ""
        c = "<change" + \
                misc.xml_attr_to_str("keychain",
                        keysyms.serialise(change.last.keychain)) + \
                misc.xml_attr_to_str("seltype", change.last.seltype) + \
                misc.xml_attr_to_str("selsize", change.last.selsize) + \
                misc.xml_attr_to_str("deleted_before",
                        change.last.deleted_before) + \
                misc.xml_attr_to_str("deleted_after",
                        change.last.deleted_after)
        if change.last.inserted:
            c += ">\n<insert>" + misc.esc_xml_text(change.last.inserted) + \
                    "</insert>\n</change>\n"
        else:
            c += "/>\n"
        return c
    
    @staticmethod
    def save_recordings():
        s = ""
        for r in list(qrec.recordings.values()):
            s += r.to_xml()
        return s
    
    @staticmethod
    def save_clipboards():
        s = ""
        for n, clip in list(clipboard.clipboards.items()):
            if clip[0]:
                s += "<clipboard" + misc.xml_attr_to_str("name", n) + \
                        misc.xml_attr_to_str("mode", clip[1]) + ">" + \
                        misc.esc_xml_text(clip[0]) + "</clipboard>\n"
        return s
    
    def save_all(self):
        h = self.save_histories()
        f = self.save_files()
        s = self.save_last_search()
        l = self.save_last_lnsearch()
        c = self.save_last_change()
        r = self.save_recordings()
        p = self.save_clipboards()
        if f or h or s or l or c or r or p:
            fp = open(self.filename, 'w')
            fp.write("<vixn_state>\n" + h + f + s + l + c + r + p +
                    "</vixn_state>\n")
            fp.close()
    
    def find_file(self, filename):
        """ Returns an existing File object for the given filename (must be
        canonical), or None if the file is unknown """
        for f in self.files:
            if f.filename == filename:
                return f
        else:
            return None
    
    def get_file(self, filename):
        """ Returns a File object for the given filename (must be canonical),
        creating a new one if necessary """
        f = self.find_file(filename)
        if f:
            self.touch_file(f)
        else:
            f = File(filename)
            self.append_file(f)
        return f
    
    def append_file(self, fobj):
        if len(self.files) >= MAX_FILES:
            self.files.pop(0)
        self.files.append(fobj)
    
    def touch_file(self, fobj):
        """ Moves the given File object to the end of the list so it's expired
        last """
        for n in range(len(self.files)):
            f = self.files[n]
            if f.filename == fobj.filename:
                del self.files[n]
                break
        self.append_file(fobj)
        
        
        
state = State(STATE_FILE)
atexit.register(state.save_all)

