import locale
import os
import select
import subprocess

from gi.repository import GLib, GObject

from . import msgs, settings

class StreamToBuffer(object):
    """ Reads from file descriptors into a buf. """

    BAD_CONDS = GLib.IOCondition.ERR | \
            GLib.IOCondition.HUP | GLib.IOCondition.NVAL

    def __init__(self, buf, stop_callback = None, fds = None,
            disposable = True):
        """ See also add_fds() below. If disposable is True, buffer is marked
            unmodified when stream is closed. """
        self.disposable = disposable
        self.stop_callback = stop_callback
        self.buffer = buf
        buf.begin_user_action()
        self.__reset()
        if fds:
            self.add_fds(fds)

    def __reset(self):
        self.polls = {}

    def add_fds(self, fds):
        """ fds may be a single file-like object (anything with a fileno()
        method), int file descriptor or a list of them. The contents of
        multiple streams will be interleaved arbitrarily; the main use of
        multiple sources will be stdout and stderr; hopefully that will work
        out OK in most cases. """
        if not isinstance(fds, list) and not isinstance(fds, tuple):
            fds = [fds]
        for f in fds:
            if not isinstance(f, int):
                f = f.fileno()
            self.polls[f] = GLib.io_add_watch(f,
                    GLib.IOCondition.IN | self.BAD_CONDS,
                    self.__fd_callback)

    def prepare(self, *args):
        return (False, -1)

    def check(self, *args):
        result = False
        for p in self.polls:
            if p.revents & GLib.IOCondition.IN:
                result = True
            if p.revents & self.BAD_CONDS:
                # Need to test for BAD_CONDS so we know when files have been
                # closed and we can stop reading, but apparently a BAD_COND
                # can arise when socket still has data to be read, so defer
                # removal.
                GLib.idle_add(self.__remove_poll, p)
        return result

    def __remove_poll(self, p):
        # p is an fd and polls is a dict of fd:source-tag
        GObject.source_remove(self.polls[p])
        del self.polls[p]
        if not self.polls:
            self.stop_reading()
        return False

    def dispatch(self, callback, *args):
        fds = []
        for p in self.polls:
            if p.revents & GLib.IOCondition.IN:
                fds.append(p.fd)
        callback(fds, args)
        return len(fds) != 0

    def __source_callback(self, fds, *args):
        for f in fds:
            self.__read_fd(f)

    def __read_fd(self, f):
        t = os.read(f, 4096)
        if t:
            self.got_text(f, str(t, "utf8"))

    def __fd_callback(self, fd, condition):
        if condition & GLib.IOCondition.IN:
            try:
                self.__read_fd(fd)
            except UnicodeDecodeError as x:
                msgs.exception("")
        if condition & self.BAD_CONDS:
            self.__remove_poll(fd)
        return True

    def stop_reading(self):
        """ Returns False so can be used as idle callback. May be safer to do
        it that way than call direct from one of GSource callbacks, but not
        sure if it's really necessary.  """
        del self.polls
        self.__reset()
        self.buffer.end_user_action()
        if self.disposable:
            self.buffer.set_modified(False)
        if self.stop_callback:
            self.stop_callback()
        return False

    def got_text(self, fd, text):
        """ Called when a bit of text has been read from one of the inputs.
        This may be overridden eg to trap compiler errors. """
        b = self.buffer
        b.insert(b.get_end_iter(), text)
        b.place_cursor(b.get_end_iter())



class TermToBuffer(StreamToBuffer):
    """ A buf used for displaying the output of a command. """
    def __init__(self, buf, stop_callback = None,
            cmd_args = None, capture_stderr = True, cwd = None, shell = False,
            disposable = True):
        self.capture_stderr = capture_stderr
        StreamToBuffer.__init__(self, buf, stop_callback,
                disposable = disposable)
        cwd = self.check_cwd(cwd)
        if cmd_args:
            self.run_command(cmd_args, capture_stderr, cwd, shell)
        else:
            self.proc = None

    def check_cwd(self, cwd):
        if not cwd:
            cwd = self.buffer.get_cwd()
        return cwd

    def run_command(self, cmd_args = None, capture_stderr = True,
            cwd = None, shell = False):
        """ If capture_stderr is True stderr is captured into buf too,
        otherwise discarded. Either way, stdout and stderr args are both PIPE.
        """
        env = os.environ.copy()
        # We're not a terminal
        if env.get('TERM'):
            del env['TERM']
        # Try to force command to output UTF-8
        lang = locale.getdefaultlocale()[0] + '.UTF-8'
        env['LANG'] = lang
        env['LANGUAGE'] = lang
        env['LC_ALL'] = lang
        env['LC_CTYPE'] = lang
        cwd = self.check_cwd(cwd)
        self.proc = subprocess.Popen(cmd_args,
                stdout = subprocess.PIPE, stderr = subprocess.PIPE,
                shell = shell, cwd = cwd, env = env)
        self.capture_stderr = capture_stderr
        self.add_fds([self.proc.stdout, self.proc.stderr])

    def stop_reading(self):
        if self.proc:
            self.proc.wait()
            self.proc = None
        return StreamToBuffer.stop_reading(self)

    def got_text(self, fd, text):
        if self.capture_stderr or not self.proc or \
                fd != self.proc.stderr.fileno():
            StreamToBuffer.got_text(self, fd, text)

