import os
import sys
from curses import ascii

from gi.repository import Gdk, Gtk

from . import change
from . import clipboard
from . import dialog
from . import format
from . import keysyms
from . import mode
from . import msgs
from . import qrec
from . import rectsel as mod_rectsel
from . import search
from .vicflags import *
from . import vixexcept
from . import vixiter


class ViCommand(object):
    """ Representation of a single vi command. """
    def __init__(self, method, flags, help = None, dflt_key = None):
        """ method is an aunbound method to be called with a ViInterpreter as
        its argument. flags is a value from vicflags. help is a string
        describing the command. dflt_key is the default keystroke (sequence)
        for the command, which may be a single string or a list of strings in
        case there is more than one keystroke (sequence); if None the method
        name is used, so it must be a valid key sequence. """
        self.method = method
        self.flags = flags
        self.help = help
        if dflt_key == None and method:
            dflt_key = method.__name__
        self.dflt_key = dflt_key


class ReboundExCommand(object):
    """ Allows user to bind ex commands as if they were vi commands. """
    def __init__(self, ex_command):
        self.flags = EXCOMMAND
        self.ex_command = ex_command

    def method(self, interp):
        interp.window.excmd.execute(self.ex_command, interp.num)


class BoundViCommand(object):
    """ ViCommand bound to its key sequence, which may require extra flags
    eg SEQUENCE. """
    def __init__(self, vi_command, extra_flags = 0):
        if vi_command:
            self.method = vi_command.method
            self.flags = vi_command.flags | extra_flags
        else:
            self.method = None
            self.flags = extra_flags


commands = {}


def command(flags, help = None, dflt_key = None):
    """ Use as a decorator to register a vi command. Args are as described
    for ViCommand's constructor. """
    def decorator(method):
        commands[method.__name__] = ViCommand(method, flags, help, dflt_key)
        return method
    return decorator


def bind_command(keydict, keystroke, vi_command):
    """ vi_command can be a method name instead of a ViCommand, in which case
    it's looked up in commands. If you add a command that clashes with a
    sequence eg g vs gqap, the last added will be used, breaking any other
    commands starting with g. However, making g a sequence again would restore
    the other g* commands. See keysyms for special keys. """
    if not hasattr(vi_command, 'method'):
        if vi_command[0] == ':':
            vi_command = ReboundExCommand(vi_command[1:])
        else:
            global commands
            s = vi_command
            vi_command = commands.get(vi_command)
            if not vi_command:
                raise vixexcept.UserError(_("No such vi command %s") % s)
    flags = vi_command.flags
    # Explicit INSERTs are also CLEARSEL, but COMPOUND INSERTs aren't
    # necessarily CLEARSEL
    if flags & INSERT:
        flags |= CHANGE | CLEARSEL
    if flags & COMPOUND:
        flags |= INSERT | CHANGE
    if (flags & NEEDSRANGE) or (flags & NEEDSARG) or (flags & DIGIT):
        flags |= INCOMPLETE
    if flags & ZERO:
        flags |= MOVEMENT
    ks = keysyms.split_command(keystroke)
    ser = keysyms.serialise(ks)
    keydict[ser] = BoundViCommand(vi_command, flags)
    # All substrings at start of sequenced command are SEQUENCEs. You need to be
    # aware that flags are ORred with SEQUENCE which could lead to
    # inconsistencies if different types of command share similar prefixes.
    if len(ks) > 1:
        acc = []
        for k in ks[:-1]:
            acc.append(k)
            s = keysyms.serialise(acc)
            old = keydict.get(s)
            if old:
                old.flags |= SEQUENCE | INCOMPLETE
            else:
                keydict[s] = BoundViCommand(None, SEQUENCE | INCOMPLETE | flags)


""" Built-in vi commands follow. """

@command(MOVEMENT, _("Move cursor backwards one character"),
        ('h', '<Left>', '<kLeft>'))
def h(interp):
    interp.view.move_horiz(-interp.num)

@command(MOVEMENT, _("Move cursor down one line"),
        ('j', '<Down>', '<kDown>'))
def j(interp):
    interp.view.move_vert(interp.num)

@command(MOVEMENT, _("Move cursor up one line"),
        ('k', '<Up>', '<kUp>'))
def k(interp):
    interp.view.move_vert(-interp.num)

@command(MOVEMENT, _("Move cursor forwards one character"),
        ('l', '<Right>', '<kRight>'))
def l(interp):
    interp.view.move_horiz(interp.num)

def _start_selection(interp, mo):
    if interp.buffer.selmode:
        interp.view.clear_selection()
    else:
        interp.save_cursor_pos()
        interp.view.start_selection(mo)

@command(SELECT | CLEARSEL, _("Initiate selection, character at a time"))
def v(interp):
    _start_selection(interp, mode.SELECT)

@command(SELECT | CLEARSEL, _("Initiate selection, line at a time"))
def V(interp):
    _start_selection(interp, mode.SELBYLINE)

@command(SELECT | CLEARSEL, _("Initiate a rectangular selection"), '<C-V>')
def rectsel(interp):
    _start_selection(interp, mode.SELRECT)

@command(MOVEMENT,
        _("Move forwards by a word defined by the [buf:wordchars] setting"))
def w(interp):
    interp.view.move_by_words(interp.num, True, False)

@command(MOVEMENT, _("Move forwards by a non-space word"))
def W(interp):
    interp.view.move_by_words(interp.num, True, True)

@command(MOVEMENT, _("Move backwards by an alphanumeric word"))
def b(interp):
    interp.view.move_by_words(-interp.num, False, False)

@command(MOVEMENT, _("Move backwards by a non-space word"))
def B(interp):
    interp.view.move_by_words(-interp.num, False, True)

@command(MOVEMENT, _("Move to end of alphanumeric word"))
def e(interp):
    interp.view.move_by_words(interp.num, False, False)

@command(MOVEMENT, _("Move to end of non-space word"))
def E(interp):
    interp.view.move_by_words(interp.num, False, True)

@command(MOVEMENT, _("Move forwards by a sentence"), ')')
def next_sentence(interp):
    interp.view.move_by_sentences(interp.num)

@command(MOVEMENT, _("Move backwards by a sentence"), '(')
def prev_sentence(interp):
    interp.view.move_by_sentences(-interp.num)

@command(MOVEMENT, _("Move forwards by a paragraph"), '}')
def next_para(interp):
    interp.view.move_by_paras(interp.num)

@command(MOVEMENT, _("Move backwards by a paragraph"), '{')
def prev_para(interp):
    interp.view.move_by_paras(-interp.num)

@command(MOVEMENT, _("Move to first non-space character on next line"),
        ('<Return>', '<kEnter>'))
def next_line(interp):
    interp.view.move_vert(interp.num)
    interp.view.sol()

@command(ZERO, _("Move to first column on current line"), ('0', '<k0>'))
def bol(interp):
    interp.view.move_to_column(0)

@command(MOVEMENT, _("Move to first non-space character on current line"),
        ('^', '<Home>', '<kHome>'))
def sol(interp):
    interp.view.sol()

@command(MOVEMENT, _("Move to end of current line"),
        ('$', '<End>', '<kEnd>'))
def eol(interp):
    interp.view.eol(interp.owner != None)

@command(MOVEMENT, _("Move to nth column on current line"), '|')
def column(interp):
    interp.view.move_to_column(interp.num - 1)

@command(MOVEMENT,
        _("Move to first line visible in window, or n lines after it"))
def H(interp):
    interp.view.move_to_top_of_window(interp.num - 1)

@command(MOVEMENT, _("Move to middle line visible in window"))
def M(interp):
    interp.view.move_to_middle_of_window()

@command(MOVEMENT,
        _("Move to last line visible in window, or n lines before it"))
def L(interp):
    interp.view.move_to_bottom_of_window(-interp.num)

@command(COMMAND, _("Scroll view down by height of window"),
        ('<C-F>', '<PageDown>', '<kPageDown>', ' '))
def scroll_down(interp):
    interp.view.scroll_by_page(interp.num)

@command(COMMAND, _("Scroll view up by height of window"),
        ('<C-B>', '<PageUp>', '<kPageUp>', '<BS>'))
def scroll_up(interp):
    interp.view.scroll_by_page(-interp.num)

@command(COMMAND, _("Scroll view down by half height of window"), '<C-D>')
def scroll_down_half(interp):
    interp.view.scroll_half_page(interp.num)

@command(COMMAND, _("Scroll view up by half height of window"), '<C-U>')
def scroll_up_half(interp):
    interp.view.scroll_half_page(-interp.num)

@command(COMMAND, _("Scroll view down by one line"), '<C-E>')
def scroll_down_line(interp):
    interp.view.scroll_by_line(interp.num)

@command(COMMAND, _("Scroll view up by one line"), '<C-Y>')
def scroll_up_line(interp):
    interp.view.scroll_by_line(-interp.num)

@command(COMMAND, _("Scroll view so cursor is at top of window"),
        ('z<Return>', 'z<kEnter>'))
def reposn_top(interp):
    interp.view.reposition(0)

@command(COMMAND, _("Scroll view so cursor is at bottom of window"), 'z.')
def reposn_bottom(interp):
    interp.view.reposition(1)

@command(COMMAND, _("Scroll view so cursor line is at middle of window"), 'z-')
def reposn_middle(interp):
    interp.view.reposition(0.5)

@command(COMMAND, _("Redraw window"), '<C-L>')
def redraw(interp):
    interp.view.queue_draw()

@command(COMMAND, _("Open mini-buf to enter an ex command"), ':')
def excmd(interp):
    interp.view.cancel_recording()
    if interp.buffer.get_has_selection():
        initial = ":'<,'>"
    else:
        initial = ':'
    interp.window.status_bar.open_input(interp.view,
            interp.buffer.get_cwd(),
            initial, interp.window.excmd.execute, interp.num)

@command(COMMAND, _("Open mini-buf to enter term to search forwards"), '/')
def search_forward(interp):
    interp.view.cancel_recording()
    interp.window.status_bar.open_input(interp.view,
            interp.buffer.get_cwd(),
            '/', interp.window.excmd.execute, interp.num)

@command(COMMAND, _("Open mini-buf to enter term to search backwards"), '?')
def search_backward(interp):
    interp.view.cancel_recording()
    interp.window.status_bar.open_input(interp.view,
            interp.buffer.get_cwd(),
            '?', interp.window.excmd.execute, interp.num)

@command(COMMAND,
        _("Open mini-buf to enter term to search with results to throwback"),
        r'\\')
def search_throwback(interp):
    interp.view.cancel_recording()
    interp.window.status_bar.open_input(interp.view,
            interp.buffer.get_cwd(),
            '\\', interp.window.excmd.execute, interp.num)

@command(MOVEMENT, _("Repeat previous search"))
def n(interp):
    interp.view.show_search_result(interp.buffer.search_forward(interp.num))

@command(MOVEMENT, _("Repeat previous search in reverse direction"))
def N(interp):
    interp.view.show_search_result(interp.buffer.search_backward(interp.num))

@command(NEEDSARG | MOVEMENT,
        _("Move forwards on current line to next instance of given character"))
def f(interp):
    search.last_lnsearch = ['f', interp.arg]
    iter = interp.forward_find_char()
    interp.one_more(iter)
    interp.view.move_cursor_to_iter(iter, True)

@command(NEEDSARG | MOVEMENT,
    _("Move backwards on current line to previous instance of given character"))
def F(interp):
    search.last_lnsearch = ['F', interp.arg]
    interp.view.move_cursor_to_iter(interp.backward_find_char(), True)

@command(NEEDSARG | MOVEMENT,
        _("Move forwards on current line to character preceding next instance "
        "of given character"))
def t(interp):
    search.last_lnsearch = ['t', interp.arg]
    iter1 = interp.buffer.get_iter_at_insert()
    iter2 = interp.forward_find_char()
    if not iter1.equal(iter2):
        iter2.backward_cursor_position()
        interp.one_more(iter2)
        interp.view.move_cursor_to_iter(iter2, True)

@command(NEEDSARG | MOVEMENT,
        _("Move backwards on current line to character following next instance "
        "of given character"))
def T(interp):
    search.last_lnsearch = ['T', interp.arg]
    iter1 = interp.buffer.get_iter_at_insert()
    iter2 = interp.backward_find_char()
    if not iter1.equal(iter2):
        iter2.forward_cursor_position()
        interp.view.move_cursor_to_iter(iter2, True)

@command(MOVEMENT, _("Repeat last current-line search for character"), ';')
def rpt_lnsearch(interp):
    global commands
    if search.last_lnsearch:
        interp.arg = search.last_lnsearch[1]
        commands[search.last_lnsearch[0]].method(interp)

@command(MOVEMENT,
    _("Repeat last current-line search for character in opposite direction"),
    ',')
def rpt_lnsearch_opposite(interp):
    if search.last_lnsearch:
        fn1 = search.last_lnsearch[0]
        fn2 = fn1
        if fn1 == 'f':
            fn1 = 'F'
        elif fn1 == 'F':
            fn1 = 'f'
        elif fn1 == 't':
            fn1 = 'T'
        elif fn1 == 'T':
            fn1 = 't'
        interp.arg = search.last_lnsearch[1]
        commands[fn1].method(interp)
        search.last_lnsearch[0] = fn2

@command(COMMAND, _("Show line number"), '<C-G>')
def show_line_number(interp):
    buf = interp.buffer
    if buf.get_modified():
        mod = _('[Modified] ')
    else:
        mod = ''
    interp.window.status_bar.show_message(_("%s %s%d lines") % \
            (os.path.basename(buf.get_default_filename()), mod,
            interp.buffer.get_iter_at_insert().get_line()))

@command(MOVEMENT, _("Move to first non-blank character on first line"))
def gg(interp):
    if interp.num:
        line = interp.num - 1
    else:
        line = 0
    interp.view.move_cursor_to_line(line)

@command(MOVEMENT,
        _("Move to first non-blank character on line n or last line"))
def G(interp):
    if interp.num > 1:
        line = interp.num - 1
    else:
        line = interp.buffer.get_line_count() - 1
    interp.view.move_cursor_to_line(line)

@command(NEEDSARG, _("Set mark in named register"))
def m(interp):
    interp.buffer.set_vi_mark(interp.arg, interp.buffer.get_iter_at_insert())

@command(NEEDSARG | MOVEMENT, _("Move to mark in named register"), '`')
def move_to_mark(interp, line = False):
    m = interp.buffer.get_vi_mark(interp.arg)
    if not m:
        interp.window.status_bar.show_message(_("Mark %s not set") % interp.arg)
        return
    i = interp.buffer.get_iter_at_mark(m)
    if line:
        vixiter.iter_to_first_non_space(i)
    interp.view.move_cursor_to_iter(i, True)

@command(NEEDSARG | MOVEMENT,
        _("Move to line containing mark in named register"), "'")
def move_to_mark_line(interp):
    move_to_mark(interp, True)

@command(COMMAND | CLEARSEL, _("Save buf and close view"))
def ZZ(interp):
    if interp.buffer.save(False):
        interp.window.destroy_view(interp.view)
        interp.window = None
    else:
        dialog.error(interp.window, _("Unable to save"))

@command(INSERT, _("Begin inserting text in overwrite mode"))
def R(interp):
    interp.begin_insert('R')

@command(INSERT, _("Begin inserting text"))
def i(interp):
    interp.begin_insert('i')

@command(INSERT, _("Begin inserting text at start of line (first non-space)"))
def I(interp):
    interp.view.sol()
    interp.begin_insert('I')

@command(INSERT, _("Begin appending text after current position"))
def a(interp):
    interp.begin_append('a')

@command(INSERT, _("Begin appending text at end of line"))
def A(interp):
    interp.view.eol()
    interp.begin_append('A')

@command(COMPOUND | CLEARSEL,
        _("Start entering text on new line after current line"))
def o(interp):
    iter = interp.buffer.get_iter_at_insert()
    if not iter.ends_line():
        iter.forward_to_line_end()
    interp.insert_nl_and_indentation(iter, True, False)
    interp.begin_insert('o')

@command(COMPOUND | CLEARSEL,
        _("Start entering text on new line before current line"))
def O(interp):
    iter = interp.buffer.get_iter_at_insert()
    iter.set_line_offset(0)
    interp.insert_nl_and_indentation(iter, True, True)
    interp.begin_insert('O')

@command(NEEDSARG, _("Next clipboard operation applies to given register"), '"')
def setreg(interp):
    interp.reg = interp.arg
    interp.arg = None
    interp.state = 0

@command(CHANGE, _("Join current and next line"))
def J(interp):
    buf = interp.buffer
    start, end = buf.get_working_range(False)
    if start.equal(end) and interp.num > 1:
        end.forward_lines(interp.num)
    # Don't join last line in a selection with next line
    join_last = (start.get_line() == end.get_line())
    if not join_last:
        end.backward_char()
    marks = []
    if not start.ends_line():
        start.forward_to_line_end()
    while start.compare(end) < 0:
        marks.append(buf.create_mark(None, start))
        start.forward_to_line_end()
    if join_last:
        marks.append(buf.create_mark(None, start))
    for m in marks:
        start = buf.get_iter_at_mark(m)
        end = start.copy()
        while True:
            end.forward_char()
            if end.equal(start):
                break
            after = end.copy()
            after.forward_char()
            if after.equal(end):
                break
            s = buf.get_text(end, after)
            if not s.isspace():
                break
        if not end.equal(start):
            buf.place_cursor(start)
            buf.delete(start, end)
            buf.insert_at_cursor(' ')
        buf.delete_mark(m)
    # mark-set signal gets sent while cursor is at eol so need to
    # give it another chance to be block cursor if desired
    #buf.set_selmode(0)

def _r_helper(buf, start, end, char):
    mark = buf.create_mark(None, start, True)
    text = buf.get_text(start, end)
    buf.delete(start, end)
    s = ""
    for c in text:
        if ord(c) < 32:
            s += c
        else:
            s += char
    start = buf.get_iter_at_mark(mark)
    buf.insert(start, s)
    buf.delete_mark(mark)

@command(CHANGE | NEEDSARG | REGMOD, _("Replace a character"))
def r(interp):
    buf = interp.buffer
    start, end = buf.get_working_range(False)
    if not interp.num:
        interp.num = 1
    if start.equal(end):
        end.forward_cursor_positions(interp.num)
    else:
        interp.num = 1
    imark = buf.create_mark(None, start, True)
    if buf.rectsel:
        marks = buf.rectsel.get_marks_by_line()
        for s, e in marks:
            _r_helper(buf, buf.get_iter_at_mark(s), buf.get_iter_at_mark(e),
                    interp.arg)
            buf.delete_mark(s)
            buf.delete_mark(e)
    else:
        _r_helper(buf, start, end, interp.arg)
    start = buf.get_iter_at_mark(imark)
    if interp.num > 1:
        start.forward_cursor_positions(interp.num - 1)
    buf.place_cursor(start)
    buf.delete_mark(imark)

@command(CHANGE, _("Cut character under cursor (delete forwards)"),
        ('x', '<Del>', '<kDel>'))
def x(interp):
    start, end = interp.get_working_range()
    interp.clip(start, end)
    # Cursor position gets deleted so force cursor to be replaced
    interp.buffer.place_cursor(interp.buffer.get_iter_at_insert())

@command(CHANGE, _("Cut character before cursor (delete backwards)"))
def X(interp):
    start, end = interp.get_working_range(False, True)
    interp.clip(start, end)

@command(COMPOUND | NEEDSRANGE | REGMOD,
        _("Delete (cut) range and start inserting new text"))
def c(interp, cmd = 'c'):
    buf = interp.buffer
    # Need to cache selmode because get_working_range() may change it
    selmode = buf.selmode
    upper = ascii.isupper(cmd)
    if selmode == mode.SELBYLINE and upper:
        cmd = cmd.lower()
        upper = False
    l = cmd.lower()
    ins = (l == 'c') or (l == 's')
    inc_lf = cmd == 'S' or cmd == 'C'
    if upper and cmd != 'S' and selmode == mode.SELRECT:
        buf.rectsel.extend_to_eol()
    if upper:
        bounds = interp.get_working_range_lines(inc_lf, cmd == 'S')
    else:
        bounds = interp.get_working_range( \
                cmd == 'c' or selmode == mode.SELBYLINE)
    if buf.rectsel:
        # vim treats rectsel same as plain seln with S
        if cmd == 'S':
            buf.rectsel.cancel()
        elif ins:
            buf.rectsel_marks = buf.rectsel.get_marks_by_line()
    elif not upper and cmd != 's' and len(bounds) < 2:
        return
    start = bounds[0]
    end = bounds[1]
    if upper and start.equal(end):
        end.forward_to_line_end()
        if inc_lf:
            end.forward_char()
    m = buf.create_mark(None, start)
    if ins:
        indentation = None
        s = interp.get_settings()
        if s.get_setting("autoindent") and \
                (cmd == 'S' or selmode == mode.SELBYLINE or interp.doubled):
            ts = s.get_setting('tabstop')
            et = s.get_setting('expandtabs')
            text = interp.buffer.get_text(start, end)
            w = get_leading_space(text, ts)[1]
            if w:
                indentation = make_leading_space(w, ts, et)
    interp.clip(start, end, l != 'y')
    if buf.rectsel_marks:
        buf.place_cursor(buf.get_iter_at_mark(buf.rectsel_marks[0][1]))
    if ins:
        if cmd =='S' or (cmd == 'C' and selmode != mode.SELRECT) \
                or selmode == mode.SELBYLINE or interp.doubled:
            interp.insert_nl_at_cursor(True)
            if indentation:
                buf.insert_at_cursor(indentation)
        interp.begin_insert(cmd)
    else:
        buf.place_cursor(buf.get_iter_at_mark(m))
    buf.delete_mark(m)

@command(COMPOUND | REGMOD,
        _("Delete (cut) to end of line and start inserting new text"))
def C(interp):
    c(interp, 'C')

@command(COMPOUND | REGMOD,
        _("Delete (cut) character under cursor and begin inserting text"))
def s(interp):
    c(interp, 's')

@command(COMPOUND | REGMOD,
        _("Delete (cut) current line and begin inserting text"))
def S(interp):
    c(interp, 'S')

@command(CHANGE | NEEDSRANGE | REGMOD, _("Delete (cut) range of characters"))
def d(interp):
    c(interp, 'd')

@command(CHANGE | REGMOD, _("Delete (cut) to end of line"))
def D(interp):
    c(interp, 'D')

@command(NEEDSRANGE | REGMOD, _("Yank range of characters (copy to clipboard)"))
def y(interp):
    c(interp, 'y')

@command(REGMOD, _("Yank rest of line (copy to clipboard)"))
def Y(interp):
    c(interp, 'Y')

@command(CHANGE | REGMOD | CLEARSEL,
    _("Insert contents of clipboard before cursor (paste)"))
def P(interp):
    interp.paste_clipboard(True)

@command(CHANGE | REGMOD | CLEARSEL,
    _("Append contents of clipboard after cursor (paste)"))
def p(interp):
    interp.paste_clipboard(False)

@command(CHANGE | NEEDSRANGE, _("Indent range by one level"), '>')
def indent(interp):
    return interp.in_or_out_dent(True)

@command(CHANGE | NEEDSRANGE, _("Remove one level of indentation from range"),
        '<')
def outdent(interp):
    return interp.in_or_out_dent(False)

@command(TAIL, _("Repeat last change at current position"),
        ('.', '<kPoint>'))
def repeat(interp):
    if change.last:
        num = interp.num
        interp.num = 1
        for n in range(num):
            interp.execute_change(change.last)
    else:
        interp.invalid()

@command(CHANGE, _("Swap case of character under cursor"), '~')
def swapcase(interp):
    interp.change_case("swapcase")

@command(CHANGE, _("Convert character under cursor to upper case"))
def U(interp):
    interp.change_case("upper")

@command(CHANGE,
    _("Convert selection to lower case or undo last change if no selection"))
def u(interp):
    if interp.change or interp.buffer.get_has_selection():
        interp.change_case("lower")
    elif interp.buffer.can_undo():
        interp.buffer.undo()
        interp.buffer.clear_selection()
        interp.view.scroll_cursor_on_screen()
    else:
        interp.invalid()

@command(COMMAND, _("Redo (reverse undo)"), '<C-R>')
def redo(interp):
    if interp.buffer.can_redo():
        interp.buffer.redo()
        interp.buffer.clear_selection()
        interp.view.scroll_cursor_on_screen()
    else:
        interp.invalid()

@command(CHANGE | NEEDSRANGE, _("Format range by wrapping long lines"))
def gq(interp):
    buf = interp.buffer
    if buf.rectsel:
        bounds = buf.rectsel.get_bounds()
    else:
        bounds = buf.get_selection_bounds()
    if len(bounds) != 2:
        msgs.warning(_("Attempting to format with no selection"))
        interp.invalid()
        return
    start, end = bounds
    # If double_key was used, format a whole paragraph, not just a line
    if interp.doubled:
        # Is selection a blank line?
        if buf.get_text(start, end) == '\n':
            # Do nothing if only want to format one paragraph, otherwise
            # format following paragraphs
            if interp.num == 1:
                interp.invalid()
                return
            else:
                start.forward_visible_line()
        # Now somewhere in first paragraph to be formatted, set bounds
        # at start of this para and end of interp.num paras later
        end = start.copy()
        vixiter.move_iter_by_paras(start, -1)
        vixiter.move_iter_by_paras(end, interp.num)
    # Skip any blank lines at start of selection
    if start.ends_line():
        start.forward_visible_line()
        start.set_visible_line_offset(0)
        if start.compare(end) >= 0:
            interp.invalid()
            return
    text = buf.get_text(start, end)
    tw = buf.get_textwidth(interp.view)
    text = format.format(text, tw, buf,
            interp.get_settings().get_setting('ts'))
    buf.delete(start, end)
    buf.insert(start, text)

@command(CHANGE, _("Format current paragraph by wrapping long lines"))
def gp(interp):
    interp.select_line()
    interp.doubled = True
    gq(interp)

@command(COMMAND, _("Enlarge text"), ('<C-+>', '<C-kPlus>'))
def zoom_in(interp):
    interp.window.zoom_in()

@command(COMMAND, _("Reduce text size"), ('<C-->', '<C-kMinus>'))
def zoom_out(interp):
    interp.window.zoom_out()

@command(COMMAND, _("Use default text size"), ('<C-0>', '<C-k0>'))
def zoom100(interp):
    interp.window.zoom100()

@command(NEEDSARG, _("Start recording keystrokes for playback with the @ " \
        "command. Press q again when not in insert mode to end recording"))
def q(interp):
    interp.view.start_recording(interp.arg)

@command(NEEDSARG, _("Playback a sequence of keystrokes recorded with the q " \
        "command"), '@')
def playback(interp):
    rec = qrec.recordings.get(interp.arg)
    if rec:
        for n in range(interp.num):
            rec.playback(interp.view)
    else:
        interp.window.status_bar.show_message(_("No recording for key '%s'") % \
                interp.arg)

""" End of built-in vi commands. """

def visym_to_disp(s):
    """ Converts a list of keystrokes with <> notation into something
    suitable for showing in a status bar etc eg <C-N> is converted to ^N,
    <A-N> to ~N. """
    result = ""
    for c in s:
        if c[0] == '<' and c[-1] == '>':
            s2 = c[1:-1]
            if len(s2) > 2 and s2[1] == '-':
                if s2[0] == 'C':
                    result += '^' + s2[2]
                    s2 = s2[2:]
                elif s2[0] == 'A':
                    result += '~' + s2[2]
                    s2 = s2[2:]
                elif s2[0] == 'M':
                    result += '~' + s2[2]
                    s2 = s2[2:]
            if s2 == 'kPoint':
                result += '.'
            elif s2 == 'kPlus':
                result += '+'
            elif s2 == 'kMinus':
                result += '-'
            elif s2 == 'kDivide':
                result += '/'
            elif s2 == 'kMultiply':
                result += '*'
            elif s2[0] == 'k':
                result += s2[1:]
            else:
                # Not really recognised
                result += c
        else:
            if c[0] == '\\':
                c = c[1:]
            result += c
    return result



class ViInterpreter(object):
    """ There are two main states: top-level commands, and movement commands
    which are slaves to top-level commands that take a range. To deal with this
    we use a slave (movement) - owner (range-taking command) relationship.
    """
    def __init__(self, keydict, view, owner = None, double_key = None):
        """ double_key is for slaves. If this key is pressed, it counts as
        selecting the whole line. """
        self.view = view
        if owner:
            # Avoid resetting status bar if we're a slave
            self.window = None
        else:
            self.window = view.get_vixwin()
        self.buffer = self.view.get_buffer()
        self.keydict = keydict
        self.invalid = self.reset
        # Initial reset mustn't also reset owner or its status bar
        self.owner = None
        self.saved_cursor_pos = None
        self.reset()
        self.owner = owner
        if owner:
            self.window = view.get_vixwin()
            self.start_iter = self.buffer.get_iter_at_insert()
            self.double_key = double_key

    def get_settings(self):
        # Doesn't matter whether we use buf, view or window; buf may
        # cause less function nesting when using set() with local = False
        return self.buffer

    def set_buffer(self, buf):
        self.buffer = buf

    def set_window(self, win):
        self.window = win

    def get_selection_mode(self):
        return self.buffer.selmode

    def reset(self):
        self.slave = None
        self.state = 0
        self.arg = None
        self.digits = ""
        # chain excludes digits, registers etc. all_chain includes everything
        self.chain = []
        self.all_chain = []
        self.num = 1
        self.reg = None
        self.double_key = None
        self.doubled = False    # Whether child of this obj got double_key
        if self.owner:
            self.owner.reset()
            self.owner = None
        elif self.window:
            self.window.status_bar.set_key_string(None)
        # self.change is set to a change object when running a repeat ('.')
        self.change = None
        self.clear_saved_cursor_pos()

    def get_status_string(self):
        if self.owner:
            s = self.owner.get_status_string()
        else:
            s = ""
        s += visym_to_disp(self.all_chain)
        return s
        #s += self.digits
        #if self.reg:
        #    s += '"' + self.reg
        #s += visym_to_disp(self.chain)
        #if self.arg:
        #    s += self.arg
        #return s

    def update_status_bar(self):
        if self.window:
            self.window.status_bar.set_key_string(self.get_status_string())

    def command_changed(self):
        """ Called when a keystroke has changed the command. """
        s = keysyms.serialise(self.chain)
        desc = self.keydict.get(s)
        if not desc:
            self.invalid()
            return
        if desc.flags & SEQUENCE:
            self.state |= SEQUENCE
            self.update_status_bar()
            return
        else:
            self.state &= ~SEQUENCE
        if desc.flags & NEEDSARG:
            self.state |= NEEDSARG
        self.update_status_bar()
        if desc.flags & NEEDSRANGE:
            self.get_range(self.chain[-1])
        if not (desc.flags & INCOMPLETE):
            self.execute()
            return

    def clear_saved_cursor_pos(self):
        if self.saved_cursor_pos:
            self.buffer.delete_mark(self.saved_cursor_pos)
            self.saved_cursor_pos = None

    def save_cursor_pos(self):
        self.clear_saved_cursor_pos()
        buf = self.buffer
        i = buf.get_iter_at_insert()
        self.saved_cursor_pos = buf.create_mark(None, i)

    def restore_cursor_pos(self):
        if self.saved_cursor_pos:
            buf = self.buffer
            i = buf.get_iter_at_mark(self.saved_cursor_pos)
            buf.place_cursor(i)
            self.view.scroll_cursor_on_screen()
            self.clear_saved_cursor_pos()

    def get_range(self, double_key = None):
        if self.buffer.get_has_selection():
            self.execute()
            return
        self.buffer.explicit_selection = None
        self.save_cursor_pos()
        self.state |= NEEDSRANGE
        #self.update_status_bar()   # shouldn't be necessary
        self.slave = ViInterpreter(self.keydict, self.view, self, double_key)

    def feed_key(self, k, estring = None):
        if k == '<Esc>':
            self.view.clear_selection()
            self.reset()
            return
        enc_ctrl = False
        enc_alt = False
        if not self.change:
            self.all_chain.append(k)
        if self.slave:
            self.slave.feed_key(k)
            return
        #self.clear_saved_cursor_pos()
        if self.double_key and k == self.double_key:
            self.owner.doubled = True
            self.owner.execute()
            self.reset()
            return
        else:
            self.double_key = None
        if self.state & SEQUENCE:
            self.chain.append(k)
            self.command_changed()
            return
        if self.state & NEEDSARG:
            if k[0] == '\\':
                k = k[1:]
            self.arg = k
            self.execute()
            return
        k_desc = self.keydict.get(k)
        if not k_desc:
            if estring:
                self.invalid()
            return
        if (k_desc.flags & DIGIT) or \
                ((k_desc.flags & ZERO) and self.digits and not self.chain):
            if k[:2] == '<k':
                k = k[2]
            self.digits += k
            self.update_status_bar()
            return
        self.chain = [k]
        if self.owner and not (k_desc.flags & MOVEMENT):
            self.invalid()
            return
        if self.get_selection_mode() and k_desc.flags & CLEARSEL:
            self.view.clear_selection()
        self.command_changed()

    def execute(self, k = None, desc = None):
        self.view.active_for_buffer(True)
        buf = self.buffer
        if self.digits:
            self.num = int(self.digits)
        else:
            self.num = 1
        if self.owner and self.owner.digits:
            self.num *= int(self.owner.digits)
        if k:
            k = keysyms.split_command(k)
        else:
            k = self.chain
        if not desc:
            desc = self.keydict.get(keysyms.serialise(k))
        if self.reg != None and (not desc or not (desc.flags & REGMOD)):
            self.reg = None
        oldreg = self.reg
        self.update_status_bar()
        if desc:
            method = desc.method
        else:
            method = None
        made_change = False
        if self.doubled:
            self.select_line()
        if method:
            # 'u' has dual role as undo or lower case, so its CHANGE flag
            # doesn't apply if there's no selection
            if (desc and (desc.flags & CHANGE)) and \
                    (method != u or buf.get_has_selection()):
                made_change = True
                if not self.change:
                    # If we used a slave to get range the selection would
                    # have been made by the slave's movement, so ignore it
                    if self.slave:
                        selmode = 0
                        selsize = 0
                    else:
                        selmode = buf.selmode
                        selsize = buf.get_selection_size()
                    buf.change = change.Change(buf, self.all_chain,
                            selmode, selsize)
            if made_change:
                buf.explicit_selection = None
                buf.begin_user_action()
            method(self)
            if desc.flags & TAIL:
                return
            if made_change and not (desc.flags & INSERT):
                buf.change_complete()
                # Selection should always be cleared after a change
                self.view.set_default_mode()
        if self.owner:
            i = buf.get_iter_at_insert()
            buf.select_range(i, self.start_iter)
            self.owner.digits = ""
            self.owner.execute()
        if not self.reg or oldreg:
            self.reset()

    def execute_change(self, chng):
        if self.view:
            self.view.active_for_buffer(True)
        buf = self.buffer
        self.change = chng
        buf.change = chng
        chng.buffer = buf
        buf.ignore_mark_set = True
        if chng.seltype:
            if buf.selmode:
                chng.seltype = buf.selmode
                chng.selsize = buf.get_selection_size()
            elif chng.seltype == mode.SELRECT:
                start = buf.get_iter_at_insert()
                left = start.get_line_offset()
                top = start.get_line()
                buf.rectsel = mod_rectsel.RectSel(buf, [left, top,
                        left + chng.selsize[0] - 1,
                        top + chng.selsize[1] - 1])
            else:
                start = buf.get_iter_at_insert()
                end = start.copy()
                if chng.seltype == mode.SELBYLINE:
                    end.forward_lines(chng.selsize)
                    vixiter.extend_iters_to_whole_lines(start, end)
                else:
                    end.forward_cursor_positions(chng.selsize)
                buf.select_range(start, end)
        elif buf.get_has_selection():
            buf.clear_selection()
        if chng.keychain == None:
            # This is caused by the user moving the cursor elsewhere during
            # an insert, logically starting a new insert. We need to
            # explicitly call begin_user_action()
            buf.begin_user_action()
            self.begin_insert()
        else:
            for k in chng.keychain:
                self.feed_key(k)
        self.view.set_default_mode()
        buf.ignore_mark_set = False

    """ The rest are mainly helper functions for the built-in commands. """

    def select_line(self):
        start = self.buffer.get_iter_at_insert()
        end = start.copy()
        start.set_line_offset(0)
        vixiter.iter_to_eol(end)
        self.view.set_selmode(mode.SELBYLINE)
        self.buffer.select_range(start, end)

    def begin_insert(self, typ = 'i'):
        self.clear_saved_cursor_pos()
        if self.change:
            self.change.repeat_deletions()
            ins = self.change.inserted
            if ins:
                self.buffer.insert_at_cursor(ins)
            self.buffer.change_complete()
            self.reset()
        else:
            if not self.buffer.change:
                msgs.warning(
                        _("Beginning insert operation without change recorder"))
            else:
                self.buffer.change.prepare_for_insert()
            if typ == 'R':
                self.view.set_mode(mode.OVERWRITE)
            else:
                self.view.set_mode(mode.INSERT)

    def begin_append(self, typ = 'a'):
        iter = self.buffer.get_iter_at_insert()
        if not iter.ends_line():
            iter.forward_cursor_position()
        self.view.move_cursor_to_iter(iter, True)
        self.begin_insert(typ)

    def insert_nl_and_indentation(self, iter, indent = False, after = False):
        if indent and self.get_settings().get_setting("autoindent"):
            start = iter.copy()
            end = iter.copy()
            if not start.starts_line():
                start.set_line_offset(0)
            if not end.ends_line():
                end.forward_to_line_end()
            text = self.buffer.get_text(start, end)
            n = 0
            for n in range(len(text)):
                if text[n] != ' ' and text[n] != '\t':
                    break
            else:
                # If we reach end, n is len(text) - 1
                n += 1
            text = text[:n]
        else:
            text = None
        self.buffer.insert(iter, '\n')
        if after:
            iter.backward_char()
        self.buffer.place_cursor(iter)
        if text:
            self.buffer.insert_at_cursor(text)

    def insert_nl_at_cursor(self, after = False):
        self.insert_nl_and_indentation(self.buffer.get_iter_at_insert(),
                False, after)

    @staticmethod
    def iter_find_char(iter, end, char, stepper = None):
        if not stepper:
            stepper = iter.forward_cursor_position
        stepper()
        while not(iter.equal(end)):
            if iter.get_char() == char:
                return True
            stepper()
        return False

    def forward_find_char(self):
        iter = self.buffer.get_iter_at_insert()
        if iter.ends_line():
            return iter
        orig = iter.copy()
        end = iter.copy()
        end.forward_to_line_end()
        n = self.num
        while n:
            if not self.iter_find_char(iter, end, self.arg):
                return orig
            n -= 1
        return iter

    def backward_find_char(self):
        iter = self.buffer.get_iter_at_insert()
        if iter.starts_line():
            return iter
        orig = iter.copy()
        end = iter.copy()
        end.set_line_offset(0)
        n = self.num
        while n:
            if not self.iter_find_char(iter, end, self.arg,
                    iter.backward_cursor_position):
                return orig
            n -= 1
        return iter

    def one_more(self, iter):
        """ If getting range for a NEEDSRANGE, movements need to move on
        one more place to ensure character under cursor is included. """
        if self.owner:
            iter.forward_cursor_position()

    def get_working_range_lines(self, inc_lf = False,
            start_starts_line = True, backward = False):
        start, end = self.buffer.get_working_range(False)
        #if not start.equal(end):
        #   self.num = 1
        if start_starts_line:
            start.set_line_offset(0)
        if backward:
            if self.num > 1:
                start.backward_lines(self.num - 1)
        else:
            if self.num > 1:
                end.forward_lines(self.num - 1)
            if inc_lf:
                vixiter.ensure_iter_eol(end)
            elif end.starts_line():
                if end.get_line() > start.get_line():
                    # This is a bit kludgy, but in every practical case when we
                    # get here we've selected one beyond what we wanted
                    end.backward_char()
            else:
                end.forward_to_line_end()
        return [start, end]

    def get_working_range(self, inc_lf = False, backward = False):
        """ Returns extended selection bounds: if selection mode is SELBYLINE
        that number of lines is multiplied by self.num, otherwise the
        number of characters in the selection is multiplied. If inc_lf is True
        and the selection is whole lines, the end iter is moved past the final
        LF. """
        if self.buffer.selmode == mode.SELBYLINE:
            return self.get_working_range_lines(inc_lf, True, backward)
        start, end = self.buffer.get_working_range(False)
        if backward:
            end.backward_cursor_positions(self.num)
        elif start.equal(end):
            end.forward_cursor_positions(self.num)
        elif self.num > 1:
            end.forward_cursor_positions(self.num - 1)
        return [start, end]

    def cut_rect(self):
        buf = self.buffer
        for start, end in buf.rectsel.get_marks_by_line():
            s = buf.get_iter_at_mark(start)
            e = buf.get_iter_at_mark(end)
            buf.delete(s, e)
            buf.delete_mark(start)
            buf.delete_mark(end)

    def clip(self, start, end, cut = True):
        buf = self.buffer
        if buf.rectsel:
            selmode = mode.SELRECT
            text = buf.rectsel.get_text()
        else:
            if buf.selmode == mode.SELBYLINE or self.doubled:
                selmode = mode.SELBYLINE
            else:
                selmode = mode.SELECT
            text = buf.get_text(start, end)
        clipboard.set_text(self.reg, text, selmode)
        # Would select_range() here and use cut_/copy_clipboard but for
        # some reason selection immediately clears itself if cursor is
        # at start of line
        if selmode == mode.SELBYLINE:
            amount = text.count('\n')
            if amount == 1:
                amount = _("1 line")
            else:
                amount = _("%d lines")  % amount
        else:
            amount = len(text)
            if amount == 1:
                amount = _("1 character")
            else:
                amount = _("%d characters") % amount
        if cut:
            if buf.rectsel:
                self.cut_rect()
            else:
                buf.delete(start, end)
            msg = _("%s deleted") % amount
        else:
            msg = _("%s yanked") % amount
        buf.clear_selection()
        self.restore_cursor_pos()
        buf.ensure_trailing_nl()
        if self.view:
            w = self.view.get_vixwin()
            if w:
                #self.view.skip_cursor_pos_update = True
                w.status_bar.show_message(msg)

    def paste_rect(self, iter, text):
        buf = self.buffer
        lines = text.split('\n')
        ln = iter.get_line()
        col = iter.get_line_offset()
        lc = buf.get_line_count()
        for l in lines:
            if ln < lc:
                i = buf.get_checked_iter(col, ln)
            else:
                i = buf.get_end_iter()
                l = '\n' + l
            buf.insert(i, l)
            ln += 1

    def paste_clipboard(self, before = False):
        buf = self.buffer
        iter = buf.get_iter_at_insert()
        try:
            text, selmode = clipboard.get_text_and_selmode(self.reg)
        except KeyError:
            self.window.status_bar.show_message(_("Register %s not set") % \
                    self.reg)
            return
        if selmode == mode.SELBYLINE:
            if not before:
                iter.forward_line()
            iter.set_line_offset(0)
        elif not before and not iter.ends_line():
            iter.forward_cursor_position()
        mk = buf.create_mark(None, iter, selmode == mode.SELBYLINE)
        # Not called during insert operation so this insert doesn't affect
        # buf.change. Instead, a repeat will cause content of
        # clipboard at that point in time to be pasted, even if it's
        # changed. This behaviour is consistent with vim.
        if text:
            if selmode == mode.SELRECT:
                self.paste_rect(iter, text)
            else:
                buf.insert(iter, text)
        i = buf.get_iter_at_mark(mk)
        if selmode == mode.SELBYLINE:
            self.view.move_iter_to_hug_column(i)
        buf.place_cursor(i)
        buf.delete_mark(mk)

    def in_or_out_dent(self, indent):
        buf = self.buffer
        buf.extend_selection_to_whole_lines()
        [start, end] = self.get_working_range()
        saved_start = buf.create_mark(None, start)
        s = self.get_settings()
        sw = s.lookup_setting('shiftwidth', 4)
        ts = s.get_setting('ts')
        et = s.get_setting('expandtabs')
        end = buf.create_mark(None, end)
        while start.compare(buf.get_iter_at_mark(end)) < 0:
            # This works by replacing all leading whitespace with a new
            # set with shiftwidth removed, ensuring mixtures of tabs and
            # spaces are handled OK. If there is less than shiftwidth
            # leading whitespace, only the whitespace is removed, no "real"
            # characters.
            # Set marks at start and end of line and get text in between
            eol = start.copy()
            eol.forward_to_line_end()
            start_mark = buf.create_mark(None, start, True)
            otext = buf.get_text(start, eol)
            # Replace leading whitespace in text with amended version
            i, w = get_leading_space(otext, ts)
            if indent:
                w += sw
            else:
                w -= sw
                if w < 0:
                    w = 0
            if w:
                ntext = make_leading_space(w, ts, et) + otext[i:]
            else:
                ntext = otext[i:]
            # Delete old version of line from buf and replace with new one
            buf.delete(start, eol)
            buf.insert(buf.get_iter_at_mark(start_mark), ntext)
            # Set start iter to start of next line
            start = buf.get_iter_at_mark(start_mark)
            buf.delete_mark(start_mark)
            start.forward_line()
            start.set_line_offset(0)
        buf.delete_mark(end)
        buf.clear_selection()
        i = buf.get_iter_at_mark(saved_start)
        vixiter.iter_to_first_non_space(i)
        buf.place_cursor(i)
        self.view.scroll_cursor_on_screen()
        buf.delete_mark(saved_start)
        #buf.set_cursor_type()

    def change_case(self, method, start = None, end = None):
        buf = self.buffer
        if not start:
            if buf.rectsel:
                self.__change_case_rect(method)
                return
            start, end = self.get_working_range()
        txt = buf.get_text(start, end)
        txt = getattr(txt, method)()
        buf.delete(start, end)
        buf.insert(start, txt)

    def __change_case_rect(self, method):
        buf = self.buffer
        for start, end in buf.rectsel.get_marks_by_line():
            self.change_case(method, buf.get_iter_at_mark(start),
                buf.get_iter_at_mark(end))
            buf.delete_mark(start)
            buf.delete_mark(end)


""" Other helper functions """

def get_leading_space(text, ts):
    # Returns (i, w) where i is number of leading space characters,
    # w is their width with tabs expanded to width ts
    l = len(text)
    i = 0   # string index
    w = 0   # width of tabs/spaces
    while i < l and (text[i] == ' ' or text[i] == '\t'):
        if text[i] == '\t':
            w += ts - w % ts
        else:
            w += 1
        i += 1
    return (i, w)

def make_leading_space(w, ts, et):
    # Returns a string of spaces w wide, consisting of tabs of width ts
    # if et is True, padded with trailing spaces if necessary; uses
    # all spaces if et is False
    if et:
        return ' ' * w
    else:
        return ('\t' * int(w / ts)) + ' ' * (w % ts)

