import bz2
import gzip
import os
import re
import sys
import traceback
from curses import ascii

from gi.repository import GLib
from gi.repository import Gdk, Gtk
from gi.repository import GtkSource

from . import buffers
from . import dialog
from .excmd import ExInterpreter
from . import hooks
from . import langs
from . import make
from . import mode
from . import misc
from . import msgs
from . import rectsel
from . import scheme
from . import search
from . import settings
from .state import state
from . import throwback
from . import vixi18n
from . import vixdbus
from . import vixexcept
from . import vixiter
from . import vixwin

# Set of Unicode characters we recognise as bullets
_BULLETS = '[*\u2022\u2023\u2043\u2219\u25e6]'

default_mode = mode.VI


_make_buffer_name_index = 0


class SearchBufExtn(throwback.BufExtn):
    def __init__(self, buf, targ_buf):
        self.targ_buf = targ_buf
        self.re = re.compile("^([0-9]+):([0-9]+):.*")
        throwback.BufExtn.__init__(self, buf)

    def parse_line(self, line, *args, **kwargs):
        m = self.re.match(line)
        if m:
            return [self.targ_buf.filename, int(m.group(1)), int(m.group(2)),
                    m.start(1), m.end(1), m.start(2), m.end(2), "find"]
        return None


class VixnBuffer(GtkSource.Buffer, settings.SettingsClient):
    """ vixn's version of SourceBuffer. """

    make_buf = None

    def __init__(self, settings_client, cwd = None,
            tag_table = None, filename = None):
        self.destroyed = False
        self.user_action_count = 0
        self.ignore_user_action = False
        self.debug_user_action = False
        self.always_debug_user_action = False
        self.excmd = None
        self.editable = True
        self.is_buffer = True
        self.active_view = None
        self.fgparents = []
        self.views = []
        self.throwbacks = []
        self.__debug_tb = []
        self.extns = {}
        self.selmode = 0
        self.fmt_bul_re = None
        self.email_ft = False
        self.make = None
        self.state = None
        self.ignore_vi_marks = False
        self.__mark_set_pending = False
        # If selection is started explicitly explicit_selection is
        # an iter where it started
        self.explicit_selection = None
        # change here is a new change being constructed by a vicmd operation
        # eg an insert
        self.change = None
        self.wchars = '_'
        self.rectsel = None
        self.rectsel_marks = None
        self.pristine = True

        self.crlf = False   # CRLF line endings

        self.settings_client = \
                settings_client.copy_settings(settings.PRIORITY_DEFAULT)
        if cwd:
            self.settings_client.set_setting('cwd', cwd)
        # rc file (settings.defaults) overrides inherited settings
        self.settings_client.merge(settings.defaults)
        # If settings_client was an existing buf give copy of settings
        # low priority, otherwise they're defaults
        if settings_client.get_buffer():
            priority = settings.PRIORITY_DEFAULT
        else:
            priority = None
        self.settings = \
                self.settings_client.get_vixbuf_settings().copy(priority)

        if tag_table:
            GtkSource.Buffer.__init__(self, tag_table)
        else:
            GtkSource.Buffer.__init__(self)

        self.search = search.Search(self)

        # buf starts off without a filename, but a suitable "Untitled..."
        # leafname is generated for it as orig_filename. orig_filename is
        # updated when a file is loaded or buf saved. If you try to save
        # over an existing file that doesn't match orig_filename you get a
        # warning. Have to do this even if filename is given here, in case
        # load fails.
        global _filename_index
        self.orig_filename, _filename_index = \
                buffers.generate_filename(self.get_cwd())
        self.filename = None
        if filename:
            self.load(filename, True)
        else:
            self.load_state(self.orig_filename)
            self.set_setting('filetype', 'text', settings.PRIORITY_DEFAULT)
            self.set_text('\n')
            self.place_cursor(self.get_start_iter())
            self.set_modified(False)

        buffers.buffers.append(self)

        self.ignore_mark_set = False
        self.connect("mark-set", self.__mark_set)

        self.connect("changed", self.__changed_handler)
        self.connect("modified-changed", self.__modified_changed_handler)

        # Look up explicitly because we want a different default from
        # gtksourceview's
        self.set_highlight_syntax(self.lookup_setting("syntax", True, True))

        settings.apply_all(self, self.settings)

    def empty_clone(self):
        return VixnBuffer(self)

    def get_buffer(self):
        return self

    def get_view(self):
        if self.active_view:
            return self.active_view
        for v in self.views:
            w = v.get_vixwin()
            if w and w.get_property("has-focus"):
                self.active_view = v
                break
        else:
            if len(self.views):
                self.active_view = self.views[-1]
        return self.active_view

    def get_vixwin(self):
        v = self.get_view()
        if v:
            return v.get_vixwin()
        else:
            return None

    def get_views(self):
        return self.views

    def get_windows(self):
        wins = []
        for v in self.views:
            wins.append(v.get_vixwin())
        return wins

    def get_vixbuf_settings(self):
        return self.settings

    def get_vixview_settings(self):
        v = self.get_view()
        if v:
            return v.get_vixview_settings()
        else:
            return self.settings_client.get_vixview_settings()

    def get_vixwin_settings(self):
        w = self.get_vixwin()
        if w:
            return w.get_vixwin_settings()
        else:
            return self.settings_client.get_vixwin_settings()

    def get_cwd(self):
        return self.lookup_setting("cwd", "")

    def __exception(self, message):
        dialog.exception(self.get_vixwin(), message)

    def debug_throwback(self, code, msg):
        try:
            if code == -1:
                msg = msg % self.get_default_filename()
            else:
                msg = msg % (code, self.get_default_filename())
        except:
            msgs.exception("Error formatting debug_throwback message '%s' " \
                    "with code %d" % (msg, code))
        else:
            self.__debug_tb.append(msg)

    def detach_throwback(self, tfs):
        self.debug_throwback(tfs.unique_id,
                "Removing throwback %d from buf %s")
        tfs.detach_from_buf()
        try:
            self.throwbacks.remove(tfs)
        except:
            for tbm in self.__debug_tb:
                sys.stderr.write(tbm + '\n')
            self.__exception(_("Detaching throwback set from buf it " \
                    "isn't attached to"))
            self.__debug_tb.append("*************")

    def add_fgparent(self, code):
        self.fgparents.append(code)
        vixdbus.add_buffer(code)

    def set_highlight_syntax(self, syn = True):
        if syn == 'on' or syn == 1:
            syn = True
        elif syn == 'off':
            syn = False
        GtkSource.Buffer.set_highlight_syntax(self, syn)

    def __insert_mode(self):
        for v in self.views:
            if v.mode & mode.INSERT_MASK:
                return True
        return False

    def __show_block_cursor(self, ct = None, ins = None):
        if mode.sel_but_not_rect(self.selmode):
            return False
        if ct == None:
            ct = self.lookup_setting('blockcursor', 0, True)
        if ct == 1 and ins or (ins == None and self.__insert_mode()):
            ct = 0
        return ct != 0

    def __cursor_at_eol(self, iter = None):
        if not iter:
            iter = self.get_iter_at_insert()
        return iter.ends_line()

    def __show_block_cursor_eol(self, ct = None, iter = None):
        if not self.__show_block_cursor(ct):
            return False
        return not self.__cursor_at_eol(iter)

    def lookup_cursor_bg_color(self, applying = False):
        return self.lookup_setting('cbg', '#cd0000', applying)

    def lookup_cursor_fg_color(self, applying = False):
        return self.lookup_setting('cfg', '#ffff00', applying)

    def update_caret_color(self, ins = None):
        # Whether to use block cursor colour for caret
        if ins == None:
            ins = self.__insert_mode()
        if not ins:
            cbg = self.lookup_cursor_bg_color(True)
            cbgg = Gdk.color_parse(cbg)
        for v in self.views:
            if not ins:
                color = cbgg
                v.cursor_color = color
            else:
                v.cursor_color = None
                color = v.get_caret_gdk_color()
            v.modify_cursor(color, color)

    def set_selmode(self, mo):
        self.selmode = mo & mode.SELECT_MASK
        if not self.selmode:
            self.explicit_selection = None
        if self.selmode == mode.SELRECT and not self.rectsel:
            self.rectsel = rectsel.RectSel(self)
        elif self.selmode != mode.SELRECT and self.rectsel:
            self.rectsel.cancel()

    def set_mode(self, mo, update_views = True):
        self.set_selmode(mo)
        if update_views:
            for v in self.views:
                v.set_mode(mo, False)

    def set_default_mode(self, update_views = True):
        self.set_mode(default_mode, update_views)

    def __mark_set(self, buf, iter, mark):
        n = mark.get_name()
        if n != 'insert' and n != 'selection_bound':
            return
        has_sel = False
        if not self.ignore_mark_set:
            if n == 'insert' and self.explicit_selection:
                if mode.sel_but_not_rect(self.selmode):
                    # Make sure selection's starting point (explicit_selection)
                    # is always included in selection. We do this by moving
                    # selection_bound to explicit_selection and then one
                    # position further on if it's same as or after insert
                    # point.
                    sel_iter = self.get_iter_at_mark( \
                            self.get_mark('selection_bound'))
                    if sel_iter.compare(iter) < 0:
                        start = sel_iter
                        end = iter
                    else:
                        start = iter
                        end = sel_iter
                    before = False
                    after = False
                    if self.explicit_selection.compare(start) < 0:
                        before = True
                    elif self.explicit_selection.compare(end) >= 0:
                        after = True
                    if iter.equal(sel_iter):
                        # In this case we're swapping iter and sel_iter so also
                        # move iter by one step otherwise it will look as if
                        # nothing has happened
                        if before:
                            if self.selmode == mode.SELBYLINE:
                                iter.forward_line()
                            else:
                                iter.forward_char()
                        elif after:
                            if self.selmode == mode.SELBYLINE:
                                iter.backward_line()
                            else:
                                iter.backward_char()
                    if before or after:
                        if self.selmode == mode.SELBYLINE:
                            sel_iter.set_line( \
                                    self.explicit_selection.get_line())
                            if sel_iter.compare(iter) >= 0:
                                sel_iter.forward_line()
                        else:
                            sel_iter = self.explicit_selection.copy()
                            if sel_iter.compare(iter) >= 0:
                                sel_iter.forward_cursor_position()
                        self.ignore_mark_set = True
                        self.select_range(iter, sel_iter)
                        self.ignore_mark_set = False
                        return
                elif self.rectsel:
                    self.rectsel.move_to_iter(iter)
            if n == "selection_bound" or n == "insert":
                bounds = self.get_selection_bounds()
                has_sel = len(bounds) > 1
                if not self.ignore_mark_set:
                    if has_sel and not self.selmode:
                        self.set_mode(mode.SELECT)
                    elif not has_sel and mode.sel_but_not_rect(self.selmode):
                        self.set_default_mode()
                    if has_sel and buf.selmode == mode.SELBYLINE:
                        self.ignore_mark_set = True
                        self.extend_selection_to_whole_lines()
                        self.ignore_mark_set = False
                        # Above causes this function to be called again
                        return
            if n == 'insert' and not self.__mark_set_pending:
                self.__mark_set_pending = True
                GLib.idle_add(self.__mark_set_idle, None)
        if has_sel:
            self.set_vi_mark('<', bounds[0])
            self.set_vi_mark('>', bounds[1])
        if not self.ignore_vi_marks and self.state:
            if n == 'insert':
                self.state.pos = [iter.get_line_offset(), iter.get_line()]
            if n and n.startswith("vi_"):
                self.record_vi_mark(n, iter)

    def __mark_set_idle(self, data):
        self.__mark_set_pending = False
        iter = self.get_iter_at_insert()
        view = self.get_view()
        ts = view.get_tab_width()
        col = view.cursor_pos_update(iter, None, ts)
        for v in self.views:
            if v == view:
                continue
            ts2 = v.get_tab_width()
            if ts2 == ts:
                col2 = col
            else:
                col2 = None
            v.cursor_pos_update(iter, col2, ts2)
        return False

    def __changed_handler(self, buf):
        if self.get_view().wrap_on_change():
            # Wrap should cause a new change signal so continuing to process
            # this one would be redundant.
            return
        self.pristine = False
        if self.search:
            self.search.invalidate()
        # Typing text doesn't generate mark-set signals so force status update
        if not self.__mark_set_pending:
            self.__mark_set_pending = True
            GLib.idle_add(self.__mark_set_idle, None)

    def add_view(self, view):
        if view in self.views:
            msgs.error("buf %s already has view %s" % (self, view))
        else:
            self.views.append(view)
            self.clear_selection()
            self.update_view_titles()
            view.scroll_cursor_on_screen()
            view.cursor_pos_update(self.get_iter_at_insert(),
                    None, view.lookup_setting('ts', 8, False))
        view.set_editable(self.editable)

    def remove_view(self, view):
        try:
            del self.views[self.views.index(view)]
        except:
            msgs.error("remove_view: buf %s has no view %s" %
                    (self, view))
        else:
            if not len(self.views):
                self.destroy()
            else:
                if view == self.active_view:
                    self.active_view = None
                self.update_view_titles()

    def notify_parent_closing(self):
        for code in self.fgparents:
            vixdbus.remove_buffer(code)
        self.fgparents = []

    def __detach_throwbacks(self):
        self.debug_throwback(len(self.throwbacks),
                "__detach_throwbacks: Removing %d throwbacks from %s")
        if not self.throwbacks:
            return
        for t in self.throwbacks:
            self.debug_throwback(t.unique_id, "Will remove %d from %s")
        # Have to copy list because it gets altered as we go along
        for t in list(self.throwbacks):
            self.debug_throwback(t.unique_id,
                    "__detach_throwbacks: Now removing %d from %s")
            self.detach_throwback(t)
        self.throwbacks = []

    def close_buf(self):
        for v in self.views:
            v.destroy()

    def destroy(self):
        """ No use relying on __del__ because we need to remove from global
        buffers list and __del__ won't get called until after removal. """
        if self.destroyed:
            raise vixexcept.CodeError(
                    _("destroy called twice for same buffer: " \
                        "previous invocation:\n%s") % '\n'.join(self.destroyed))
        self.destroyed = traceback.format_stack()
        if VixnBuffer.make_buf == self:
            VixnBuffer.make_buf = None
        for e in list(self.extns.values()):
            e.destroy()
        if self.state:
            state.touch_file(self.state)
        if self.change:
            self.change_complete()
        self.debug_throwback(-1, "destroy('%s')")
        del self.search
        self.search = None
        self.notify_parent_closing()
        self.debug_throwback(-1,  "destroy calling detach_throwbacks for %s")
        self.__detach_throwbacks()
        try:
            del buffers.buffers[buffers.buffers.index(self)]
        except:
            msgs.error("buf %s not in global list" % self)
        buffers.check_quit()

    def get_default_filename(self):
        """ If the buf doesn't have a current filename, this returns
        orig_filename instead. """
        if self.filename:
            return self.filename
        else:
            return self.orig_filename

    def save_range(self, filename, start, end, check = 0):
        """ Possible values of check:
        0: Don't warn if file exists, and don't prompt for filename
        1: Warn if existing file doesn't match self.orig_filename
        2: Always warn if file exists
        Returns whether save succeeded. """
        if not os.path.isabs(filename):
            filename = os.path.join(self.get_cwd(), filename)
        if check and os.path.exists(filename) and \
                (filename != self.orig_filename or check == 2) and \
                not dialog.check_overwrite(self.get_vixwin(), filename):
            return False
        text = self.get_text(start, end, True)
        if text == None:
            return False
        compression = langs.compression_test(filename)[1]
        try:
            if compression == 'gz':
                fp = gzip.open(filename, 'wb')
            elif compression == 'bz2':
                fp = bz2.BZ2File(filename, 'w')
            else:
                fp = open(filename, 'w')
            fp.write(text)
            fp.close()
        except:
            self.__exception(_("Unable to save to '%s'") % str(filename))
            return False
        return True

    def save_as(self, filename, check = 0):
        """ See save_range for explanation of check. """
        if self.lookup_setting('trim', True):
            self.trim()
        filename = self.expandpath(filename)
        result = self.save_range(filename,
                self.get_start_iter(), self.get_end_iter(), check)
        if result:
            if self.state:
                self.state.filename = filename
                state.touch_file(self.state)
            else:
                self.state = state.State(filename)
            new_filename = self.filename != filename
            for t in self.throwbacks:
                if new_filename:
                    t.reset_target_pos()
                else:
                    t.confirm_target_pos()
            if new_filename:
                self.debug_throwback(-1,
                        "Removing throwbacks from %s for saving")
                self.__detach_throwbacks()
            self.filename = filename
            self.orig_filename = filename
            # If buffer had been modified modified-changed signal will cause
            # title update, otherwise it needs to be forced
            if not self.get_modified():
                self.update_view_titles()
            self.set_modified(False)
            if new_filename:
                self.debug_throwback(-1,
                        "Reregistering throwbacks for %s after saving")
                throwback.register_buffer(self)
        return result

    def __modified_changed_handler(self, buf):
        self.update_view_titles()

    def update_view_titles(self):
        if self.views:
            for v in self.views:
                v.update_title()

    def save(self, force = True):
        """ If force is True, it saves even if we might be overwriting existing
        file. """
        if force:
            if not self.get_modified():
                return
            check = 0
        else:
            check = 1
        filename = self.filename
        if not filename:
            filename = os.path.join(self.get_cwd(),
                    self.orig_filename)
        return self.save_as(filename, check)

    def __load_file(self, filename, whole_buffer = True):
        compression = langs.compression_test(filename)[1]
        try:
            if compression == 'gz':
                fp = gzip.open(filename, 'rt')
            elif compression == 'bz2':
                fp = bz2.open(filename, 'rt')
            else:
                fp = open(filename, 'rt')
            text = fp.read()
            fp.close()
            #fp = open("tmp", 'w')
            #fp.write(text)
            #fp.close()
        except:
            self.__exception(_("Unable to load '%s'") % str(filename))
            return None
        return text

    def __load_and_insert(self, filename, whole_buffer):
        text = self.__load_file(filename, whole_buffer)
        if not text:
            return text
        if whole_buffer:
            self.set_text(text)
            start_check = self.get_start_iter()
            end_check = self.get_end_iter()
        else:
            iter = self.get_iter_at_insert()
            iter.forward_line()
            iter.set_line_offset(0)
            s = self.create_mark(None, iter, True)
            e = self.create_mark(None, iter, False)
            self.insert(iter, text)
            start_check = self.get_iter_at_mark(s)
            end_check = self.get_iter_at_mark(e)
            self.delete_mark(s)
            self.delete_mark(e)
        if not text:
            return text
        if text[-1] != '\n':
            if whole_buffer:
                iter = self.get_end_iter()
            else:
                iter = self.get_iter_at_insert()
            self.insert(iter, '\n')
        return text

    def insert_file(self, filename):
        self.begin_user_action()
        filename = self.expandpath(filename)
        self.__load_and_insert(filename, False)
        self.end_user_action()

    def __del_vi_marks(self):
        if self.state:
            for n in list(self.state.marks.keys()):
                m = self.get_mark("vi_" + n)
                if m:
                    self.delete_mark(m)

    def load(self, filename, ignore_noexist = False):
        self.__del_vi_marks()
        for t in self.throwbacks:
            t.reset_target_pos()
        self.debug_throwback(-1, "Removing throwbacks from %s for loading")
        self.__detach_throwbacks()
        self.pristine = False
        self.begin_not_undoable_action()
        filename = self.expandpath(filename)
        if not ignore_noexist or os.path.exists(filename):
            text = self.__load_and_insert(filename, True)
        else:
            text = ""
            # set_text forces modified flag to ensure title bar is updated
            self.set_text("")
        if text != None or (ignore_noexist and not os.path.exists(filename)):
            self.__run_hooks(hooks.filename, filename)
            filetype = langs.guess_filetype(filename, text)
            self.set_setting('filetype', filetype,
                    settings.PRIORITY_AUTO_FILENAME)
            self.filename = filename
            self.orig_filename = filename
            self.set_modified(False)
        self.end_not_undoable_action()
        self.load_state(filename)
        self.debug_throwback(-1,
                "Reregistering throwbacks for %s after loading")
        throwback.register_buffer(self)

    def get_line_length(self, line_num):
        """ Including terminator; returns -1 if line invalid """
        if line_num >= self.get_line_count():
            return -1
        i = self.get_iter_at_line(line_num)
        if i.ends_line():
            return 0
        i.forward_to_line_end()
        return i.get_line_offset()

    def get_checked_iter(self, column, line):
        l = self.get_line_length(line)
        if l >= column:
            return self.get_iter_at_line_offset(line, column)
        elif l > 0:
            return self.get_iter_at_line_offset(line, l)
        elif l != -1:
            return self.get_iter_at_line(line)
        else:
            #msgs.warning(_("Bad position (%d, %d)") % (column, line))
            return None

    def load_state(self, filename):
        self.state = state.get_file(filename)
        self.ignore_vi_marks = True
        if self.state.pos:
            i = self.get_checked_iter(*self.state.pos)
            if not i:
                self.state.pos = [0, 0]
                i = self.get_iter_at_line(0)
            self.place_cursor(i)
            v = self.get_view()
            if v:
                v.scroll_cursor_on_screen()
                v.cursor_pos_update(i, None, v.lookup_setting('ts', 8, False))
        for n, m in list(self.state.marks.items()):
            i = self.get_checked_iter(*m)
            if i:
                self.set_vi_mark(n, i)
            else:
                del(self.state.marks[n])
        self.ignore_vi_marks = False

    def get_iter_at_insert(self):
        return self.get_iter_at_mark(self.get_insert())

    def extend_selection_to_whole_lines(self):
        if self.rectsel:
            ins_iter, sel_iter = self.rectsel.get_bounds()
        else:
            ins_iter = self.get_iter_at_mark(self.get_insert())
            sel_iter = self.get_iter_at_mark(self.get_selection_bound())
        vixiter.extend_iters_to_whole_lines(ins_iter, sel_iter)
        self.select_range(ins_iter, sel_iter)
        self.set_mode(mode.SELBYLINE)

    def start_search(self, pattern, backward = False, tback = False, num = 1):
        """ Starts a new /, ? or \ search. Returns next match after iter as
        [start, end] iters of match, or None if no match. """
        if search.get_ignore_case(self, pattern):
            flags = re.I
        else:
            flags = 0
        self.search.get_matches(pattern, backward, tback, flags)
        if not self.search.results:
            self.get_vixwin().status_bar.show_message( \
                    _("No match for search"))
            return
        iter = self.get_iter_at_insert()
        if backward:
            iter.backward_char()
            return self.search.search_backward(iter, num)
        elif tback:
            self.__open_search_throwback()
            return None
        else:
            iter.forward_char()
            return self.search.search_forward(iter, num)

    def __validate_search(self):
        if self.search.is_valid():
            return True
        elif search.last_search:
            self.search.get_matches(search.last_search[0],
                    search.last_search[2], False, search.last_search[1])
            return True
        else:
            return False

    def __search_forward(self, num):
        iter = self.get_iter_at_insert()
        if not iter.forward_char():
            iter = self.get_start_iter()
        return self.search.search_forward(iter, num)

    def __search_backward(self, num):
        iter = self.get_iter_at_insert()
        if not iter.backward_char():
            iter = self.get_end_iter()
        return self.search.search_backward(iter, num)

    def search_forward(self, num = 1):
        if not self.__validate_search():
            return None
        if search.last_search[2]:
            return self.__search_backward(num)
        else:
            return self.__search_forward(num)

    def search_backward(self, num = 1):
        if not self.__validate_search():
            return None
        if search.last_search[2]:
            return self.__search_forward(num)
        else:
            return self.__search_backward(num)

    def __open_search_throwback(self):
        global _make_buffer_name_index
        tbfilename, _make_buffer_name_index = \
                buffers.generate_filename(self.get_cwd(),
                        _("_Search_%d"), _make_buffer_name_index,
                        lambda x, y: x + y)
        tbbuf = VixnBuffer(self, self.get_cwd(), None, tbfilename)
        tb = SearchBufExtn(tbbuf, self)
        vixwin.VixnWindow(tbbuf, settings.PRIORITY_DEFAULT).present()
        throwback.add_view_extn_to_buf(tbbuf)
        tb.add_line(_("Search results for %s") % self.filename)
        for result in self.search.results:
            line = result[0].get_line()
            column = result[0].get_visible_line_offset()
            result[0].set_line_offset(0)
            end = result[1].copy()
            end.forward_to_line_end()
            if end.get_line() != line:
                end = result[1]
            text = self.get_text(result[0], end, False)
            tb.add_line("%d:%d:%s" % (line + 1, column + 1, text))
        tbbuf.set_modified(False)

    def get_working_range(self, forward = True):
        """ Returns a pair of iters; either selection bounds or two copies of
        insert point (forward = False) or insert point and next character
        position (forward = True). """
        if self.rectsel:
            return self.rectsel.get_bounds()
        bounds = self.get_selection_bounds()
        if len(bounds) > 1:
            return bounds
        else:
            start = self.get_iter_at_insert()
            end = start.copy()
            if forward:
                end.forward_char()
            return [start, end]

    def clear_selection(self, iter = None):
        if self.rectsel:
            self.rectsel.cancel()
        else:
            if not iter:
                iter = self.get_iter_at_insert()
            self.place_cursor(iter)

    def get_has_selection(self):
        return self.rectsel or GtkSource.Buffer.get_has_selection(self)

    def get_selection_size(self):
        r = self.rectsel
        if r:
            return [r.right + 1 - r.left, r.bottom + 1 - r.top]
        selection = self.get_selection_bounds()
        if len(selection) == 2:
            if self.selmode == mode.SELBYLINE:
                return abs(selection[1].get_line() - \
                        selection[0].get_line())
            else:
                return abs(selection[1].get_offset() - \
                        selection[0].get_offset())
        else:
            return 0

    def change_complete(self):
        if self.change:
            chng = self.change
            self.change.finalise()
        else:
            chng = None
        if self.rectsel_marks:
            m = self.create_mark(None, self.get_iter_at_insert())
            for start, end in self.rectsel_marks[1:]:
                if chng and (chng.inserted or \
                        chng.deleted_before or chng.deleted_after):
                    self.place_cursor(self.get_iter_at_mark(end))
                    chng.repeat_deletions()
                    if chng.inserted:
                        self.insert_at_cursor(chng.inserted)
                self.delete_mark(start)
                self.delete_mark(end)
            self.rectsel_marks = None
            self.place_cursor(self.get_iter_at_mark(m))
            self.delete_mark(m)
        self.explicit_selection = None
        self.ensure_trailing_nl()
        self.end_user_action()
        #self.set_selmode(0)
        # See comment in git commit 1504f18209a9f38ddf79ea9048668d0fa392347c
        v = self.get_view()
        if v:
            v.cursor_pos_update(self.get_iter_at_insert(),
                    None, v.get_tab_width())

    def __get_lang(self, lang = False):
        if lang == False:
            try:
                return self.get_language()
            except:
                return None
        else:
            return lang

    def get_filetype(self):
        if self.email_ft:
            return "email"
        lang = self.__get_lang()
        if lang:
            return lang.get_name()
        else:
            return None

    def get_filetype_name(self, lang = False):
        lang = self.__get_lang(lang)
        if lang:
            return lang.get_name()
        elif self.email_ft:
            return 'Email'
        else:
            return 'Text'

    def show_language_status(self, win = None, lang = False):
        l = self.get_filetype_name(lang)
        if win:
            win.status_bar.set_filetype(l)
            win.status_bar.set_dos(self.crlf)
        else:
            for v in self.views:
                w = v.get_vixwin()
                if w:
                    w.status_bar.set_filetype(l)
                    w.status_bar.set_dos(self.crlf)

    def set_language(self, lang):
        GtkSource.Buffer.set_language(self, lang)
        self.show_language_status(None, lang)

    def set_filetype(self, ft):
        if ft:
            ft = ft.lower()
        if ft == 'email':
            self.email_ft = True
            ft = None
        else:
            self.email_ft = False
        if not ft or ft == 'text' or ft == 'none':
            if ft == 'none':
                ft = None
            self.set_language(None)
        else:
            lang = langs.get_lang_by_name(ft)
            if lang:
                self.set_language(lang)
            else:
                dialog.warning(self.get_vixwin(),
                        _("Filetype '%s' not recognised") % ft)
                return
        if ft:
            self.__run_hooks(hooks.filetype, ft)
        elif self.email_ft:
            self.__run_hooks(hooks.filetype, 'email')
        else:
            self.__run_hooks(hooks.filetype, 'text')

    def get_fmt_bullet_re(self):
        # This returns an re object, not a string
        if not self.fmt_bul_re:
            r = self.lookup_setting("fmtbulletre",
                    r"(\s*)(" + _BULLETS + r")(\s+)")
            self.fmt_bul_re = re.compile(r)
        return self.fmt_bul_re

    def __run_hooks(self, hook, match_with):
        w = self.get_vixwin()
        if w:
            ex = w.excmd
        elif self.excmd:
            ex = self.excmd
        else:
            ex = ExInterpreter(self, hook.get_settings_priority())
            self.excmd = ex
            # Settings will be applied once buf is added to view/window
        ex.run_hooks(hook, match_with)

    def __run_hooks_idle_cb(self, args):
        self.__run_hooks(args[0], args[1])
        # Only run once at glib idle callback
        return False

    def get_vi_mark(self, name):
        return self.get_mark('vi_' + name)

    def record_vi_mark(self, name, iter):
        if name.startswith('vi_'):
            name = name[3:]
        self.state.marks[name] = [iter.get_line_offset(), iter.get_line()]

    def set_vi_mark(self, name, iter = None):
        if not iter:
            iter = self.get_iter_at_insert()
        mark = self.get_vi_mark(name)
        if mark:
            self.move_mark(mark, iter)
        else:
            self.create_mark('vi_' + name, iter)
            # Creating a mark doesn't cause "mark-set"
            self.record_vi_mark(name, iter)

    def set_scheme_id(self, id):
        scm = scheme.get_scheme(id)
        self.set_style_scheme(scm)
        if scm:
            id = scm.get_id()
        else:
            id = 'classic'
        # Defer to idle handler otherwise caretcolor doesn't override scheme
        GLib.idle_add(self.__run_hooks_idle_cb, (hooks.scheme, id))

    def get_scheme_id(self):
        scm = self.get_style_scheme()
        if scm:
            return scm.get_id()
        else:
            return "classic"

    def set_editable(self, editable = True):
        self.editable = editable
        for v in self.views:
            v.set_editable(editable)

    def get_editable(self):
        v = self.get_view()
        if v:
            return v.get_editable()
        else:
            return self.editable

    def __launch_build(self, build_class, caption, arg = None):
        global _make_buffer_name_index
        if VixnBuffer.make_buf and not VixnBuffer.make_buf.get_modified():
            makebuf = VixnBuffer.make_buf
            makebuf.extns['throwback'].destroy()    # Removes old marks
            s = makebuf.get_start_iter()
            e = makebuf.get_end_iter()
            makebuf.delete(s, e)
            makewin = makebuf.get_vixwin()
        else:
            if self.pristine:
                makebuf = self
                makewin = self.get_vixwin()
            else:
                makebuf = VixnBuffer(self)
                makewin = vixwin.VixnWindow(makebuf, settings.PRIORITY_DEFAULT)
            VixnBuffer.make_buf = makebuf
        makebuf.orig_filename, _make_buffer_name_index = \
                    buffers.generate_filename(self.get_cwd(),
                            caption + "_%d", _make_buffer_name_index)
        m = build_class(makebuf, None, self.get_vixwin(), arg)
        if m.dir:
            makebuf.make = m
            makebuf.set_editable(False)
            makewin.set_current_view(makebuf.get_view())
            makewin.present()
            m.run()
            return True
        else:
            return False

    def launch_make(self, arg = None):
        return self.__launch_build(make.Make, _("Make"), arg)

    def launch_scons(self, arg = None):
        return self.__launch_build(make.SCons, _("SCons"), arg)

    def begin_user_action(self):
        self.user_action_count += 1
        if self.user_action_count > 1 and not self.ignore_user_action:
            if not self.debug_user_action:
                dialog.error(self.get_vixwin(), "user_action_count > 1")
            self.debug_user_action = True
        if self.debug_user_action or self.always_debug_user_action:
            sys.stderr.write("user_action_count increased to " + \
                    self.user_action_count + '\n')
            traceback.print_stack()
        GtkSource.Buffer.begin_user_action(self)

    def end_user_action(self):
        self.user_action_count -= 1
        if self.debug_user_action or self.user_action_count < 0 \
                or self.always_debug_user_action:
            sys.stderr.write("user_action_count decreased to " + \
                    self.user_action_count + '\n')
            traceback.print_stack()
        if not self.user_action_count:
            self.debug_user_action = False
        if self.user_action_count < 0:
            dialog.error(self.get_vixwin(), "user_action_count < 0")
            self.user_action_count = 0
        else:
            GtkSource.Buffer.end_user_action(self)

    def get_wchars(self):
        return self.wchars

    def set_wchars(self, wchars):
        self.wchars = wchars

    def ensure_trailing_nl(self):
        end = self.get_end_iter()
        i = self.get_iter_at_insert()
        move_cursor = i.equal(end)
        end.backward_char()
        if end.get_char() != '\n':
            end.forward_char()
            self.insert(end, '\n')
            if move_cursor:
                i = self.get_end_iter()
                i.backward_char()
                self.place_cursor(i)

    def get_text(self, start, end, hidden = True):
        return GtkSource.Buffer.get_text(self, start, end, hidden)

    def get_textwidth(self, view = None):
        """ Returns the buffer's textwidth, or the given/current view's width
        if tw is not set. """
        tw = self.lookup_setting('tw', 0)
        if not tw:
            if not view:
                view = self.get_view()
            tw = view.get_visible_columns()
        return tw

    def trim(self):
        self.begin_user_action()
        if self.rectsel:
            start, end = self.rectsel.get_bounds()
        else:
            bounds = self.get_selection_bounds()
            if bounds:
                start, end = bounds
            else:
                start = self.get_start_iter()
                end = self.get_end_iter()
        emk = self.create_mark(None, end)
        lmk = self.create_mark(None, start)
        ls = start
        while ls.compare(end) < 0:
            if ls.ends_line():
                ls.forward_line()
                continue
            le = ls.copy()
            le.forward_to_line_end()
            ss = le.copy()
            ss.backward_char()
            c = ss.get_char()
            while c.isspace() and ss.compare(ls) > 0:
                ss.backward_char()
                c = ss.get_char()
            if not c.isspace():
                ss.forward_char()
            c = le.get_char()
            if c != '\n' and c != '\r':
                le.forward_char()
            if ss.compare(le) < 0:
                self.delete(ss, le)
                ls = self.get_iter_at_mark(lmk)
                end = self.get_iter_at_mark(emk)
            ls.forward_line()
            self.delete_mark(lmk)
            lmk = self.create_mark(None, ls)
        self.delete_mark(lmk)
        self.delete_mark(emk)
        self.end_user_action()

def rss(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_BUFFER, mnk = False):
    settings.register(keys, str, get_method, set_method, desc, scope, mnk)

def rsi(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_BUFFER, mnk = False):
    settings.register(keys, int, get_method, set_method, desc, scope, mnk)

def rsb(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_BUFFER, mnk = False):
    settings.register(keys, bool, get_method, set_method,
            desc, scope, mnk)

rsb(('smartcase', 'scs'), None, None, _("Detect upper-case characters in " \
        "search patterns to override ignorecase option"))
rsb(('ignorecase', 'ic'), None, None, _("Ignore case when searching"))
rsb(('syntax', 'syn'),
        VixnBuffer.get_highlight_syntax,
        VixnBuffer.set_highlight_syntax,
        _("Whether to use syntax highlighting"))
rss(('filetype', 'ft', 'lang', 'language'),
        VixnBuffer.get_filetype, VixnBuffer.set_filetype,
        _("File type for syntax highlighting"))
rsi(('textwidth', 'tw'), None, None,
        _("Width to use when formatting text"))
rss('fmtbulletre', None, None,
    _("Regular expression for formatter to detect bullets at start of line"))
rss('cwd', None, None,
        _("Buffer's working directory, set automatically to that of the " \
        "process vixn is run from"))
rss(('scheme', 'colorscheme', 'stylescheme'),
        VixnBuffer.get_scheme_id, VixnBuffer.set_scheme_id,
        _("GtkSourceView style scheme name"))
rss(('wordchars', 'wchars'), VixnBuffer.get_wchars, VixnBuffer.set_wchars,
        _("Characters, in addition to alphanumerics, which are considered " \
        "part of a word"))
rsb('trim', None, None, _("Whether to automatically trim trailing space at " \
        "the ends of lines when saving"))
"""
rsb('fastload', None, None, _("If your files are UTF-8 compliant and don't " \
        "have CRLF line-endings this option makes files load faster by " \
        "skipping the conversion process"),
        settings.SCOPE_NONE)
"""
