""" vixn's interface to D-BUS """

import dbus
import dbus.service
# Enable dbus-glib main loop
if dbus.version >= (0, 80, 0):
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default = True)
else:
    import dbus.glib

from gi.repository import Gtk

from . import grep
from . import msgs
from . import settings
from . import throwback
from .vixdomain import *

session_bus = None

dbus_ok = False

def init_dbus():
    try:
        global session_bus, dbus_ok
        session_bus = dbus.SessionBus()
        dbus_ok = True
        return True
    except:
        msgs.exception(_("Failed to connect to D-BUS"))
        return False

def _request_bus_name(name, replace = False):
    """ Returns:
    BusName object if request succeeds.
    None if another instance owns name.
    Exception if there's an error.
    """
    global dbus     # Doesn't work without this for some unknown reason!
    global dbus_ok
    import dbus.exceptions
    if dbus.version >= (0, 80, 0):
        from dbus.service import BusName
    else:
        # Override standard BusName with __new__ based on 0.80.x's
        # because earlier versions don't allow us to use some flags we need
        from dbus import dbus_bindings
        class BusName(dbus.service.BusName):
            def __new__(cls, name, bus, allow_replacement = False,
                    replace_existing = False, do_not_queue = False):
                if name in bus._bus_names:
                    return bus._bus_names[name]
                name_flags = (
                    (allow_replacement and
                            dbus_bindings.NAME_FLAG_ALLOW_REPLACEMENT or 0) |
                    (replace_existing and
                            dbus_bindings.NAME_FLAG_REPLACE_EXISTING or 0) |
                    (do_not_queue and
                            dbus_bindings.NAME_FLAG_DO_NOT_QUEUE or 0))
                retval = dbus_bindings.bus_request_name(
                        bus.get_connection(), name, name_flags)
                if retval == dbus_bindings.REQUEST_NAME_REPLY_PRIMARY_OWNER:
                    pass
                elif retval == dbus_bindings.REQUEST_NAME_REPLY_IN_QUEUE:
                    pass
                elif retval == dbus_bindings.REQUEST_NAME_REPLY_EXISTS:
                    raise dbus.exceptions.NameExistsException(name)
                elif retval == dbus_bindings.REQUEST_NAME_REPLY_ALREADY_OWNER:
                    pass
                else:
                    raise dbus.exceptions.RuntimeError(_("Requesting bus " \
                            "name %s returned unexpected value %s") %
                            (name, str(retval)))
                bus_name = object.__new__(cls)
                bus_name._bus = bus
                bus_name._name = name
                bus._bus_names[name] = bus_name
                return bus_name

    try:
        name_owner = BusName(name, session_bus, True, replace, True)
    except dbus.exceptions.NameExistsException:
        return None
    except:
        dbus_ok = False
        raise
    return name_owner


def _get_interface(name, path, interface):
    global session_bus
    obj = session_bus.get_object(name, path)
    return dbus.Interface(obj, interface)


def run_vixn_via_dbus(cwd, fgparent, argv, always_reload):
    # Tries to get a remote instance of dbus to run this argv.
    # Returns False if that doesn't work.
    try:
        if not fgparent:
            fgparent = ''
        vixn_interface = _get_interface(BUS_NAME, PATH, INTERFACE)
        vixn_interface.run_vixn(cwd, fgparent, argv, always_reload)
        return True
    except:
        msgs.exception(_("Failed to get existing instance of " + \
                "vixn to handle this request"))
        return False


def start_throwback_session(cwd):
    vixn_interface = _get_interface(BUS_NAME, PATH, INTERFACE)
    s = vixn_interface.start_throwback_session(cwd)
    return s

def add_throwback_line(session, text):
    vixn_interface = _get_interface(BUS_NAME, PATH, INTERFACE)
    vixn_interface.add_throwback_line(session, text)

def throwback_session_finished(session):
    vixn_interface = _get_interface(BUS_NAME, PATH, INTERFACE)
    vixn_interface.throwback_session_finished(session)


class Listener(dbus.service.Object):
    def __init__(self, bus_name, path):
        global session_bus
        if dbus.version >= (0, 80, 0):
            dbus.service.Object.__init__(self, None, path, bus_name)
        else:
            dbus.service.Object.__init__(self, bus_name, path)


class VixnOnDBus(Listener):
    @dbus.service.method(INTERFACE, in_signature = 'ssasb')
    def run_vixn(self, cwd, fgparent, argv, always_reload):
        if fgparent == '':
            fgparent = None
        from . import main
        main.run(cwd, False, fgparent, argv, always_reload)

    @dbus.service.method(INTERFACE, in_signature = 's', out_signature = 'i')
    def start_throwback_session(self, cwd):
        s = grep.open_session(cwd)
        return s

    @dbus.service.method(INTERFACE, in_signature = 'is')
    def add_throwback_line(self, session, text):
        grep.sessions[session].add_line(text)

    @dbus.service.method(INTERFACE, in_signature = 'i')
    def throwback_session_finished(self, session):
        grep.sessions[session].finished()


# This makes sure the main service doesn't get culled by GC
_service = None

def start_main_service(replace = False):
    """ Returns True if starts service, False if another instance has the name,
    or an exception. """
    global _service
    bus_name = _request_bus_name(BUS_NAME, replace)
    if bus_name:
        _service = VixnOnDBus(bus_name, PATH)
        return _service
    else:
        return False


""" Enable parent to wait in foreground until all child buffers are closed. """

def _coded_bus_name(code):
    return BUS_NAME + '.fg' + code

def _coded_path_name(code):
    return PATH + '/fg' + code

class Waiter(Listener):
    def __init__(self, code):
        if not dbus_ok:
            init_dbus()
        full_name = _coded_bus_name(code)
        bus_name = _request_bus_name(full_name, False)
        if not bus_name:
            raise dbus.exceptions.RuntimeError(_("Bus name '%s' already " + \
                    "has an owner") % full_name)
        self.buffers = 0
        Listener.__init__(self, bus_name, _coded_path_name(code))

    @dbus.service.method(INTERFACE)
    def add_buffer(self):
        self.buffers += 1

    @dbus.service.method(INTERFACE)
    def remove_buffer(self):
        self.buffers -= 1
        if self.buffers <= 0:
            Gtk.main_quit()


""" Methods for children of --foreground to send buffer count messages
to parent. """

def _change_buffer_count(code, add):
    iface = _get_interface(_coded_bus_name(code), _coded_path_name(code),
            INTERFACE)
    if add:
        iface.add_buffer()
    else:
        iface.remove_buffer()

def add_buffer(code):
    _change_buffer_count(code, True)

def remove_buffer(code):
    _change_buffer_count(code, False)

