import os
import re
import subprocess
import sys

from gi.repository import GLib, GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GtkSource
from gi.overrides import keysyms as ks
from gi.repository import Pango

from . import bindings
from . import dialog
from . import format
from . import keysyms
from . import misc
from . import mode
from . import msgs
from . import qrec
from . import search
from . import settings
from . import vicmd
from . import vixi18n
from . import vixiter

URL_WITH_PROTOCOL = r'(https?|ftps?|file|git|ssh|git\+ssh)://[^> "\t\n\r\f\v]+'
URL_HOST_TRAILING = r"([-a-z0-9]*\.)+[a-z0-9]{2,}"
URL_HOST = r"(www|ftp)" + URL_HOST_TRAILING
URL_PATH = r"[/?][-A-Za-z0-9_$.+!*(),;:@&=?/~#%]*"
URL_WITHOUT_PROTOCOL = URL_HOST + '(' + URL_PATH + ')?'
URL_RE = "(" + URL_WITH_PROTOCOL + ")|(" + URL_WITHOUT_PROTOCOL + ")"
MAILTO_RE = r"(mailto:)?[-a-z0-9.+]+@" + URL_HOST_TRAILING

def pango_pixels(units):
    return int((units + Pango.SCALE / 2) / Pango.SCALE)

class VixnView(GtkSource.View, settings.SettingsClient):
    """ vixn's version of GtkSource.View """

    def __init__(self, buf, priority = None):
        """ buf may be some other SettingsClient instead. See settings.py
        copy() methods for meaning of priority. """
        if not hasattr(buf, "is_buffer"):
            from . import vixbuf
            buf = vixbuf.VixnBuffer(buf)
        self.label = None
        self.settings = buf.get_vixview_settings().copy(priority)
        self.__set_within_margin(*self.get_set_size())
        GtkSource.View.__init__(self)
        GtkSource.View.set_buffer(self, buf)

        self.applied_font_desc = None
        self.mode = 0
        self.ignore_hug = False
        self.hug_column = 0
        self.url_re = None
        self.url_re_s = URL_RE
        self.mailto_re = None
        self.mailto_re_s = MAILTO_RE
        self.browser = None
        self.mailer = None
        self.qrec = None
        self.auto_wrap = False
        self.auto_wrapping = False
        self.extns = {}
        self.hiline = 1
        #self.skip_cursor_pos_update = False
        self.vicmd = vicmd.ViInterpreter(bindings.defaults, self)

        buf.add_view(self) # Automatically calls update_title()

        self.set_default_mode()

        self.connect("destroy", self.__destroy_cb)
        self.connect("key-press-event", self.__key_pressed)
        self.connect("focus-in-event", self.__focus_in_cb)
        self.connect("focus-out-event", self.__focus_out_cb)
        self.connect("button-press-event", self.__button_press_cb)
        self.connect("map", self.__map_cb)
        self.connect("size-allocate", self.__size_alloc_cb)
        self.connect("parent-set", self.__parent_set)

        self.set_indent_width(4)

        settings.apply_all(self, self.settings)

    # def get_buffer(self):
    # Implemented in base class

    def __parent_set(self, widget, old_parent):
        if not old_parent:
            settings = self.get_settings()
            settings.set_string_property("gtk-key-theme-name", "Emacs", "vixn")

    def __size_alloc_cb(self, widget, alloc):
        self.__set_within_margin(*self.get_visible_size())

    def __set_within_margin(self, w, h):
        m = min(w, h)
        if m:
            self.within_margin = 1.0 / m

    def get_view(self):
        return self

    def get_vixwin(self):
        win = self.get_toplevel()
        if isinstance(win, Gtk.Window):
            return win
        return None

    def get_vixbuf_settings(self):
        return self.get_buffer().get_vixbuf_settings()

    def get_vixview_settings(self):
        return self.settings

    def get_vixwin_settings(self):
        w = self.get_vixwin()
        if w:
            return w.get_vixwin_settings()
        else:
            # If not bound to a window yet we should get window settings from
            # buf; if buf already had a window it will get those
            # settings, otherwise defaults
            return self.get_buffer().get_vixwin_settings()

    def cursor_pos_update(self, iter, col, ts):
        """ col (column) is visible column which may differ from char offset
        due to tabs etc. If col is None, it's worked out from iter and ts. """
        #if self.skip_cursor_pos_update:
        #    self.skip_cursor_pos_update = False
        #    return
        if col == None:
            col = vixiter.get_column(iter, self.get_tab_width())
        if self.ignore_hug:
            self.ignore_hug = False
        else:
            self.hug_column = col
        win = self.get_vixwin()
        if win and self == win.view:
            win.status_bar.set_position(col, iter.get_line())
        return col

    def __check_iter_eol(self, iter, force = False):
        """ If an iter is at the end of a line move it back to last visible
        character. """
        if (force or not self.lookup_setting("virtualedit", False)) and \
                iter.ends_line() and not iter.starts_line() and \
                not (self.mode & mode.SELECT_MASK):
            iter.backward_cursor_position()

    def move_iter_to_column(self, iter, column, space_only = False):
        """ Moves to 'display column' ie a tab can span several display
        columns. Starts at first column. If space_only is True, stop when we
        reach a character that isn't a space or tab.
        Returns actual display column moved to which may not be same as column
        in, if tabstop isn't same as shiftwidth or we reach eol. """
        ts = self.get_tab_width()
        iter.set_visible_line_offset(0)
        n = 0
        while n < column and not iter.ends_line():
            c = iter.get_char()
            if c == '\t':
                n += ts - n % ts
            elif c != ' ' and space_only:
                return n
            else:
                n += 1
            iter.forward_cursor_position()
            self.__check_iter_eol(iter)
        return n

    def move_iter_to_hug_column(self, iter):
        self.move_iter_to_column(iter, self.hug_column)

    def active_for_buffer(self, active = True):
        buf = self.get_buffer()
        if active:
            buf.active_view = self
            # Make sure no other view on same buf is in insert mode to avoid
            # possibility of confusing user_action
            for v in buf.views:
                if v != self and (v.mode & mode.INSERT_MASK):
                    v.set_mode(v.mode & ~mode.INSERT_MASK)
                    break
        elif buf.active_view == self:
            buf.active_view = None

    def __destroy_cb(self, widget = None):
        for e in list(self.extns.values()):
            e.destroy()
        buf = self.get_buffer()
        if self.vicmd:
            self.vicmd.clear_saved_cursor_pos()
            self.vicmd.set_window(None)
        buf.remove_view(self)

    def get_set_size(self):
        return [self.lookup_setting("columns", 80),
                self.lookup_setting("lines", 30)]

    def init_size(self):
        """ Must be called between adding to top-level window and showing
        latter to work correctly. """
        set_size = self.get_set_size()
        columns = os.environ.get("COLUMNS")
        if not columns:
            columns = set_size[0]
        else:
            self.set_setting("columns", columns)
        lines = os.environ.get("LINES")
        if not lines:
            lines = set_size[1]
        else:
            self.set_setting("lines", lines)
        self.apply_font()
        self.apply_columns_and_lines(int(columns), int(lines))

    def __key_pressed(self, widget, event):
        ctrl = (event.get_state() & Gdk.ModifierType.CONTROL_MASK) != 0
        alt = (event.get_state() & Gdk.ModifierType.MOD1_MASK) != 0
        if self.qrec:
            if event.string == 'q' and not ctrl and not alt:
                # FIXME: Shouldn't hardwire 'q' if not bound
                self.cancel_recording()
                return True
            self.qrec.add_keystroke(event)
        if (self.mode & mode.INSERT_MASK):
            if event.keyval == ks.Escape:
                self.set_mode(mode.VI)
                return True
            elif event.keyval == ks.BackSpace and self.get_indent_on_tab():
                # Change key to shift-tab if deleting space at start of line
                # in smarttab mode so we perform an outdent instead
                buf = self.get_buffer()
                i1 = buf.get_iter_at_insert()
                i0 = i1.copy()
                i0.set_line_offset(0)
                if buf.get_text(i0, i1).isspace():
                    event.keyval = ks.Tab
                    event.state = Gdk.ModifierType.SHIFT_MASK
                    try:
                        # Throws RuntimeError: unable to set value for field
                        event.string ='\t'
                    except:
                        pass
                return False
            else:
                return False
        else:
            try:
                k = keysyms.encode_event_key(event.keyval, ctrl, alt)
                if k != None:
                    self.vicmd.feed_key(k, event.string)
            except:
                dialog.exception(self.get_vixwin(),
                        _("Exception handling keystroke"))
            return True
        return False

    def __focus_in_cb(self, widget, event):
        if self.hiline == 1:
            GtkSource.View.set_highlight_current_line(self, False)
        elif self.hiline == 2:
            GtkSource.View.set_highlight_current_line(self, True)
        self.active_for_buffer(True)
        return False

    def __focus_out_cb(self, widget, event):
        if self.hiline == 1:
            GtkSource.View.set_highlight_current_line(self, True)
        elif self.hiline == 2:
            GtkSource.View.set_highlight_current_line(self, False)
        return False

    def set_hiline(self, val):
        self.hiline = val
        focus = self.has_focus()
        if val == 0 or (val == 1 and focus) or (val == 2 and not focus):
            GtkSource.View.set_highlight_current_line(self, False)
        elif val == 3 or (val == 1 and not focus) or (val == 2 and focus):
            GtkSource.View.set_highlight_current_line(self, True)

    def get_hiline(self, val):
        return self.hiline

    def __get_iter_at_button_event(self, event):
        try:
            x, y = self.window_to_buffer_coords( \
                    self.get_window_type(event.window),
                    int(event.x), int(event.y))
            return self.get_iter_at_location(x, y)
        except:
            msgs.exception(_("Unable to get iter at clicked position"))
            return None

    @staticmethod
    def __is_sel_by_word_char(c, cswchars):
        return c.isalnum() or c in cswchars

    def __double_click(self, event, start):
        if not start:
            return False
        end = start.copy()
        n = start.get_line_offset()
        start.set_line_offset(0)
        if not end.ends_line():
            end.forward_to_line_end()
        line = self.get_buffer().get_text(start, end)
        # First look for URI
        if not self.url_re:
            self.url_re = re.compile(self.url_re_s, re.I)
        # Can't use findall because complex regexps cause overlaps
        offset = 0
        umatch = None
        while offset < len(line) and not umatch:
            mtch = self.url_re.search(line, offset)
            if not mtch:
                break
            if mtch.start() <= n and mtch.end() > n:
                umatch = mtch
            else:
                offset += mtch.end()
        if not umatch:
            if not self.mailto_re:
                self.mailto_re = re.compile(self.mailto_re_s, re.I)
            offset = 0
            mmatch = None
            while offset < len(line) and not mmatch:
                mtch = self.mailto_re.search(line, offset)
                if not mtch:
                    break
                if mtch.start() <= n and mtch.end() > n:
                    mmatch = mtch
                else:
                    offset += mtch.end()
        if umatch:
            start.set_line_offset(umatch.start())
            end.set_line_offset(umatch.end())
        elif mmatch:
            start.set_line_offset(mmatch.start())
            end.set_line_offset(mmatch.end())
        else:
            l = len(line)
            m = n
            wchars = self.lookup_setting('wchars', '_')
            while m and self.__is_sel_by_word_char(line[m - 1], wchars):
                m -= 1
            while n < l and self.__is_sel_by_word_char(line[n], wchars):
                n += 1
            start.set_line_offset(m)
            end.set_line_offset(n)
        self.get_buffer().select_range(start, end)
        if (event.get_state() & Gdk.ModifierType.CONTROL_MASK) and \
                (umatch or mmatch):
            uri = self.get_buffer().get_text(start, end)
            if umatch:
                prog = self.get_browser()
            else:
                prog = self.get_mailer()
            if not prog:
                if umatch:
                    msg = _("Unable to find a browser")
                else:
                    msg = _("Unable to find a mail program")
                dialog.warning(self.get_vixwin(), msg)
                return True
            if '%s' in prog:
                prog = prog % uri
            else:
                prog = prog + ' ' + uri
            subprocess.Popen(prog, shell = True)
        return True

    def __triple_click(self, event, iter):
        def set_selbyline(view):
            view.set_selmode(mode.SELBYLINE)
            return False
        GLib.idle_add(set_selbyline, self)
        return False

    def __paste_click(self, clipboard, text, arg):
        if not text:
            return
        event, iter = arg
        w = self.get_vixwin()
        if w and w.status_bar and \
                hasattr(w.status_bar.showing, "set_text_unselected"):
            input_bar = w.status_bar.showing
            wtext = input_bar.get_text()
            i = input_bar.get_position()
            t = wtext[:i] + text + wtext[i + 1:]
            input_bar.set_text_unselected(t, i + len(text))
            return
        if not iter:
            return
        buf = self.get_buffer()
        iua = buf.ignore_user_action
        buf.ignore_user_action = True
        buf.begin_user_action()
        if self.mode & mode.OVERWRITE:
            end = iter.copy()
            end.forward_chars(len(text))
            # Confine replacement to one line
            if end.get_line() != iter.get_line():
                end = iter.copy()
                if not end.ends_line():
                    end.forward_to_line_end()
            buf.delete(iter, end)
        buf.insert(iter, text)
        buf.end_user_action()
        buf.ignore_user_action = iua

    def __button_press_cb(self, widget, event):
        iter = self.__get_iter_at_button_event(event)
        for e in list(self.extns.values()):
            if e.click_event(event, iter):
                return True
        if event.button == 1:
            if event.type == Gdk.EventType._2BUTTON_PRESS:
                return self.__double_click(event, iter)
            elif event.type == Gdk.EventType._3BUTTON_PRESS:
                return self.__triple_click(event, iter)
        elif event.button == 2 and event.type == Gdk.EventType.BUTTON_PRESS:
            Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY).request_text( \
                    self.__paste_click, [event.copy(), iter])
            return True
        return False

    def set_selmode(self, mo, update_buf = True):
        mask = ~mode.SELECT_MASK
        if mo:
            mask &= ~mode.INSERT_MASK
        self.set_mode((self.mode & mask) | mo, update_buf)

    def set_mode(self, mo, update_buf = True):
        """ Previous state of RECORDING flag is preserved """
        mo |= self.mode & mode.RECORDING
        buf = self.get_buffer()
        if mo == self.mode:
            buf.set_selmode(mo)
            return
        was_ins = self.mode & mode.INSERT_MASK
        mo |= self.mode & mode.RO_MASK
        self.mode = mo
        if update_buf:
            buf.set_selmode(mo)
        if was_ins and not (mo & mode.INSERT_MASK):
            buf.change_complete()
            iter = buf.get_iter_at_insert()
            self.__check_iter_eol(iter)
            self.move_cursor_to_iter(iter, True)
        self.update_cursor_state()
        win = self.get_vixwin()
        if win:
            win.status_bar.set_mode(mo)

    def set_default_mode(self):
        from . import vixbuf
        self.set_mode(vixbuf.default_mode)

    def __map_cb(self, widget, event = None):
        self.update_cursor_state()
        return False

    def update_cursor_state(self, bc = None, cc = None, ic = None):
        if not self.get_realized():
            # Wait until we get called again at map-event otherwise
            # set_overwrite causes assertion warnings
            return
        if bc == None:
            bc = self.lookup_setting('bc', True, True)
        mo = self.mode & mode.INSERT_MASK
        self.set_overwrite(mo == mode.OVERWRITE or (bc and not mo))
        if mo:
            if ic:
                c = ic
            else:
                c = self.lookup_setting('inscolor', '#800080', True)
        else:
            if cc:
                c = cc
            else:
                c = self.lookup_setting('curscolor', '#cd0000', True)
        rgb = Gdk.RGBA()
        if rgb.parse(c):
            self.override_cursor(rgb, rgb)

    def set_window(self, window):
        self.vicmd.set_window(window)

    def set_buffer(self, buf):
        old_buf = self.get_buffer()
        GtkSource.View.set_buffer(self, buf)
        if old_buf and hasattr(old_buf, 'remove_view'):
            old_buf.remove_view(self)
        buf.add_view(self)
        if self.vicmd:
            self.vicmd.set_buffer(buf)
        win = self.get_vixwin()
        if win and win.excmd:
            win.excmd.buffer = buf

    def update_title(self):
        buf = self.get_buffer()
        # Somehow this can be a TextBuffer instead of a VixnBuffer
        # when closing window
        if not getattr(buf, 'orig_filename', None):
            return
        self.win_title = buf.filename
        if not self.win_title:
            self.win_title = os.path.basename(buf.orig_filename)
        self.tab_title = os.path.basename(self.win_title)
        if buf.get_modified():
            self.win_title = '* ' + self.win_title
            self.tab_title = '* ' + self.tab_title
        l = len(buf.views)
        if l > 1:
            suffix = ' [%d/%d]' % (buf.views.index(self) + 1, l)
            self.win_title += suffix
            self.tab_title += suffix
        if self.label:
            self.label.set_text(self.tab_title + ' ')
        win = self.get_vixwin()
        if win and win.view == self:
            if not buf.get_modified():
                for v in win.views:
                    if v.get_buffer().get_modified():
                        self.win_title = '+ ' + self.win_title
                        break
            win.set_title(self.win_title)

    def get_em_size(self, columns = 1, lines = 1):
        """ Returns the size [w, h] of the specified number of characters in
        pixels, including line spacing (assuming the setting is the same for
        paragraphs and wrapped lines). Width is approximate for non-monospace
        fonts. """
        pc = self.get_pango_context()
        # See comment for self.modify_font
        if self.applied_font_desc:
            desc = self.applied_font_desc
        else:
            desc = pc.get_font_description()
        metrics = pc.get_metrics(desc, pc.get_language())
        return [pango_pixels(metrics.get_approximate_char_width() * columns),
                pango_pixels((metrics.get_ascent() + metrics.get_descent()) *
                        lines) +
                self.get_pixels_inside_wrap() * lines]

    def get_visible_size(self):
        rect = self.get_visible_rect()
        ems = self.get_em_size()
        return [int(rect.width / ems[0]), int(rect.height / ems[1])]

    def get_visible_columns(self):
        return self.get_visible_size()[0]

    def get_visible_lines(self):
        return self.get_visible_size()[1]

    def get_total_border_size(self):
        """ Returns [w, h] size of all fixed parts of the widget that don't
        contain text, (I think) excluding padding caused by resizing the
        window by a fraction of a character size. """
        # FIXME: The +2 is a bit of a fudge, may depend on theme
        return [self.get_indent() +
                self.get_left_margin() + self.get_right_margin() +
                self.get_border_window_size(Gtk.TextWindowType.LEFT) +
                self.get_border_window_size(Gtk.TextWindowType.RIGHT) + 2,
                self.get_border_window_size(Gtk.TextWindowType.TOP) +
                self.get_border_window_size(Gtk.TextWindowType.BOTTOM) + 2]

    def get_window_size(self):
        allocation = self.get_allocation()
        [w, h] = [allocation.width, allocation.height]
        borders = self.get_total_border_size()
        w -= borders[0]
        h -= borders[1]
        return [w, h]

    def get_columns_and_lines(self, only_if_realized = False):
        """ Gets the number of lines and columns available in the widget. May
        only be approximate. """
        win = self.get_vixwin()
        if win and win.get_realized():
            [w, h] = self.get_window_size()
            em_size = self.get_em_size()
            return [int(w / em_size[0]), int(h / em_size[1])]
        elif not only_if_realized:
            return self.get_set_size()
        else:
            return None

    def apply_columns_and_lines(self, columns, lines):
        """ NB set_size_request() only works before widget is first shown. """
        em_size = self.get_em_size(columns, lines)
        borders = self.get_total_border_size()
        w = em_size[0] + borders[0]
        h = em_size[1] + borders[1]
        if hasattr(self, 'scroller'):
            self.scroller.resize(w, h)
        return [w, h]

    def get_difference_between_self_and_parent(self, win = None):
        if not win:
            win = self.get_vixwin()
        [cw, ch] = self.get_size_request()
        [pw, ph] = win.get_size_request()
        return [pw - cw, ph - ch]

    def modify_font(self, desc):
        """ Reading desc from widget's context returns the old or default
        desc, so we need to remember what we've actually set. """
        GtkSource.View.modify_font(self, desc)
        self.applied_font_desc = desc

    def apply_font(self, font = None, zoom = None):
        win = self.get_vixwin()
        realized = (win != None) and win.get_realized()
        resize = (font != None or zoom != None) \
                and realized and win.view == self
        if not font:
            font = self.lookup_setting("guifont", "Monospace 10", True)
        if not zoom:
            zoom = self.lookup_setting("zoom", 1.0, True)
        if resize:
            [w, h] = self.get_columns_and_lines()
            if realized:
                [dw, dh] = self.get_difference_between_self_and_parent(win)
        # Most likely sort of error is a bad user-supplied font description
        # so change exception type. Bear in mind all views in a window should
        # share this setting
        try:
            desc = Pango.FontDescription(font)
            if zoom != 1:
                desc.set_size(int(desc.get_size() * zoom))
            self.modify_font(desc)
            if resize:
                [w, h] = self.apply_columns_and_lines(w, h)
                win.resize(w + dw, h + dh)
        except:
            dialog.exception(self.get_vixwin(),
                    _("Unable to apply font '%s'") % font)

    def get_columns(self, key = None):
        size = self.get_columns_and_lines(True)
        if size:
            return size[0]
        else:
            return None

    def get_lines(self, key = None):
        size = self.get_columns_and_lines(True)
        if size:
            return size[1]
        else:
            return None

    def __move_mark_to_iter(self, mark, iter, scroll = True):
        buf = self.get_buffer()
        buf.move_mark(mark, iter)
        if scroll:
            self.scroll_to_mark(mark, self.within_margin, False, 0, 0)

    def __move_insert_mark_to_iter(self, mark, iter, save_hug = False,
            scroll = True):
        """ Cursor needs to be treated specially, because moving insert mark
        without moving selection_bound mark creates a selection. """
        if not save_hug:
            self.ignore_hug = True
        if mode.sel_but_not_rect(self.mode):
            self.get_buffer().move_mark(mark, iter)
        else:
            self.get_buffer().place_cursor(iter)
        if scroll:
            self.scroll_to_mark(mark, self.within_margin, False, 0, 0)

    def move_cursor_to_iter(self, iter, save_hug = False, scroll = True):
        self.__move_insert_mark_to_iter(self.get_buffer().get_insert(), iter,
                save_hug, scroll)

    def move_cursor_to_line(self, line,  scroll = True):
        """ line numbering starts from 0. """
        buf = self.get_buffer()
        total = buf.get_line_count()
        if line >= total:
            line = total - 1
        iter = buf.get_iter_at_line(line)
        #self.move_iter_to_hug_column(iter)
        vixiter.iter_to_first_non_space(iter)
        self.move_cursor_to_iter(iter, False, scroll)

    def __get_mark_and_iter_to_move(self):
        buf = self.get_buffer()
        mark = buf.get_insert()
        return (mark, buf.get_iter_at_mark(mark))

    def start_selection(self, selmode = mode.SELECT):
        self.set_mode(selmode)
        buf = self.get_buffer()
        mark = buf.get_insert()
        iter = buf.get_iter_at_mark(mark)
        buf.explicit_selection = iter.copy()
        if selmode == mode.SELECT:
            iter.forward_cursor_position()
            self.__move_insert_mark_to_iter(mark, iter)
        elif selmode == mode.SELBYLINE:
            iter.set_line_offset(0)
            end = iter.copy()
            vixiter.iter_to_eol(end)
            self.get_buffer().select_range(iter, end)
        # SELRECT initiated by buf.set_selmode() via self.set_mode()

    def clear_selection(self):
        self.set_default_mode()
        buf = self.get_buffer()
        if self.vicmd and self.vicmd.saved_cursor_pos:
            self.vicmd.restore_cursor_pos()
        else:
            ins = buf.get_iter_at_insert()
            buf.place_cursor(ins)

    def move_vert(self, lines, scroll = True):
        buf = self.get_buffer()
        mark, iter = self.__get_mark_and_iter_to_move()
        if lines > 0:
            iter.forward_lines(lines)
        elif lines < 0:
            iter.backward_lines(-lines)
        self.ignore_hug = True
        if (self.mode & mode.SELECT_MASK) == mode.SELBYLINE:
            ins = buf.get_iter_at_insert()
            if ins.get_line() > iter.get_line():
                iter.set_line_offset(0)
            elif not iter.starts_line():
                iter.forward_line()
                iter.set_line_offset(0)
        else:
            l1 = iter.get_line()
            self.move_iter_to_hug_column(iter)
            if l1 != iter.get_line():
                # column was beyond end of line so iter has been moved to
                # start of next line. Put it back one.
                iter.backward_cursor_position()
        self.__move_insert_mark_to_iter(mark, iter, False, scroll)

    def move_horiz(self, columns):
        if (self.mode & mode.SELECT_MASK) == mode.SELBYLINE:
            return
        mark, iter = self.__get_mark_and_iter_to_move()
        orig_iter = iter.copy()
        if columns > 0:
            if iter.ends_line():
                return
            iter.forward_cursor_positions(columns)
        elif columns < 0:
            iter.backward_cursor_positions(-columns)
        if iter.get_line() != orig_iter.get_line():
            if columns > 0:
                iter = orig_iter.copy()
                if not iter.ends_line():
                    iter.forward_to_line_end()
            else:
                iter = orig_iter
                iter.set_line_offset(0)
        self.__check_iter_eol(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def move_by_words(self, count, next = None, anynonspace = None):
        mark, iter = self.__get_mark_and_iter_to_move()
        if anynonspace:
            predicate = lambda c: not c.isspace()
        else:
            cset = self.lookup_setting('wchars', '_')
            predicate = lambda c: c.isalnum() or c in cset
        if count < 0:
            move = iter.backward_char
            count = -count
        else:
            move = iter.forward_char
        l = iter.get_line()
        while count:
            # Make sure we're already in a word
            while count and not predicate(iter.get_char()):
                prev = iter.copy()
                if not move() or iter.get_line() != l:
                    count = 0
            if not count:
                iter = prev
                break
            while count and predicate(iter.get_char()):
                prev = iter.copy()
                if not move() or iter.get_line() != l:
                    count = 0
            if not count:
                iter = prev
                break
            count -= 1
        if not next:
            iter = prev
        self.__check_iter_eol(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def move_by_sentences(self, count):
        mark, iter = self.__get_mark_and_iter_to_move()
        if count > 0:
            iter.forward_sentence_ends(count)
        else:
            iter.backward_sentence_starts(-count)
        self.__check_iter_eol(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def move_by_paras(self, count):
        mark, iter = self.__get_mark_and_iter_to_move()
        vixiter.move_iter_by_paras(iter, count)
        self.__check_iter_eol(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def move_to_column(self, column):
        mark, iter = self.__get_mark_and_iter_to_move()
        iter.set_visible_line_offset(column)
        self.__check_iter_eol(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def sol(self):
        mark, iter = self.__get_mark_and_iter_to_move()
        vixiter.iter_to_first_non_space(iter)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def eol(self, on_lf = True):
        """ If on_lf cursor is always moved to LF, otherwise it may be moved
        back to last visible char depending on circumstances. """
        mark, iter = self.__get_mark_and_iter_to_move()
        if not iter.ends_line():
            iter.forward_to_line_end()
        if not on_lf:
            self.__check_iter_eol(iter, True)
        self.__move_insert_mark_to_iter(mark, iter, True)

    def get_iter_at_window_coords(self, x, y, line_offset = 0, hug = False):
        x, y = self.window_to_buffer_coords(Gtk.TextWindowType.WIDGET, x, y)
        iter = self.get_iter_at_location(x, y)
        if line_offset > 0:
            iter.forward_lines(line_offset)
        elif line_offset < 0:
            iter.backward_lines(-line_offset)
        if hug:
            self.move_iter_to_hug_column(iter)
        return iter

    def move_to_window_coords(self, x, y, line_offset = 0, hug = False):
        """ Here 'hug' means whether to honour existing hug. """
        iter = self.get_iter_at_window_coords(x, y, line_offset, hug)
        self.ignore_hug = not hug
        self.__check_iter_eol(iter)
        self.move_cursor_to_iter(iter, not hug)

    def move_to_top_of_window(self, offset = 0):
        self.move_to_window_coords(0, 0, offset, True)

    def move_to_bottom_of_window(self, offset = 0):
        """ Actually moves to one line below bottom of window,
        so correct with offset. """
        self.move_to_window_coords(0, self.get_window_size()[1],
                offset, True)

    def move_to_middle_of_window(self):
        self.move_to_window_coords(0, int(self.get_window_size()[1] / 2),
                0, True)

    def __get_adjustments(self):
        parent = self.get_parent()
        if hasattr(parent, "get_vadjustment"):
            return [parent.get_hadjustment(), parent.get_vadjustment()]
        else:
            return None

    def scroll_by_pixels(self, count):
        scroll_to = self.get_iter_at_window_coords(0, count)
        n_vis_lines = self.get_lines()
        offset = int(self.within_margin * n_vis_lines)
        if count < 0:
            offset = n_vis_lines - offset
        move_to = self.get_iter_at_window_coords(0, count, offset, True)
        self.ignore_hug = True
        buf = self.get_buffer()
        ins_mark = buf.get_insert()
        ins_iter = buf.get_iter_at_mark(ins_mark)
        if (count > 0 and ins_iter.compare(move_to) < 0) or \
                (count < 0 and ins_iter.compare(move_to) > 0):
            self.__check_iter_eol(move_to)
            self.__move_insert_mark_to_iter(ins_mark, move_to, False, False)
        vadj = self.__get_adjustments()[1]
        new_val = vadj.get_value() + count
        allocation = self.get_allocation()
        max_val = vadj.get_upper() - allocation.height
        if new_val < 0:
            new_val = 0
        elif new_val > max_val:
            new_val = max_val
        vadj.set_value(new_val)

    def scroll_by_page(self, count = 1):
        allocation = self.get_allocation()
        self.scroll_by_pixels(count * (allocation.height -
                2 * self.get_em_size()[1]))

    def scroll_half_page(self, count = 1):
        allocation = self.get_allocation()
        scroll_by = self.lookup_setting("scroll", 0)
        if scroll_by:
            scroll_by *= self.get_em_size()[1] * count
        else:
            scroll_by = int(allocation.height / 2) * count
        self.scroll_by_pixels(scroll_by)

    def scroll_by_line(self, count = 1):
        self.scroll_by_pixels(count * self.get_em_size()[1])

    def reposition(self, yalign):
        buf = self.get_buffer()
        mark = buf.get_insert()
        iter = buf.get_iter_at_mark(mark)
        # Work out current screen X pos so it's preserved
        irect = self.get_iter_location(iter)
        vrect = self.get_visible_rect()
        xalign = float((irect.x - vrect.x)) / vrect.width
        self.scroll_to_mark(mark, 0, True, xalign, yalign)

    def show_search_result(self, result):
        """ result is pair of iters marking start and end of match. """
        win = self.get_vixwin()
        if win:
            sbar = win.status_bar
        else:
            sbar = None
        if result:
            self.__check_iter_eol(result[0])
            self.move_cursor_to_iter(result[0], True)
            status =  self.get_buffer().search.status
            if sbar:
                if status == search.WRAPPED_FORWARD:
                    sbar.show_message(_("Search hit BOTTOM, wrapped to TOP"))
                elif status == search.WRAPPED_BACKWARD:
                    sbar.show_message(_("Search hit TOP, wrapped to BOTTOM"))
        elif sbar:
            sbar.show_message(_("No match for search"))

    def scroll_cursor_on_screen(self):
        """ Make sure cursor is within viewport. """
        self.scroll_to_mark(self.get_buffer().get_insert(),
                self.within_margin, False, 0, 0)

    def start_recording(self, reg):
        self.set_mode(mode.RECORDING)
        self.qrec = qrec.Recording(reg)

    def cancel_recording(self):
        self.qrec = None
        self.mode &= ~mode.RECORDING
        self.get_vixwin().status_bar.set_mode(self.mode)

    def set_block_cursor(self, bc):
        self.update_cursor_state(bc = bc)

    def set_insert_color(self, color):
        self.update_cursor_state(ic = color)

    def set_cursor_color(self, color):
        self.update_cursor_state(cc = color)

    def set_editable(self, editable):
        if editable:
            self.set_mode(self.mode & ~mode.READONLY)
        else:
            self.set_mode(self.mode | mode.READONLY)
        GtkSource.View.set_editable(self, editable)

    def get_wrap(self):
        return self.get_wrap_mode() != Gtk.WrapMode.NONE

    def set_wrap(self, wm):
        if wm:
            wm = Gtk.WrapMode.CHAR
        else:
            wm = Gtk.WrapMode.NONE
        self.set_wrap_mode(wm)

    def get_url_re(self):
        return self.url_re_s

    def set_url_re(self, rex):
        rexc = re.compile(rex, re.I)
        if rexc:
            self.url_re_s = rex
            self.url_re = rexc

    def get_mailto_re(self):
        return self.mailto_re_s

    def set_mailto_re(self, rex):
        rexc = re.compile(rex, re.I)
        if rexc:
            self.mailto_re_s = rex
            self.mailto_re = rexc

    @staticmethod
    def __get_prog(candidates, env = None):
        if env:
            env = os.environ.get(env)
            if env:
                candidates = [env] + candidates
        return misc.find_exe_in_path(candidates)

    def get_browser(self):
        if not self.browser:
            self.browser = self.__get_prog(['x-www-browser', 'firefox',
                    'iceweasel', 'opera', 'galeon', 'konqueror', 'epiphany'],
                    "BROWSER")
        return self.browser

    def set_browser(self, browser):
        self.browser = browser

    def get_mailer(self):
        if not self.mailer:
            self.mailer = self.__get_prog(['claws-mail', 'thunderbird',
                    'icedove', 'evolution', 'kmail'])
        return self.mailer

    def set_mailer(self, mailer):
        self.mailer = mailer

    def set_auto_wrap(self, aw):
        self.auto_wrap = aw

    def get_auto_wrap(self):
        return self.auto_wrap

    def wrap_on_change(self):
        if self.auto_wrapping:
            return False
        self.auto_wrapping = True
        result = self.__wrap_on_change()
        self.auto_wrapping = False
        return result

    def __wrap_on_change(self):
        if not self.auto_wrap or not (self.mode & mode.INSERT_MASK):
            return False
        buf = self.get_buffer()
        it = buf.get_iter_at_insert()
        # Most editors only wrap when appending to a line
        if not it.ends_line():
            return False
        tw = buf.get_textwidth(self)
        if not tw:
            return False
        ts = self.get_tab_width()
        col = vixiter.get_column(it, ts)
        if col < tw:
            return False
        start = it.copy()
        start.set_line_offset(0)
        while vixiter.get_column(it, ts) >= tw:
            if not it.backward_find_char(lambda x, y: x == y, ' ', start):
                return False
        m = buf.create_mark(None, it, False)
        after = it.copy()
        after.forward_cursor_position()
        buf.delete(it, after)
        it = buf.get_iter_at_mark(m)
        buf.insert(it, '\n')
        buf.delete_mark(m)
        return True


class ScrolledView(Gtk.ScrolledWindow):
    """ A VixnView with scrollbars. The actual VixnView widget is attribute
    'view'. """
    def __init__(self, view_or_buffer, priority = None):
        """ Create either with single arg an existing VixnView or same args as
        VixnView's constructor to create a new view. """
        if isinstance(view_or_buffer, VixnView):
            self.view = view_or_buffer
        else:
            self.view = VixnView(view_or_buffer, priority)
        GObject.GObject.__init__(self)
        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.ALWAYS)
        self.add(self.view)
        self.view.scroller = self

    def resize(self, w, h):
        self.set_min_content_width(w)
        self.set_min_content_height(h)
        GLib.idle_add(self.relax_min_size)

    def relax_min_size(self):
        self.set_min_content_width(-1)
        self.set_min_content_height(-1)
        return False

    def get_vixbuf_settings(self):
        return self.view.get_vixbuf_settings()

    def get_vixview_settings(self):
        return self.view.settings

    def get_vixwin_settings(self):
        return self.view.get_vixwin_settings()


def rsi(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_VIEW, mnk = False):
    settings.register(keys, int, get_method, set_method, desc, scope, mnk)

def rsb(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_VIEW, mnk = False):
    settings.register(keys, bool, get_method, set_method,
            desc, scope, mnk)

def rss(keys, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_VIEW, mnk = False):
    settings.register(keys, str, get_method, set_method, desc, scope, mnk)

rsi(('tabstop', 'ts'),
        VixnView.get_tab_width, VixnView.set_tab_width,
        _("Tabs are displayed as equivalent to this number of spaces"))
rsi(('shiftwidth', 'sw'), VixnView.get_indent_width, VixnView.set_indent_width,
        _("Indenting and outdenting remove/add this much whitespace"))
rsi(('scroll', 'scr'), None, None, _("How many lines C-D and C-U scroll by; " \
        "unset or 0 for half window height"))
rsb(('autoindent', 'ai', 'smartindent', 'si', 'cindent', 'cin'),
        VixnView.get_auto_indent, VixnView.set_auto_indent,
        _("Whether to use auto indentation"))
rsb(('virtualedit', 've'), None, None,
        _("Whether to allow the cursor to be positioned after the " \
                "last character on a line"))
rsb(('expandtabs', 'et'), VixnView.get_insert_spaces_instead_of_tabs,
        VixnView.set_insert_spaces_instead_of_tabs,
        _("Insert 'tabstop' number of spaces instead of a tab character when " \
        "Tab is pressed in Insert mode"))
rsb(('smarttab', 'sta'), VixnView.get_indent_on_tab, VixnView.set_indent_on_tab,
        _("Insert 'shiftwidth' instead of 'tabstop' number of spaces when " \
        "Tab is pressed in Insert mode"))
rsb(('blockcursor', 'bc'), None, VixnView.set_block_cursor,
        _("Use block cursor when in vi command mode"))
rss(('insertcolor', 'inscolor', 'insertcolour', 'inscolour'),
        None, VixnView.set_insert_color,
        _("Colour of the insert caret/cursor in #rgb format"))
rss(('cursorcolour', 'cursorcolor', 'curscolour', 'curscolor'),
        None, VixnView.set_cursor_color,
        _("Colour (#rgb) of cursor when not in insert mode."))
rsb('wrap', VixnView.get_wrap, VixnView.set_wrap,
        _("Whether to wrap the display of long lines to fit the view; " \
        "otherwise scroll the view"))
rss(('urlre'), VixnView.get_url_re, VixnView.set_url_re,
        _("Regular expression for matching URLs (case insensitive)"))
rss(('mailtore'), VixnView.get_mailto_re, VixnView.set_mailto_re,
        _("Regular expression for matching email addresses " \
        "or mailto URIs (case insensitive)"))
rss(('browser'), VixnView.get_browser, VixnView.set_browser,
        _("Browser to launch when a URL is Ctrl-double-clicked. May contain " \
        "'%s' to substitute with the URL, otherwise it's added to the end"))
rss(('mailer'), VixnView.get_mailer, VixnView.set_mailer,
        _("Mail program to launch when an email address or mailto: is " \
        "Ctrl-double-clicked. May contain '%s' to substitute with the " \
        "address, otherwise it's added to the end"))
rsb(('autowrap', 'aw'), VixnView.get_auto_wrap, VixnView.set_auto_wrap,
        _("Whether to insert line breaks when the cursor reaches the end of " \
        "a line as defined by the buffer's textwidth setting. If textwidth " \
        "is not set wrap at the window edge. Will not wrap if there is no " \
        "whitespace in the line."))
rsi(('highlightline', 'hiline', 'hlline', 'hll'),
        VixnView.get_hiline, VixnView.set_hiline,
        _("Whether to highlight the line containing the cursor: " \
        "0 = Never, 1 = Only when unfocused, 2 = Only when focused, " \
        "3 = Always"))

# All views in a window should share these settings
rsi(('columns', 'co'), VixnView.get_columns, None,
        _("Initial number of columns (approx)"), settings.SCOPE_VIEW)
rsi('lines', VixnView.get_lines, None,
        _("Initial number of lines (approx)"), settings.SCOPE_VIEW)
