from gi.repository import GObject, GLib, Gdk, Gtk

import struct

from . import bottombar
from . import closebutton
from . import dialog
from . import menus
from . import mode
from . import msgs
from . import settings
from . import shortcuts
from . import vicmd
from . import vixi18n
from . import vixview

# These seem to be missing from Pango gi
SCALE_XX_LARGE = 1.728
SCALE_XX_SMALL = 0.5787037037037

windows = []

changed_key_theme = False

VTAB = 1448362306

class VixnWindow(Gtk.Window, settings.SettingsClient):
    """ A top-level vixn window. Currently holds one view; may hold multiple
    tabs in future. """
    def __init__(self, v_b_or_s, priority = None, create_tab = True):
        """ v_b_or_s is one of VixnView, ScrolledView, VixnBuffer,
        SettingsDefaults. See settings.py copy() methods for meaning of
        priority. """
        self.is_window = True
        self.excmd = None
        self.views = []
        self.destroying = False
        self.excmd = None

        GObject.GObject.__init__(self)

        self.accel_group = Gtk.AccelGroup()
        self.add_accel_group(self.accel_group)
        [self.menu_bar, self.menus_dict, self.items_dict ] = \
                menus.create_menu_tree(self, self.accel_group)
        shortcuts.connect_handler()

        self.status_bar = bottombar.BottomBar()
        self.notebook = Gtk.Notebook()
        self.notebook.set_group_name("vixn")
        self.notebook.set_show_border(False)
        self.notebook.set_scrollable(True)

        self.connect("delete_event", self.close_request)
        self.connect("destroy", self.__destroy)
        self.notebook.connect("switch-page", self.__switch_page_cb)
        self.notebook.connect("page-removed", self.__page_removed_cb)
        self.notebook.connect("page-reordered", self.__page_reordered_cb)
        self.notebook.connect("page-added", self.__page_added_cb)
        self.notebook.connect("create-window", self.__create_window_cb)

        vbox = Gtk.VBox()
        vbox.pack_start(self.menu_bar, False, True, 0)
        vbox.pack_start(self.notebook, True, True, 0)
        vbox.pack_start(self.status_bar, False, False, 2)

        if create_tab:
            self.new_tab(v_b_or_s, priority)
        self.settings = v_b_or_s.get_vixwin_settings().copy(priority)
        settings.mark_all_unapplied(self.settings)

        self.menu_bar.show()
        self.notebook.show_all()
        self.status_bar.show()

        vbox.show()
        self.add(vbox)
        if create_tab:
            self.view.init_size()
            self.view.set_window(self)
            self.set_title(self.view.win_title)

        global windows
        windows.append(self)

        settings.apply_all(self, self.settings)

    def new_tab(self, v_b_or_s, priority = None):
        if isinstance(v_b_or_s, vixview.ScrolledView):
            self.viewport = v_b_or_s
        else:
            self.viewport = vixview.ScrolledView(v_b_or_s, priority)
        v = self.viewport.view
        self.view = v
        v.vicmd.set_window(self)
        # If this isn't the first view it needs font set from window
        if self.views:
            v.apply_font()

        buf = self.view.get_buffer()
        if not self.excmd:
            self.excmd = buf.excmd
        if self.excmd:
            self.excmd.set_window(self)
            buf.excmd = None
        else:
            from . import excmd
            self.excmd = excmd.ExInterpreter(self)

        if not v.label:
            v.label_grid = Gtk.Grid()
            v.label_ebox = Gtk.EventBox()
            v.label_ebox.set_visible_window(False)
            v.label_grid.attach(v.label_ebox, 0, 0, 1, 1);
            v.label = Gtk.Label()
            v.label_ebox.add(v.label)
            v.closebutton = closebutton.CloseButton()
            v.closebutton.connect("clicked", self.__close_tab_cb, self.viewport)
            v.closebutton.set_property("expand", False)
            v.label_grid.attach(v.closebutton, 1, 0, 1, 1);
            v.label_grid.show_all()
        self.viewport.show_all()
        self.notebook.append_page(self.viewport, v.label_grid)
        v.label_grid.get_parent().set_can_focus(False)
        # By now page-added cb has added page to views
        self.notebook.set_current_page(len(self.views) - 1)

        GLib.idle_add(v.scroll_cursor_on_screen)

    def set_current_view(self, view):
        self.notebook.set_current_page(self.notebook.page_num(view))

    def __switch_page_cb(self, notebook, page, num):
        self.viewport = page
        v = page.view
        self.view = v
        self.excmd.set_view(v)
        self.status_bar.set_mode(v.mode)
        buf = v.get_buffer()
        i = buf.get_iter_at_mark(buf.get_insert())
        self.status_bar.set_position_from_iter(i)
        buf.show_language_status(self)
        v.grab_focus()
        v.update_title()

    @staticmethod
    def __close_tab_cb(btn, viewport):
        view = viewport.view
        buf = view.get_buffer()
        if not buf.get_modified() or len(buf.views) > 1 or \
                not dialog.dcs_one_file(buf, view.get_vixwin()):
            view.get_vixwin().destroy_view(view)

    def destroy_view(self, view):
        self.notebook.remove_page(view.page_num)
        view.label_grid.destroy()
        view.get_parent().destroy()

    def __remove_and_reorder(self, num):
        del self.views[num]
        for n in range(num, len(self.views)):
            self.views[n].page_num = n

    def __insert_and_reorder(self, num, page):
        self.views.insert(num, page)
        for n in range(num, len(self.views)):
            self.views[n].page_num = n

    def __page_removed_cb(self, notebook, viewport, num):
        v = viewport.view
        v.disconnect(v.ddr_tag)
        self.__remove_and_reorder(num)
        if not self.destroying and not len(self.views):
            self.destroy()
        else:
            self.view.update_title()

    def __page_reordered_cb(self, notebook, viewport, num):
        v = viewport.view
        old = v.page_num
        self.__remove_and_reorder(old)
        self.__insert_and_reorder(num, v)

    def __page_added_cb(self, notebook, viewport, num):
        self.viewport = viewport
        v = viewport.view
        v.set_window(self)
        self.view = v
        self.__insert_and_reorder(num, v)
        notebook.set_tab_reorderable(viewport, True)
        notebook.set_tab_detachable(viewport, True)
        if not self.excmd:
            from . import excmd
            self.excmd = excmd.ExInterpreter(self)
        tl = v.drag_dest_get_target_list()
        tl.add(Gdk.Atom.intern("GTK_NOTEBOOK_TAB", False),
                Gtk.TargetFlags.SAME_APP, VTAB)
        v.ddr_tag = v.connect("drag-data-received",
                self.__drag_data_received_cb)

    def __drag_data_received_cb(self, view, dc, x, y, seldata, info, when):
        if info == VTAB:
            vp_id = struct.unpack('P', seldata.get_data())[0]
            global windows
            success = False
            for w in windows:
                for v in w.views:
                    if hash(v.scroller) == vp_id:
                        v.get_vixwin().notebook.remove_page(v.page_num)
                        self.notebook.append_page(v.scroller, v.label_grid)
                        self.notebook.set_current_page(len(self.views) - 1)
                        success = True
                        break
                if success:
                    break
            dc.finish(success, True, when)

    def __create_window_cb(self, notebook, viewport, x, y):
        win = VixnWindow(viewport, None, False)
        win.set_position(Gtk.WindowPosition.MOUSE)
        win.set_default_size(*viewport.view.get_vixwin().get_size())
        win.present()
        return win.notebook

    def get_buffer(self):
        return self.view.get_buffer()

    def get_view(self):
        return self.view

    def get_vixwin(self):
        return self

    def get_vixbuf_settings(self):
        return self.get_buffer().get_vixbuf_settings()

    def get_vixview_settings(self):
        return self.get_view().get_vixview_settings()

    def get_vixwin_settings(self):
        return self.settings

    def close_request(self, widget = None, event = None):
        result = False
        for buf in [v.get_buffer() for v in self.views]:
            if buf.get_modified():
                # If buf has another view in a different window we don't
                # need to DCS it
                for w in [ov.get_vixwin() for ov in buf.views]:
                    if w != self:
                        break
                else:
                    if len(self.views) == 1:
                        result = dialog.dcs_one_file(buf, self)
                    else:
                        result = dialog.dcs_win_multi_files(self)
                    break
        return result

    def __destroy(self, widget = None):
        self.destroying = True
        global windows
        try:
            del windows[windows.index(self)]
        except:
            msgs.error("window %s not in global list" % self)

    """
    def show(self):
        Gtk.Window.show(self)
        global changed_key_theme
        if not changed_key_theme:
            settings = self.view.get_settings()
            settings.set_string_property("gtk-key-theme-name", "Emacs", "vixn")
            changed_key_theme = True
    """

    def activate_vi_method(self, gobject, method_name):
        method = vicmd.commands.get(method_name)
        if method:
            self.view.vicmd.execute(None, method)
        else:
            msgs.warning(_("No such vi method '%s'") % method_name)

    def activate_ex_method(self, gobject, method_name):
        self.excmd.execute(method_name)

    def get_font(self):
        desc = self.view.get_pango_context().get_font_description()
        zoom = self.lookup_setting("zoom", 1.0)
        if zoom != 1.0:
            desc = desc.copy_static()
            desc.set_size(int(desc.get_size() / zoom))
        return desc.to_string()

    def apply_font(self, font = None, zoom = None):
        if zoom != None:
            self.set_setting("zoom", zoom)
        for v in self.views:
            v.apply_font(font, zoom)

    def zoom_in(self):
        zoom = self.__lookup_zoom()
        if zoom < SCALE_XX_LARGE * 1.2 * 1.2:
            self.apply_font(zoom = zoom * 1.2)

    def zoom_out(self):
        zoom = self.__lookup_zoom()
        if zoom > SCALE_XX_SMALL / 1.2 / 1.2:
            self.apply_font(zoom = zoom / 1.2)

    def zoom100(self):
        zoom = self.__lookup_zoom()
        if zoom != 1:
            self.apply_font(zoom = 1.0)

    def __lookup_zoom(self):
        return self.lookup_setting("zoom", 1.0, True)



def quit_all():
    global windows
    while len(windows):
        windows[-1].destroy()
    Gtk.main_quit()



def rs(keys, type = str, get_method = None, set_method = None,
        desc = None, scope = settings.SCOPE_WINDOW, mnk = False):
    settings.register(keys, type,
            get_method, set_method, desc, scope, mnk)

# You might expect these to apply to a view rather than a window, but if
# a window has multiple views they should share these settings
rs(('guifont', 'gfn', 'font'), str,
        VixnWindow.get_font, VixnWindow.apply_font,
        _("Main font used for text"))
rs('zoom', float, desc = _("Zoom factor for font (1.0 = normal)"))

